----------------------------------------------------------------------------
-- Module Name:        IASW_COMMON - package specification
-- Project:            INTEGRAL, JEM-X, DPE
-- Author:             Grzegorz Juchnikowski
-- Creation Date:      16.III.1998
-- Change History:
--    (Version/Date/Author/Description)
--    see Release Notes - the separate document
-- Function:
--    (Say what the component does)
--    Provides commonly used constants and types
-- Interfaces:
--    (Define the inputs and outputs)
-- Dependencies:
--    (Describe the preconditions for using this component)
-- Processing:
--    (Summarise the processing using pseudo-code or PDL)
----------------------------------------------------------------------------

with SYSTEM;
with UNCHECKED_CONVERSION;
with CSSW_COMMON_TYPES;
with CSSW_IF_TYPES;

package IASW_COMMON is

   package CIT renames CSSW_IF_TYPES;
   package CCT renames CSSW_COMMON_TYPES;

   -- Common exceptions

   EX_SW_PARAM_ERROR_ID,
   EX_SW_PARAM_ERROR_LENGTH,
   EX_SW_PARAM_ERROR_VALUE
      : exception;

   --K_FAST_MEM_COPY : constant BOOLEAN := FALSE;
   --deleted in ver.1.62 (this variable no more used)

   -- Common types

   type T_ARRAY_OF_WORDS is array(INTEGER range<>) of CCT.T_UNSIGNED_16;

   subtype T_COMPRESSED_ENERGY_TABLE is T_ARRAY_OF_WORDS(1..256);

   type T_PACKED_ARRAY_OF_BYTES is array(INTEGER range<>) of CCT.T_UNSIGNED_8;
   pragma pack(T_PACKED_ARRAY_OF_BYTES);

   subtype T_BIT_NO is INTEGER range 0..15;
   type T_16_BITS is array(T_BIT_NO) of BOOLEAN;
   pragma pack(T_16_BITS);

   subtype T_PACKET_TYPE is CCT.T_UNSIGNED_4;
   subtype T_PACKET_SUBTYPE is CCT.T_UNSIGNED_4;

   type T_ON_RQ_TM_HEADER is record
      TYP : IASW_COMMON.T_PACKET_TYPE;
      SUB : IASW_COMMON.T_PACKET_SUBTYPE;
      SPARE : CCT.T_UNSIGNED_8;
      TIME1,TIME2 : CCT.T_UNSIGNED_8;
      -- the 16-bits field TIME is split into two bytes
      -- to facilitate manipulations.
   end record;
   for T_ON_RQ_TM_HEADER use record
      TYP    at 0 range 0..3;
      SUB    at 0 range 4..7;
      SPARE  at 0 range 8..15;
      TIME1  at 1 range 0..7;
      TIME2  at 1 range 8..15;
   end record;

   subtype T_TID is CCT.T_UNSIGNED_8;
   subtype T_FID is CCT.T_UNSIGNED_8;

   type T_PARAM_ID is record
       TID: T_TID;
       FID: T_FID;
   end record;
   for T_PARAM_ID use record
       TID at 0 range 0..7;
       FID at 0 range 8..15;
   end record;

   K_MAX_TC_PARAM_VAL_LEN : constant INTEGER := CIT.T_TC_APPL_DATA'length
      - T_PARAM_ID'size / 16;

   K_MAX_TM_PARAM_VAL_LEN : constant INTEGER := CIT.T_SCIENCE_DATA_FIELD'length
      - T_ON_RQ_TM_HEADER'size / 16
      - T_PARAM_ID'size / 16 - 1;    -- (-1 for CRC)

   subtype T_TC_PARAM_VAL is T_ARRAY_OF_WORDS(1..K_MAX_TC_PARAM_VAL_LEN);
   subtype T_TM_PARAM_VAL is T_ARRAY_OF_WORDS(1..K_MAX_TM_PARAM_VAL_LEN);
   subtype T_SW_PARAM_LEN is INTEGER;

   -- APID's

   APID_BASE  : constant INTEGER := CIT.K_CSSW_APID;

   APID_TC : constant CIT.T_IASW_APID := APID_BASE + 1;

   APID_TM_SC_DEBUG           : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#1000101#;
   APID_TM_SC_LSL_DUMP        : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#1000110#;
   APID_TM_SC_TEST            : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#1000111#;


   --APID_TM_SC_IDLE            : constant CIT.T_IASW_APID
   --                                         := APID_BASE + 2#1000000#;  --not used
   --deleted in ver.1.62 (this constant not used)


   APID_TM_SC_COUNT_RATE      : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#1111000#;
   APID_TM_SC_CALIBRATION     : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#1000001#;
   APID_TM_SC_DIAGNOSTIC      : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#1000010#;
   APID_TM_SC_FULL            : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#1001000#;
   APID_TM_SC_SPECT_TIM       : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#1010000#;
   APID_TM_SC_RESTRICTED      : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#1011000#;
   APID_TM_SC_TIMING          : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#1100000#;
   APID_TM_SC_SPECTRUM        : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#1101000#;
   APID_TM_5_4                : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#0000101#;
   APID_TM_6_2                : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#0000110#;
   APID_TM_6_3                : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#0000110#;
   APID_TM_9_1                : constant CIT.T_IASW_APID
                                            := APID_BASE + 2#0001001#;

   -- Task identifiers:

   TASK_MAIN_ID        : constant CIT.T_IASW_TASK_ID
                               := CIT.T_IASW_TASK_ID'FIRST;
   TASK_TC_ID          : constant CIT.T_IASW_TASK_ID
                               := CIT.T_IASW_TASK_ID'FIRST+1;
   TASK_HSL_ID         : constant CIT.T_IASW_TASK_ID
                               := CIT.T_IASW_TASK_ID'FIRST+2;
   TASK_SC_OUT_ID      : constant CIT.T_IASW_TASK_ID
                               := CIT.T_IASW_TASK_ID'FIRST+3;
   TASK_HK_ID          : constant CIT.T_IASW_TASK_ID
                               := CIT.T_IASW_TASK_ID'FIRST+4;
   TASK_REQ_OUT_ID     : constant CIT.T_IASW_TASK_ID
                               := CIT.T_IASW_TASK_ID'FIRST+5;
   TASK_STATE_ID       : constant CIT.T_IASW_TASK_ID
                               := CIT.T_IASW_TASK_ID'FIRST+6;

   type T_TASK_NO is (  -- it is used to construct array of tasks
      TASK_MAIN_NO,
      TASK_TC_NO,
      TASK_HSL_NO,
      TASK_SC_OUT_NO,
      TASK_HK_NO,
      TASK_spare,
      TASK_REQ_OUT_NO,
      TASK_STATE_NO
   );

   -- Counters of activity of tasks.
   type T_TASK_ALIVE_CNT is array(T_TASK_NO) of CCT.T_UNSIGNED_8;
   pragma pack(T_TASK_ALIVE_CNT);

   -- Task priorities:
   TASK_MAIN_PRIORITY        : constant CIT.T_IASW_TASK_PRIORITY := 15;
   TASK_TC_PRIORITY          : constant CIT.T_IASW_TASK_PRIORITY := 11;
   TASK_HSL_PRIORITY         : constant CIT.T_IASW_TASK_PRIORITY := 4;
   TASK_SC_OUT_PRIORITY      : constant CIT.T_IASW_TASK_PRIORITY := 3;
   TASK_HK_PRIORITY          : constant CIT.T_IASW_TASK_PRIORITY := 13;
   TASK_REQ_OUT_PRIORITY     : constant CIT.T_IASW_TASK_PRIORITY := 10;
   TASK_STATE_PRIORITY       : constant CIT.T_IASW_TASK_PRIORITY := 12;

   -- Events which should have the same numbers in all tasks
   MRTU_ACQUISITION_COMPLETED_EVENT  : constant CIT.T_EVENT_NUMBER := 16;
   LSL_EOT_EVENT                     : constant CIT.T_EVENT_NUMBER := 15;
   LSL_COMPLETED_EVENT               : constant CIT.T_EVENT_NUMBER := 13;
   LSL_PARITY_ERROR_EVENT            : constant CIT.T_EVENT_NUMBER := 12;
   LSL_OVERRUN_ERROR_EVENT           : constant CIT.T_EVENT_NUMBER := 11;
   LSL_FRAMING_ERROR_EVENT           : constant CIT.T_EVENT_NUMBER := 10;

   -- types related to broadcast packet:

   type T_BCPKT_OBT is record
      HIGH, LOW : CCT.T_UNSIGNED_16;
   end record;

   subtype T_FLAG is CCT.T_UNSIGNED_1;

   type T_PID is record
      REVOLUTION:  CCT.T_UNSIGNED_16;
      POINTING:    CCT.T_UNSIGNED_16;
   end record;

   type T_BROADCAST_PACKET is record
      NEXT_BELT_ENTRY:     T_BCPKT_OBT;
      NEXT_BELT_EXIT:      T_BCPKT_OBT;
      NEXT_ECLIPSE_ENTRY:  T_BCPKT_OBT;
      NEXT_ECLIPSE_EXIT:   T_BCPKT_OBT;
      DRMC:                T_FLAG;
      PID:                 T_PID;
      PDURATION:           CIT.T_U32_BITS;
      TM_SHARE_SPI:        CCT.T_UNSIGNED_8;
      TM_SHARE_IBIS:       CCT.T_UNSIGNED_8;
      TM_SHARE_JEMX1:      CCT.T_UNSIGNED_8;
      TM_SHARE_JEMX2:      CCT.T_UNSIGNED_8;
      TM_SHARE_OMC:        CCT.T_UNSIGNED_8;
      SWITCH_OFF_IMM:      T_FLAG;
      GS_HAND_OVER:        T_FLAG;
      RAD_MONITOR_1:       CCT.T_UNSIGNED_16;
      RAD_MONITOR_2:       CCT.T_UNSIGNED_16;
      RAD_MONITOR_3:       CCT.T_UNSIGNED_16;
      OTF:                 T_FLAG;
      AOCS_MODE:           CCT.T_UNSIGNED_4;
      ESAM:                T_FLAG;
      AOCS_SUBMODE:        CCT.T_UNSIGNED_5;
      SPARE3:              T_ARRAY_OF_WORDS(1..3);
   end record;
   for T_BROADCAST_PACKET use record
      NEXT_BELT_ENTRY     at 0 range 0..31;
      NEXT_BELT_EXIT      at 2 range 0..31;
      NEXT_ECLIPSE_ENTRY  at 4 range 0..31;
      NEXT_ECLIPSE_EXIT   at 6 range 0..31;
      DRMC                at 8 range 0..0;
      -- spare            at 8 range 1..15
      -- spare 4 words    at 0 range 0..63;
      PID                 at 13 range 0..31;
      PDURATION           at 15 range 0..31;
      TM_SHARE_SPI        at 17 range 0..7;
      TM_SHARE_IBIS       at 17 range 8..15;
      TM_SHARE_JEMX1      at 18 range 0..7;
      TM_SHARE_JEMX2      at 18 range 8..15;
      TM_SHARE_OMC        at 19 range 0..7;
      SWITCH_OFF_IMM      at 19 range 8..8;
      GS_HAND_OVER        at 19 range 9..9;
      -- spare            at 19 range 10..15
      -- spare 6 words    at 20 range 0..95;
      RAD_MONITOR_1       at 26 range 0..15;
      RAD_MONITOR_2       at 27 range 0..15;
      RAD_MONITOR_3       at 28 range 0..15;
      OTF                 at 29 range 0..0;
      -- spare            at 29 range 1..11
      AOCS_MODE           at 29 range 12..15;
      -- spare            at 30 range 0..1
      ESAM                at 30 range 2..2;
      -- spare            at 30 range 3..10
      AOCS_SUBMODE        at 30 range 11..15;
      SPARE3              at 31 range 0..47;
   end record;

   -- this record is element of IASW house-keeping
   -- its fields are filled by a set of procedures GET_DPE_STATE
   -- in differente packages.  all these procedures are called in
   -- IASW_HK.DPE_STATE_TO_HK, which is called in IASW_MAIN.TASK_MAIN.
   type T_DPE_STATE is record
      EVENT_SIZE        : CCT.T_UNSIGNED_16;
      SHUTDOWN_LEVEL    : T_PACKED_ARRAY_OF_BYTES(1..2);
      SCF               : T_PACKED_ARRAY_OF_BYTES(1..2);
      ACTIVE_SCF        : INTEGER;   -- 1 or 2
      GF                : T_PACKED_ARRAY_OF_BYTES(1..2);
      HSL_BUFFER_SIZE   : CCT.T_UNSIGNED_16;
      HSL_BUFFER_INDEX  : CCT.T_UNSIGNED_16;
      TIME_SINCE_READOUT: CCT.T_UNSIGNED_16;
      EVENTS_TRANSFERRED: CCT.T_UNSIGNED_16;
      AUT_RECV_INFO     : CCT.T_UNSIGNED_16;
      AMOUNT_OF_EVENTS  : CCT.T_UNSIGNED_16;
      DPE_EVENTS        : CIT.T_U32_BITS;
      MEMORY_PATCHES    : INTEGER;
      PARAMETERS_OBT    : CIT.T_OBT;
   end record;

   -- diagnostic HK of the package COMMON.
   subtype T_DIAGNOSTIC_HK is T_TASK_ALIVE_CNT;

   -- header of telecommand TC(6,1)-Load memory
   type T_TC_LOAD_MEM_HDR is record
      MID   : CCT.T_UNSIGNED_16;        -- =255 for DFEE's memory
      ADDR2 : CCT.T_UNSIGNED_16;        -- not used part of 32-bits address
      ADDR  : CCT.T_UNSIGNED_16;        -- address of DFEE memory
   end record;

   -- header of telecommands TC(6,2) and TC(6,3)
   type T_TC_MEM_HDR is record
      MID   : CCT.T_UNSIGNED_16;        -- =255 for DFEE's memory
      ADDR2 : CCT.T_UNSIGNED_16;        -- not used part of 32-bits address
      ADDR  : CCT.T_UNSIGNED_16;        -- address of DFEE memory
      LEN   : INTEGER;     -- this field doesn't exists in TC(6,1)-Load
   end record;

   type T_MRTU_INPUT is (
      DET_PRESSURE_1,
      DET_PRESSURE_2,
      DET_TEMP_1,
      DET_TEMP_2,
      VOLTAGE_P5V_DIG,
      CURRENT_P5V_DIG,
      VOLTAGE_P5V_ANA,
      VOLTAGE_M5V_ANA,
      VOLTAGE_P12V,
      CURRENT_P12V,
      VOLTAGE_M12V,
      CURRENT_M12V,
      TEMP_1,
      TEMP_2,
      TEMP_3,
      TEMP_4,
      TEMP_5,
      TEMP_6,
      FIFO_FLAG,
      HV_PS_TEMP_1,
      HV_PS_TEMP_2,
      SPARE_ANALOG_CHANNEL
   );

   type T_MRTU_HK is array(T_MRTU_INPUT) of CCT.T_UNSIGNED_16;

   -- T_EVENT - the array which can contain any event type.  Only 5, 36,
   -- or 42 first words are valid, depending on current data mode.

   subtype T_EVENT_SIZE is CCT.T_UNSIGNED_16;
   type T_EVENT_WRAPPER is array(T_EVENT_SIZE range <>) of CCT.T_UNSIGNED_16;
   subtype T_EVENT is T_EVENT_WRAPPER (0 .. 42);
   -- T_EVENT covers 42+1 words.  +1 is for event marker F000, which
   -- is also stored, but only in case of test data mode.

   -- Types of dummy events:

   type T_DUMMY_EVENT is (
      UNKNOWN_DUMMY_EVENT,
      NORMAL_EVENT,
      -- dummy events generated in DFEE
      END_OF_DATA_MARKER,
      GREY_FILTER_REPORT,
      CALIBRATION_AMPLITUDE_REPORT,
      -- dummy event generated in TASK_HSL
      CHANGE_FORMAT_REQUEST
   );


   --------------------------------------------------------------------
   -- Local types and data, made global because of possible patching --
   --------------------------------------------------------------------

   -- Counters of activity of tasks.
   TASK_ALIVE_CNT : T_TASK_ALIVE_CNT;


   -----------------------
   -- Global procedures --
   -----------------------

   --The function performs P-operation on the binary semaphore.
   --For V-operation use simply assignement to zero.
   --SEM should point to a 16-bit word of which only bit#0 is used.
   function CAN_ENTER_CRITSEC(SEM : in SYSTEM.ADDRESS) return BOOLEAN;
   pragma INTERFACE (Assembly, CAN_ENTER_CRITSEC, "OP_CAN_ENTER_CRITSEC");

   procedure START_TASK( TASK_ID : in CIT.T_IASW_TASK_ID );
   -- this version of START_TASK can be used without checking if
   -- the task is in the state when starting is legal.

   procedure I_AM_ALIVE( TASK_NO : in T_TASK_NO );
   -- this is called by all tasks.  it increments diagnostic counters.

   -- Z := X - Y - (C); C := borrow
   procedure SUBB(
      X,Y : in     CCT.T_UNSIGNED_16;
      Z   :    out CCT.T_UNSIGNED_16;
      C   : in out BOOLEAN);

   -- Z := X + Y + (C); C := carry;
   procedure ADDC(
      X,Y : in     CCT.T_UNSIGNED_16;
      Z   :    out CCT.T_UNSIGNED_16;
      C   : in out BOOLEAN);

   -- increments X, but if it is = FFFF, then sets to X_AFTER_FFFF.
   procedure INC (
      X            : in out CCT.T_UNSIGNED_16;
      INCREMENT    : in CCT.T_UNSIGNED_16 := 1;
      X_AFTER_FFFF : in CCT.T_UNSIGNED_16 := 16#FFFF#);

   procedure INC32( X : in out CIT.T_U32_BITS );

   function INC_MODULO_FFFF( X : in CCT.T_UNSIGNED_16 ) return CCT.T_UNSIGNED_16;

   function INT2LONG( X : in INTEGER ) return LONG_INTEGER;

   procedure GET_HK_DIAG(
      FLAGS : out IASW_COMMON.T_16_BITS;
      DIAG  : out T_DIAGNOSTIC_HK);

   function GET_DUMMY_EVENT_TYPE (EVENT : in T_EVENT) return T_DUMMY_EVENT;

   procedure CONSTRUCTOR;

   -- Conversions for general use

   function CVT is new UNCHECKED_CONVERSION(
      SOURCE => INTEGER, TARGET => CCT.T_UNSIGNED_16);

   function CVT is new UNCHECKED_CONVERSION(
      SOURCE => CCT.T_UNSIGNED_16, TARGET => INTEGER);

   function CVT is new UNCHECKED_CONVERSION(
      SOURCE => SYSTEM.UNSIGNED, TARGET => CCT.T_UNSIGNED_16);

   function CVT is new UNCHECKED_CONVERSION(
      SOURCE => CCT.T_UNSIGNED_16, TARGET => SYSTEM.UNSIGNED);

   function CVT is new UNCHECKED_CONVERSION(
      SOURCE => INTEGER, TARGET => SYSTEM.UNSIGNED);

   function CVT is new UNCHECKED_CONVERSION(
      SOURCE => SYSTEM.UNSIGNED, TARGET => INTEGER);

end IASW_COMMON;
