----------------------------------------------------------------------------
-- Module Name:        IASW_DFEE - package body
-- Project:            INTEGRAL, JEM-X, DPE
-- Author:             Grzegorz Juchnikowski
-- Creation Date:      3.III.1998
-- Change History:
--    (Version/Date/Author/Description)
--    see Release Notes - the separate document
-- Function:
--    (Say what the component does)
--    Provides the service of dialog DPE-DFEE.
-- Interfaces:
--    (Define the inputs and outputs)
-- Dependencies:
--    (Describe the preconditions for using this component)
--    Before using any procedure from this package the package should
--    be initialised using procedure IASW_DFEE.CONSTRUCTOR;
--    Afterthat any task that will use communication with DFEE should
--    call IASW_DFEE.CONNECT_TO_LSL.
-- Processing:
--    (Summarise the processing using pseudo-code or PDL)
--    This package presents to the caller a set of procedures for
--    commanding DFEE.  They are formulated on the functional level, eg.
--    SET_STATE, GET_HK, LOAD_MEMORY, and so on.
--    The front end procedures call DPE_DFEE_DIALOG() giving it
--    the command to be sent and the buffer and expected type and length
--    for the answer.  Buffers, both output and input are described
--    in terms of memory address (see comment below).
--    The procedure DPE_DFEE_DIALOG() calls LSL_TX_RX(), which in turn
--    calls DPE_CSSW.TRANSMIT_AND_RECEIVE_LSL_MESSAGES().
--    Job for LSL_TX_RX() is to add header and trailer words to the
--    output packet, and selection of buffers for output and input.
--    Job for DPE_DFEE_DIALOG(): calculations of CRC-s, both output and input,
--    checking if the answer corresponds to the command, and repeatition
--    of transmission if necessary.
--    In case of permanent malfunction, or negative answer from DFEE the
--    exception is generated.  OEMs are produced together with exceptions
--    caused by the problems on Low Speed Line.
--
-- Comment on use of the ADA construction 'use at':
--    Commands to DFEE and responses from DFEE have differente formats
--    and differente lengths for differente functions.
--    They are defined as ADA records.
--    The procedure for data taransmission there and back over the
--    low speed line is off course common for all entries on the
--    functional level.  This implies that communication buffers for
--    output and input must be defined in one universal way.  There is
--    a problem with this, because ADA doesn't give any simple and safe
--    mechanizm for redefining of type of structures of differente length
--    (anything equivalent to 'union' in C, or 'cased record' in Pascal).
--    Solution found makes use of the ADA construction :
--       'for <variable> use at <address>'
--    which makes possible to see a variable of a given type in the 
--    memory location occupied by the variable defined another way.  
--    It is used in the two situations:
--    - When transmitting command to DFEE, the command block, which is
--      defined as an ADA record is "seen" as the communication output
--      buffer,
--    - After reception of response from DFEE the communication input
--      buffer is "seen" as the answer block - another ADA record.
--    This construction is however intrinsicly unsafe, because the compiler
--    has no chance to check that one tries to write to random locations
--    in memory.  Therefore special care must be undertaken, when changing
--    anything related to variables pointed in 'for ... use at ...'.
--
----------------------------------------------------------------------------

with UNCHECKED_CONVERSION;
with DPE_CSSW;
with CRC;
with IASW_TM;
with IASW_PATCHES;
with IASW_REQ_OUT;
--the two next necessary only when CSSW v1.2 modified by GJ is used.
--with SW_DRV_GLOBAL;
--with LOW_SPEED_LINE;

package body IASW_DFEE is

   -- local procedures

   procedure ENTER_LSL_CRITSEC is
   begin
      loop
         exit when IASW_COMMON.CAN_ENTER_CRITSEC(LSL_SEMAPHORE'address);
         DPE_CSSW.DELAY_RELATIVE(1);
      end loop;
   end ENTER_LSL_CRITSEC;

   procedure LEAVE_LSL_CRITSEC is
   begin
      LSL_SEMAPHORE := 0;
   end LEAVE_LSL_CRITSEC;


   function GET_CRC_IN_MY_AS(   -- local
   -- this function does the same as CRC.GET_CRC, but accepts addresses
   -- both from ROOT and AS1
      ADDRESS : in CIT.T_LOGICAL_ADDRESS;
      LENGTH  : in INTEGER
   ) return CCT.T_UNSIGNED_16 is
      USERS_BUF : IASW_COMMON.T_ARRAY_OF_WORDS(1..LSL_MAX_LONG_TRANSFER);
         for USERS_BUF use at ADDRESS;
      -- Comment on 'use at': caller of GET_CRC_IN_MY_AS should ensure
      -- that the buffer pointed by ADDRESS has at least LENGTH words.
      BUF_ON_STACK : IASW_COMMON.T_ARRAY_OF_WORDS(1..LSL_MAX_LONG_TRANSFER);
   begin
      if LENGTH not in 1..LSL_MAX_LONG_TRANSFER then
         return 0;
      end if;
      for I in 1..LENGTH loop
         BUF_ON_STACK(I) := USERS_BUF(I);
      end loop;
      return CRC.GET_CRC( BUF_ON_STACK'address, LENGTH );
   end GET_CRC_IN_MY_AS;


   -- the two following procedures are an interface to procedures
   -- IASW_PATCHES.ADD_PATCH and IASW_PATCHES.GET_PATCH, which
   -- work in another AS, and therefore can not raise exceptions.
   -- these two raise exceptions for them.

   procedure ADD_PATCH(
      HDR  : in T_MEM_BLOCK_SPEC;
      DATA : in T_MEM_BLOCK
   ) is
      MEM_BLOCK_HDR : T_MEM_BLOCK_SPEC;
      MEM_BLOCK     : T_MEM_BLOCK(1..K_MAX_LOAD_MEM_BLK_LEN);
      RETCODE       : IASW_PATCHES.T_PATCH_RETCODE;
   begin
      MEM_BLOCK_HDR := HDR;
      for I in 1..MEM_BLOCK_HDR.LENGTH loop
         MEM_BLOCK(I) := DATA(I);
      end loop;

      IASW_PATCHES.ADD_PATCH(MEM_BLOCK_HDR, MEM_BLOCK, RETCODE);

      -- simulate as if IASW_PATCHES.ADD_PATCH() generated an exception
      -- IASW_PATCHES.ADD_PATCH() itself can't do it because of problems with
      -- exceptions generated in differente AS.
      case RETCODE is
   when IASW_PATCHES.PATCHES_TABLE_FULL =>
         raise EX_PATCHES_TABLE_FULL;
   when IASW_PATCHES.PATCHES_LENGTH_ERROR =>
         raise EX_PATCHES_LENGTH_ERROR;
   when others =>
         -- just for case, no other exceptions are expected
         null;
      end case;
   end ADD_PATCH;

   procedure GET_PATCH(
      NUM  : in INTEGER;
      HDR  : out T_MEM_BLOCK_SPEC;
      DATA : out T_MEM_BLOCK
   ) is
      MEM_BLOCK_HDR : T_MEM_BLOCK_SPEC;
      MEM_BLOCK     : T_MEM_BLOCK(1..K_MAX_LOAD_MEM_BLK_LEN);
      RETCODE       : IASW_PATCHES.T_PATCH_RETCODE;
   begin
      IASW_PATCHES.GET_PATCH(NUM, MEM_BLOCK_HDR, MEM_BLOCK, RETCODE);

      -- simulate as if IASW_PATCHES.GET_PATCH() generated an exception
      -- IASW_PATCHES.GET_PATCH() itself can't do it because of problems with
      -- exceptions generated in differente AS.
      case RETCODE is
   when IASW_PATCHES.PATCHES_INDEX_ERROR =>
         raise EX_PATCHES_INDEX_ERROR;
   when IASW_PATCHES.PATCHES_LENGTH_ERROR =>
         raise EX_PATCHES_LENGTH_ERROR;
   when others =>
         -- just for case, no other exceptions are expected
         null;
      end case;

      HDR := MEM_BLOCK_HDR;
      for I in 1..MEM_BLOCK_HDR.LENGTH loop
         DATA(I) := MEM_BLOCK(I);
      end loop;
   end GET_PATCH;


   procedure SET_LSL_PADDING1(
      BUF : out IASW_COMMON.T_ARRAY_OF_WORDS;
      J   : in out INTEGER
   ) is
   begin
      -- K_P1_LEN words should be inserted
      -- procedure must be changed if K_P1_LEN differente than 0 or 2
pragma If(K_P1_LEN=0);
      null;
pragma Else;
      J := J + 1;
      BUF(J) := 16#DDDD#;
      J := J + 1;
      BUF(J) := 16#DDDD#;
pragma End if;
   end SET_LSL_PADDING1;


   procedure SET_LSL_PADDING2(
      BUF : out IASW_COMMON.T_ARRAY_OF_WORDS;
      J   : in out INTEGER
   ) is
   begin
      -- K_P2_LEN words should be inserted
      -- procedure must be changed if K_P2_LEN differente than 0 or 2
pragma If(K_P2_LEN=0);
      null;
pragma Else;
      J := J + 1;
      BUF(J) := 16#3333#;
      J := J + 1;
      BUF(J) := 16#6666#;
pragma End if;
   end SET_LSL_PADDING2;


   procedure LSL_TX_RX (
   --
   -- this procedure performs single LSL dialog (TX+RX).  sets RESULT
   -- according to the result of the communication.  it also does:
   -- - adds padding words,
   -- - sets global variable LSL_DIAG_CODE,
   -- - increments propper diagnostic counters,
   -- - sends OEMs if necessary,
   -- - blocks any exceptions.
   --
      TX_BUF : in CIT.T_LOGICAL_ADDRESS;
      TX_LEN : in INTEGER;
      RX_BUF : in CIT.T_LOGICAL_ADDRESS;
      RX_LEN : in INTEGER;
      TIMEOUT : in CIT.T_RTC_TICKS;
      REAL_RX_LEN : out INTEGER;
      RESULT : out T_LSL_RESULT
   ) is
      TXB : IASW_COMMON.T_ARRAY_OF_WORDS(1..LSL_MAX_LONG_TRANSFER);
         for TXB use at TX_BUF;
      RXB : IASW_COMMON.T_ARRAY_OF_WORDS(1..LSL_MAX_LONG_TRANSFER);
         for RXB use at RX_BUF;
      -- See general comment on use of 'use at' in the file header.
      EV_LSL : CIT.T_EVENT_LIST := (others => FALSE);
      EV : CIT.T_EVENT_LIST;
      type T_RX_BUFFER_TYPE is ( RX_SHORT, RX_LONG );
      RX_BUFFER_TYPE : T_RX_BUFFER_TYPE;
      J : INTEGER;
      ANSWER_READY : BOOLEAN;

      -- communication TX- and RX-buffers on stack
      -- program decides which of them will be used for TX and which for RX.
      -- this complication is to save stack by not allocating two long buffers.
      SHORT_BUF : IASW_COMMON.T_ARRAY_OF_WORDS(1..LSL_MAX_SHORT_TRANSFER+K_PADD_LEN);
      LONG_BUF  : IASW_COMMON.T_ARRAY_OF_WORDS(1..LSL_MAX_LONG_TRANSFER+K_PADD_LEN);
      COMMID, ANSID : INTEGER;

      function CVT is new UNCHECKED_CONVERSION(
         SOURCE => CIT.T_EVENT_LIST, TARGET => CCT.T_UNSIGNED_16);
   begin
      DPE_CSSW.DELAY_RELATIVE(1);   -- slow down the LSL dialog
      --preliminary
      RESULT := LSL_RESULT_ERROR;
      REAL_RX_LEN := RX_LEN;

      -- clear BUSY flag in LSL status
      SET_LSL_MODE;

      EV_LSL(IASW_COMMON.LSL_COMPLETED_EVENT)     := TRUE;
      EV_LSL(IASW_COMMON.LSL_EOT_EVENT)           := TRUE;
      EV_LSL(IASW_COMMON.LSL_PARITY_ERROR_EVENT)  := TRUE;
      EV_LSL(IASW_COMMON.LSL_OVERRUN_ERROR_EVENT) := TRUE;
      EV_LSL(IASW_COMMON.LSL_FRAMING_ERROR_EVENT) := TRUE;

      DPE_CSSW.CLEAR_EVENTS( EVENTS => EV_LSL );

      IASW_COMMON.INC( DFEE_DIAGNOSTIC.TX_TOTAL_NO );
      COMMID := IASW_COMMON.CVT(TXB(1));

      if TX_LEN <= LSL_MAX_SHORT_TRANSFER then

         RX_BUFFER_TYPE := RX_LONG;
         J := 0;
         SET_LSL_PADDING1(SHORT_BUF, J);
         for I in 1..TX_LEN loop
            J := J + 1;
            SHORT_BUF(J) := TXB(I);
         end loop;
         SET_LSL_PADDING2(SHORT_BUF, J);
         -- clear input buffer
         LONG_BUF(1) := 0;
         LONG_BUF(3) := 0;
         --
         DPE_CSSW.TRANSMIT_AND_RECEIVE_LSL_MESSAGES(
            TX_ADDRESS      => SHORT_BUF'address,
            TX_WORD_LENGTH  => IASW_COMMON.CVT(J),
            RX_ADDRESS      => LONG_BUF'address,
            RX_WORD_LENGTH  => IASW_COMMON.CVT(RX_LEN)
         );

      elsif RX_LEN <= LSL_MAX_SHORT_TRANSFER then

         RX_BUFFER_TYPE := RX_SHORT;
         J := 0;
         SET_LSL_PADDING1(LONG_BUF, J);
         for I in 1..TX_LEN loop
            J := J + 1;
            LONG_BUF(J) := TXB(I);
         end loop;
         SET_LSL_PADDING2(LONG_BUF, J);
         -- clear input buffer
         SHORT_BUF(1) := 0;
         SHORT_BUF(3) := 0;
         --
         DPE_CSSW.TRANSMIT_AND_RECEIVE_LSL_MESSAGES(
            TX_ADDRESS      => LONG_BUF'address,
            TX_WORD_LENGTH  => IASW_COMMON.CVT(J),
            RX_ADDRESS      => SHORT_BUF'address,
            RX_WORD_LENGTH  => IASW_COMMON.CVT(RX_LEN)
         );

      else
         -- no buffers provided for such transmission
         raise EX_DFEE_CALLER_ERROR;
      end if;

      LSL_DIAG_CODE := 0;
      ANSWER_READY := FALSE;

      loop   --until event /= EOT
         DPE_CSSW.WAIT_AND_CLEAR_EVENTS(
            EVENTS   => EV_LSL,
            TIMEOUT  => TIMEOUT,
            OCCURRED => EV );

         LSL_DIAG_CODE := LSL_DIAG_CODE + CVT(EV);
         -- there should rather be binary OR instead of +

         case RX_BUFFER_TYPE is
      when RX_SHORT => ANSID := IASW_COMMON.CVT(SHORT_BUF(1));
      when RX_LONG  => ANSID := IASW_COMMON.CVT(LONG_BUF(1));
         end case;

         if EV(IASW_COMMON.LSL_COMPLETED_EVENT) then
            ANSWER_READY := TRUE;

         elsif EV(IASW_COMMON.LSL_PARITY_ERROR_EVENT) then
            IASW_TM.PUT_OEM(IASW_TM.EV_LSL_PARITY, COMMID, ANSID);
            IASW_COMMON.INC(DFEE_DIAGNOSTIC.RX_ERROR_NO);
            exit;

         elsif EV(IASW_COMMON.LSL_FRAMING_ERROR_EVENT) then
            IASW_TM.PUT_OEM(IASW_TM.EV_LSL_FRAMING, COMMID, ANSID);
            IASW_COMMON.INC(DFEE_DIAGNOSTIC.RX_ERROR_NO);
            exit;

         elsif EV(IASW_COMMON.LSL_OVERRUN_ERROR_EVENT) then
            IASW_TM.PUT_OEM(IASW_TM.EV_LSL_OVERRUN, COMMID, ANSID);
            IASW_COMMON.INC(DFEE_DIAGNOSTIC.RX_ERROR_NO);
            exit;

         elsif EV(IASW_COMMON.LSL_EOT_EVENT) then
            null;    --ignore EOT, keep looping

         else  -- timeout
            -- maybe we receive Rq.Ack, instead of a longer block
            case RX_BUFFER_TYPE is
         when RX_SHORT => ANSWER_READY := (SHORT_BUF(1) = ANS_CODE_RQ_ACK);
         when RX_LONG  => ANSWER_READY := (LONG_BUF(1) = ANS_CODE_RQ_ACK);
            end case;
            if ANSWER_READY then
               -- yes, this is Rq.Ack.
               REAL_RX_LEN := T_LSL_ANS_RQACK'size / 16;
            else
               -- no, this is true timeout
               IASW_TM.PUT_OEM(IASW_TM.EV_LSL_TIMEOUT, COMMID, ANSID);
               IASW_COMMON.INC(DFEE_DIAGNOSTIC.RX_TIMEOUT_NO);
               RESULT := LSL_RESULT_TIMEOUT;
               exit;
            end if;
         end if;

         if ANSWER_READY then
            case RX_BUFFER_TYPE is
         when RX_SHORT =>
               for I in 1..RX_LEN loop
                  RXB(I) := SHORT_BUF(I);
               end loop;
         when RX_LONG =>
               for I in 1..RX_LEN loop
                  RXB(I) := LONG_BUF(I);
               end loop;
            end case;
            RESULT := LSL_RESULT_OK;
            IASW_TM.LSL_ERROR_CONDITION_OFF;
            exit;
         end if;
      end loop;

   exception

      when CIT.EX_LOW_SPEED_LINE_IS_DISABLED =>
         IASW_TM.PUT_OEM(IASW_TM.EV_LSL_DISABLED);
         LSL_DIAG_CODE := 16#FF01#;
         DPE_CSSW.ENABLE_LOW_SPEED_LINE;
      when CIT.EX_LOW_SPEED_LINE_IS_BUSY =>
         IASW_TM.PUT_OEM(IASW_TM.EV_LSL_BUSY);
         LSL_DIAG_CODE := 16#FF02#;
      when others =>
         LSL_DIAG_CODE := 16#FF03#;
   end LSL_TX_RX;


   procedure DPE_DFEE_DIALOG(
   --
   -- this LSL dialog procedure uses LSL_TX_RX and additionaly does:
   -- - repeats transmission if no success,
   -- - calculates CRCs both on TX and RX side,
   -- - increments diagnostic counter dealing CRC errors and LSL failures,
   -- - produces OEM dealing CRC errors, R.A. /= 0 and unexp. answers,
   -- - checks if the answer corresponds to the question,
   --
      COMMAND     : in CIT.T_LOGICAL_ADDRESS;
      COMMAND_LEN : in INTEGER;
      ANSWER      : in CIT.T_LOGICAL_ADDRESS;
      ANSWER_LEN  : in INTEGER := 0;
      ANSWER_ID   : in CCT.T_UNSIGNED_16 := 0;
      OEM_RA6     : in BOOLEAN := TRUE
   ) is
      comm : IASW_COMMON.T_ARRAY_OF_WORDS(1..LSL_MAX_LONG_TRANSFER);
        for comm use at COMMAND;
      ans : IASW_COMMON.T_ARRAY_OF_WORDS(1..LSL_MAX_LONG_TRANSFER);
        for ans use at ANSWER;
      -- See general comment on use of 'use at' in the file header.
      rept_cnt : INTEGER;
      anslen, real_anslen : INTEGER;
      RA_expected : BOOLEAN;
      lsl_result : T_LSL_RESULT;
   begin
      COMM_ID := IASW_COMMON.CVT(comm(1));
      comm(COMMAND_LEN) := GET_CRC_IN_MY_AS(COMMAND, COMMAND_LEN-1);

      if ANSWER_LEN > 0 then
         RA_expected := FALSE;
         anslen := ANSWER_LEN;
      else
         RA_expected := TRUE;
         anslen := 3;
      end if;

      -- set the number of repetitions, it will be back-counted
      if CONTEXT.STATE >= K_STATE_SAFE then
         rept_cnt := LSL_MAX_REPT_NO;
      else
         -- try only once, when just after reset of DFEE
         -- in this case it is the request for HK, and many repetitions are not desired.
         rept_cnt := 1;
      end if;

      loop
         -- this loop can only be terminated by 'exit' when success, or
         -- by 'raise exc.'

         LSL_TX_RX (
            TX_BUF => COMMAND,
            TX_LEN => COMMAND_LEN,
            RX_BUF => ANSWER,
            RX_LEN => anslen,
            TIMEOUT => LSL_TIMEOUT,
            REAL_RX_LEN => real_anslen,
            RESULT => lsl_result );

         ANS_ID := IASW_COMMON.CVT(ans(1));

         if lsl_result = LSL_RESULT_OK then

            if ans(INTEGER(real_anslen)) /= GET_CRC_IN_MY_AS(ANSWER, real_anslen-1) then
               IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_CRC, COMM_ID,ANS_ID);
               IASW_COMMON.INC(DFEE_DIAGNOSTIC.RX_CRC_NO);
            else   --CRC is ok.
               IASW_TM.DFEE_CRC_ERROR_CONDITION_OFF;
               if ans(1) = ANS_CODE_RQ_ACK then
                  case ans(2) is
      when RA_CODE_OK =>
                     if RA_expected then
                        exit;   -- Success
                     else
                        raise EX_DFEE_UNEXP_ANSWER;
                     end if;
      when RA_CODE_CRC_ERROR =>
                     IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_RA_1, COMM_ID,ANS_ID);
                     -- transmission will be repeated
      when RA_CODE_RQ_UNKNOWN =>
                     IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_RA_2, COMM_ID,ANS_ID);
                     raise EX_DFEE_DETECTS_ERROR;
      when RA_CODE_RQ_NOT_VALID =>
                     IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_RA_3, COMM_ID,ANS_ID);
                     raise EX_DFEE_DETECTS_ERROR;
      when RA_CODE_ID_NOT_VALID =>
                     IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_RA_4, COMM_ID,ANS_ID);
                     raise EX_DFEE_DETECTS_ERROR;
      when RA_CODE_PARAM_NOT_VALID =>
                     IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_RA_5, COMM_ID,ANS_ID);
                     raise EX_DFEE_DETECTS_ERROR;
      when RA_CODE_BUSY =>
                     if OEM_RA6 then
                        IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_RA_6, COMM_ID,ANS_ID);
                     end if;
                     raise EX_DFEE_BUSY;
      when others =>
                     IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_RA_X, COMM_ID,IASW_COMMON.CVT(ans(2)));
                     raise EX_DFEE_UNEXP_ANSWER;
                  end case;
               elsif RA_expected then
                  raise EX_DFEE_UNEXP_ANSWER;
               elsif ans(1) = ANSWER_ID then
                  exit;    -- Success
               else
                  raise EX_DFEE_UNEXP_ANSWER;
               end if;
            end if;  -- ends 'CRC ok'
         end if;  -- ends 'lsl_result = OK'

         -- if the program is here, the transmission was not correct
         -- and should be repeated.

         -- count repetitions
         rept_cnt := rept_cnt - 1;
         if rept_cnt <= 0 then
            raise EX_LSL_DEAD;
         end if;

         -- transmission will be repeated, wait a while before that.
         DPE_CSSW.DELAY_RELATIVE (1);

      end loop;

      -- Success
      DFEE_ANSWERED;

   exception

      when EX_DFEE_UNEXP_ANSWER =>
         IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_UNEXP_ANSWER, COMM_ID,ANS_ID);
         IASW_COMMON.INC(DFEE_DIAGNOSTIC.RX_FAILURE_NO);
         raise;
      when EX_LSL_DEAD =>
         DFEE_DOES_NOT_ANSWER;
         IASW_COMMON.INC(DFEE_DIAGNOSTIC.RX_FAILURE_NO);
         raise;
      when others =>
         IASW_COMMON.INC(DFEE_DIAGNOSTIC.RX_FAILURE_NO);
         raise;
   end DPE_DFEE_DIALOG;


   procedure SET_LSL_MODE is
   begin
      DPE_CSSW.SET_LOW_SPEED_LINE_MODE( SYNCHRONOUS => FALSE );
   end SET_LSL_MODE;


   -- public procedures

   procedure SET_LSL_PARAMS (
      CLK_FREQ_ID   : in CCT.T_UNSIGNED_16;
      BDR_FACTOR_ID : in CCT.T_UNSIGNED_16 := 1
   ) is
      subtype T_CLK_FREQ_ID is CCT.T_UNSIGNED_16 range 1..6;
      CLK_FREQ : array(T_CLK_FREQ_ID) of CIT.T_BAUD_RATE := (
         KBPS_02, KBPS_04, KBPS_08, KBPS_16, KBPS_32, KBPS_64
      );
   begin
      if (CLK_FREQ_ID not in T_CLK_FREQ_ID) then
         return;
      end if;

      SET_LSL_MODE;

      DPE_CSSW.CONFIGURE_LOW_SPEED_LINE(
         CHANNEL   => LSL_CHANNEL,
         BAUD_RATE => CLK_FREQ(CLK_FREQ_ID) );

      DPE_CSSW.ENABLE_LOW_SPEED_LINE;
   end SET_LSL_PARAMS;


   procedure CONNECT_TO_LSL is
   -- Connects the calling task to the LSL HW events
   -- Each of the tasks using LSL should call IASW_DFEE.CONNECT_TO_LSL
   -- at the start of the task.
   begin
      DPE_CSSW.CONNECT_TO_HW_EVENT(
         HW_EVENT   => CIT.LSL_END_OF_RECEPTION,
         TASK_EVENT => IASW_COMMON.LSL_COMPLETED_EVENT );
      DPE_CSSW.CONNECT_TO_HW_EVENT(
         HW_EVENT   => CIT.LSL_END_OF_TRANSMISSION,
         TASK_EVENT => IASW_COMMON.LSL_EOT_EVENT );
      DPE_CSSW.CONNECT_TO_HW_EVENT(
         HW_EVENT   => CIT.LSL_PARITY_ERROR,
         TASK_EVENT => IASW_COMMON.LSL_PARITY_ERROR_EVENT );
      DPE_CSSW.CONNECT_TO_HW_EVENT(
         HW_EVENT   => CIT.LSL_OVERRUN_ERROR,
         TASK_EVENT => IASW_COMMON.LSL_OVERRUN_ERROR_EVENT );
      DPE_CSSW.CONNECT_TO_HW_EVENT(
         HW_EVENT   => CIT.LSL_FRAMING_ERROR,
         TASK_EVENT => IASW_COMMON.LSL_FRAMING_ERROR_EVENT );
   end CONNECT_TO_LSL;


   procedure TEST_LSL_DIALOG(
   -- executes as result of TC(5,3),TID=0,FID=255
      PARAM_VAL: in IASW_COMMON.T_TC_PARAM_VAL;
      PARAM_LEN: in IASW_COMMON.T_SW_PARAM_LEN
   ) is
      lsl_result : T_LSL_RESULT;
      TM_DATA : CIT.T_SCIENCE_DATA_FIELD;
      L1,L2,L2B : INTEGER;
      TM_DONE : BOOLEAN;
   begin
      ENTER_LSL_CRITSEC;

      L1 := IASW_COMMON.CVT(PARAM_VAL(1));   -- length of sent block
      L2 := IASW_COMMON.CVT(PARAM_VAL(2));   -- expected length of received block
      if (L1 < 2) or (L1 > 100) or (L2 < 2) or (L2 > 100) or
         (PARAM_LEN < (2+L1)) then
         raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
      end if;
      for i in 1..L1 loop
         GL_MEM_BLOCK(i) := PARAM_VAL(2+i);
      end loop;
      if GL_MEM_BLOCK(L1) = 0 then
         GL_MEM_BLOCK(L1) := GET_CRC_IN_MY_AS(GL_MEM_BLOCK'address, L1-1);
      end if;
      IASW_TM.INIT_SCIENCE_TM(TM_DATA);
      LSL_TX_RX(
         TX_BUF => GL_MEM_BLOCK'address,
         TX_LEN => L1,
         RX_BUF => TM_DATA'address,
         RX_LEN => L2,
         TIMEOUT => LSL_TIMEOUT,
         REAL_RX_LEN => L2B,
         RESULT => lsl_result );
      if lsl_result = LSL_RESULT_OK then
         TM_DATA(L2B+1) := IASW_COMMON.CVT(GET_CRC_IN_MY_AS(TM_DATA'address, L2B-1));
         IASW_TM.PUT_TM(IASW_COMMON.APID_TM_SC_LSL_DUMP, TM_DATA, TM_DONE);
      else
         IASW_REQ_OUT.REPORT_DIAG_CODE( LSL_DIAG_CODE, 16#2222# );
      end if;

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end TEST_LSL_DIALOG;


   procedure GET_HK ( HK : out T_DFEE_HK ) is
      req : T_LSL_QUESTION;
   begin
      ENTER_LSL_CRITSEC;

      req.INDICATOR := COMM_CODE_HK_RQ;
      req.PARAMETER := 0;   -- TBD parameter
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => LSL_ANS_HK_BLOCK'address,
         ANSWER_LEN  => LSL_ANS_HK_BLOCK'size / 16,
         ANSWER_ID   => ANS_CODE_HK
      );
      HK := LSL_ANS_HK_BLOCK.HK_BLOCK;
      DFEE_STATE_IN_HK_VALID := TRUE;

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end GET_HK;


   procedure GET_STATUS ( STATUS : out T_DFEE_STATUS ) is
   -- this procedure is not used - no one was interested in the status
      req : T_LSL_QUESTION;
      ans : T_LSL_ANS_STATUS;
   begin
      ENTER_LSL_CRITSEC;

      req.INDICATOR := COMM_CODE_STATUS_RQ;
      req.PARAMETER := 0;   -- TBD parameter
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => ans'address,
         ANSWER_LEN  => ans'size / 16,
         ANSWER_ID   => ANS_CODE_STATUS
      );
      STATUS := ans.STATUS;

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end GET_STATUS;


   procedure DUMP_MEMORY (
      MEM_BLOCK_HDR  : in  T_MEM_BLOCK_SPEC;
      MEM_BLOCK      : out T_MEM_BLOCK
   ) is
      req : T_LSL_COMM_GET_MEMORY;
   begin
      ENTER_LSL_CRITSEC;

      if (MEM_BLOCK_HDR.LENGTH not in 1..K_MAX_DUMP_MEM_BLK_LEN) or
         (MEM_BLOCK_HDR.LENGTH > MEM_BLOCK'length) or
         (MEM_BLOCK_HDR.ADDR not in T_DFEE_MEM_ADDRESS) or
         ((MEM_BLOCK_HDR.ADDR + IASW_COMMON.CVT(MEM_BLOCK_HDR.LENGTH)-1)
         not in T_DFEE_MEM_ADDRESS)
      then
         raise EX_DFEE_CALLER_ERROR;
      end if;
      req.INDICATOR     := COMM_CODE_MEMORY_DUMP_RQ;
      req.MEM_BLOCK_HDR := MEM_BLOCK_HDR;
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => LSL_ANS_MEMORY'address,
         ANSWER_LEN  => MEM_BLOCK_HDR.LENGTH + K_LSL_ANS_MEMORY_FRAME_LEN,
         ANSWER_ID   => ANS_CODE_MEMORY_DUMP
      );
      if LSL_ANS_MEMORY.MEM_BLOCK_HDR /= MEM_BLOCK_HDR then
         IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_UNEXP_ANSWER, COMM_ID,ANS_ID);
         raise EX_DFEE_UNEXP_ANSWER;
      end if;
      for i in 1..MEM_BLOCK_HDR.LENGTH loop
         MEM_BLOCK(i) := LSL_ANS_MEMORY.MEM_BLOCK(i);
      end loop;

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end DUMP_MEMORY;


   procedure GET_MEMORY_CRC (
      MEM_BLOCK_HDR  : in  T_MEM_BLOCK_SPEC;
      CRC            : out CCT.T_UNSIGNED_16
   ) is
      req : T_LSL_COMM_GET_MEMORY;
      ans : T_LSL_ANS_MEMORY_CRC;
   begin
      ENTER_LSL_CRITSEC;

      if (MEM_BLOCK_HDR.LENGTH < 1) or
         (MEM_BLOCK_HDR.ADDR not in T_DFEE_MEM_ADDRESS) or
         ((MEM_BLOCK_HDR.ADDR + IASW_COMMON.CVT(MEM_BLOCK_HDR.LENGTH)-1)
         not in T_DFEE_MEM_ADDRESS)
      then
         raise EX_DFEE_CALLER_ERROR;
      end if;
      req.INDICATOR     := COMM_CODE_MEMORY_CRC_RQ;
      req.MEM_BLOCK_HDR := MEM_BLOCK_HDR;
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => ans'address,
         ANSWER_LEN  => ans'size / 16,
         ANSWER_ID   => ANS_CODE_MEMORY_CRC
      );
      if ans.MEM_BLOCK_HDR /= MEM_BLOCK_HDR then
         IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_UNEXP_ANSWER, COMM_ID,ANS_ID);
         raise EX_DFEE_UNEXP_ANSWER;
      end if;
      CRC := ans.MEMORY_CRC;

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end GET_MEMORY_CRC;


   procedure UPDATE_IF_OVERLAYS(
   --
   -- this procedure checks if the memory block up-loaded to DFEE
   -- overlays one of the parameters fields (target), which are
   -- also stored in DPE.  the overlaing portion of data is copied
   -- to the target DPE's mirror area.
   --
      MEM_BLOCK_HDR : in T_MEM_BLOCK_SPEC;
      MEM_BLOCK     : in T_MEM_BLOCK;
      TARGET_DFEE_ADDR : in T_DFEE_MEM_ADDRESS;
      TARGET_DPE_ADDR  : in SYSTEM.ADDRESS;
      TARGET_LEN    : in INTEGER
   ) is
      TARGET : array(0..1000) of CCT.T_UNSIGNED_16;
         for TARGET use at TARGET_DPE_ADDR;
      -- Comment on 'use at': the overlayed RAM area is specified
      -- by its address and length, because the buffers for which
      -- the procedure UPDATE_IF_OVERLAYS is used are of various types
      -- and lengths.  Care should be undertaken when specifying TARGET_LEN,
      -- which should not exceed the real length of the given target buffer.
      -- The number 1000 has been chosen so, that it is not less than the
      -- greatest possible TARGET_LEN.
      a1,a2,b1,b2 : CCT.T_UNSIGNED_16;
      i1,i2,j : INTEGER;
   begin
      a1 := MEM_BLOCK_HDR.ADDR;
      a2 := a1 + IASW_COMMON.CVT(MEM_BLOCK_HDR.LENGTH) - 1;
      b1 := TARGET_DFEE_ADDR;
      b2 := b1 + IASW_COMMON.CVT(TARGET_LEN) - 1;
      if (a1 > b2) or (b1 > a2) then
         return;   -- does not overlap
      end if;
      if a1 > b1 then
         i1 := INTEGER(a1 - b1);
         j := 0;
      else
         i1 := 0;
         j := INTEGER(b1 - a1);
      end if;
      if a2 < b2 then
         i2 := TARGET_LEN - 1 - INTEGER(b2 - a2);
      else
         i2 := TARGET_LEN - 1;
      end if;
      j := j + MEM_BLOCK'first;
      for i in i1..i2 loop
         TARGET(i) := MEM_BLOCK(j);
         j := j + 1;
      end loop;
   end UPDATE_IF_OVERLAYS;


   procedure LOAD_MEMORY (
      MEM_BLOCK_HDR : in T_MEM_BLOCK_SPEC;
      MEM_BLOCK     : in T_MEM_BLOCK;
      UPD_CONTEXT   : in BOOLEAN := TRUE
   ) is
      ans : T_LSL_ANS_RQACK;
   begin
      ENTER_LSL_CRITSEC;

      if (MEM_BLOCK_HDR.LENGTH not in 1..K_MAX_LOAD_MEM_BLK_LEN) or
         (MEM_BLOCK_HDR.LENGTH > MEM_BLOCK'length) or
         (MEM_BLOCK_HDR.ADDR not in T_DFEE_MEM_ADDRESS) or
         ((MEM_BLOCK_HDR.ADDR + IASW_COMMON.CVT(MEM_BLOCK_HDR.LENGTH)-1)
         not in T_DFEE_MEM_ADDRESS)
      then
         raise EX_DFEE_CALLER_ERROR;
      end if;
      LSL_RQ_LOAD_MEMORY.INDICATOR     := COMM_CODE_MEMORY_LOAD;
      LSL_RQ_LOAD_MEMORY.MEM_BLOCK_HDR := MEM_BLOCK_HDR;
      for i in 1..MEM_BLOCK_HDR.LENGTH loop
        LSL_RQ_LOAD_MEMORY.MEM_BLOCK(i) := MEM_BLOCK(i);
      end loop;
      DPE_DFEE_DIALOG(
         COMMAND     => LSL_RQ_LOAD_MEMORY'address,
         COMMAND_LEN => MEM_BLOCK_HDR.LENGTH + K_LSL_COMM_MEMORY_FRAME_LEN,
         ANSWER      => ans'address,
         ANSWER_LEN  => 0,
         ANSWER_ID   => 0
      );
      if UPD_CONTEXT then
         -- Check if int. or float params tables should be updated
         UPDATE_IF_OVERLAYS(
            MEM_BLOCK_HDR, MEM_BLOCK,
            SW_PARAMS_INT_ADDR, CONTEXT.SWPI'address,
            SW_PARAMS_INT_NO);
         UPDATE_IF_OVERLAYS(
            MEM_BLOCK_HDR, MEM_BLOCK,
            SW_PARAMS_FLOAT_ADDR, CONTEXT.SWPF'address,
            SW_PARAMS_FLOAT_NO * 2);
         UPDATE_IF_OVERLAYS(
            MEM_BLOCK_HDR, MEM_BLOCK,
            ENERGY_TABLE_ADDR, CONTEXT.ENTAB'address,
            CONTEXT.ENTAB'length);
         -- Add patch to the list
         ADD_PATCH(MEM_BLOCK_HDR, MEM_BLOCK);
      end if;

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end LOAD_MEMORY;


   procedure GET_SW_PARAM_INT (
   -- This procedure will probably not be used
      PARAM_ID    : in CCT.T_UNSIGNED_16;
      PARAM_VALUE : out T_SW_PARAM_INT
   ) is
      req : T_LSL_QUESTION;
      ans : T_LSL_ANS_SW_PARAM_INT;
   begin
      ENTER_LSL_CRITSEC;

      req.INDICATOR := COMM_CODE_PARAM_INT_RQ;
      req.PARAMETER := PARAM_ID;
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => ans'address,
         ANSWER_LEN  => ans'size / 16,
         ANSWER_ID   => ANS_CODE_PARAM_INT
      );
      PARAM_VALUE := ans.PARAM_VAL;

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end GET_SW_PARAM_INT;


   procedure GET_SW_PARAM_FLOAT (
   -- This procedure will probably not be used
      PARAM_ID    : in CCT.T_UNSIGNED_16;
      PARAM_VALUE : out T_SW_PARAM_FLOAT
   ) is
      req : T_LSL_QUESTION;
      ans : T_LSL_ANS_SW_PARAM_FLOAT;
   begin
      ENTER_LSL_CRITSEC;

      req.INDICATOR := COMM_CODE_PARAM_FLOAT_RQ;
      req.PARAMETER := PARAM_ID;
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => ans'address,
         ANSWER_LEN  => ans'size / 16,
         ANSWER_ID   => ANS_CODE_PARAM_FLOAT
      );
      PARAM_VALUE := ans.PARAM_VAL;

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end GET_SW_PARAM_FLOAT;


   procedure GET_SW_INT_PARAMS(
      SWPI    : out T_SW_PARAMS_INT_FIELD;
      USE_SEM : in BOOLEAN := TRUE
   ) is
      req : T_LSL_QUESTION;
   begin
      if USE_SEM then
         ENTER_LSL_CRITSEC;
      end if;

      req.INDICATOR := COMM_CODE_PAR_FIELD_RQ;
      req.PARAMETER := 1;
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => LSL_SW_PARAMS_INT_FIELD'address,
         ANSWER_LEN  => LSL_SW_PARAMS_INT_FIELD'size / 16,
         ANSWER_ID   => ANS_CODE_PAR_INT_FIELD
      );
      SWPI := LSL_SW_PARAMS_INT_FIELD.PAR_FIELD;

      if USE_SEM then
         LEAVE_LSL_CRITSEC;
      end if;
   exception
      when others =>
         if USE_SEM then
            LEAVE_LSL_CRITSEC;
         end if;
         raise;
   end GET_SW_INT_PARAMS;


   procedure GET_SW_FLOAT_PARAMS(
      SWPF    : out T_SW_PARAMS_FLOAT_FIELD;
      USE_SEM : in BOOLEAN := TRUE
   ) is
      req : T_LSL_QUESTION;
   begin
      if USE_SEM then
         ENTER_LSL_CRITSEC;
      end if;

      req.INDICATOR := COMM_CODE_PAR_FIELD_RQ;
      req.PARAMETER := 2;
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => LSL_SW_PARAMS_FLOAT_FIELD'address,
         ANSWER_LEN  => LSL_SW_PARAMS_FLOAT_FIELD'size / 16,
         ANSWER_ID   => ANS_CODE_PAR_FLOAT_FIELD
      );
      SWPF := LSL_SW_PARAMS_FLOAT_FIELD.PAR_FIELD;

      if USE_SEM then
         LEAVE_LSL_CRITSEC;
      end if;
   exception
      when others =>
         if USE_SEM then
            LEAVE_LSL_CRITSEC;
         end if;
         raise;
   end GET_SW_FLOAT_PARAMS;


   procedure CHG_SW_PARAM_INT (
      PARAM_ID    : in CCT.T_UNSIGNED_16;
      PARAM_VALUE : in T_SW_PARAM_INT
   ) is
      req : T_LSL_COMM_SW_PARAM_INT;
      ans : T_LSL_ANS_RQACK;
   begin
      ENTER_LSL_CRITSEC;

      if (IASW_COMMON.CVT(PARAM_ID) not in T_SW_PARAM_INT_ID) then
         raise EX_DFEE_CALLER_ERROR;
      end if;
      req.INDICATOR := COMM_CODE_PARAM_INT_CHG;
      req.PARAM_ID  := PARAM_ID;
      req.PARAM_VAL := PARAM_VALUE;
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => ans'address,
         ANSWER_LEN  => 0,
         ANSWER_ID   => 0
      );
      -- Update context
      CONTEXT.SWPI(IASW_COMMON.CVT(PARAM_ID)) := PARAM_VALUE;

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end CHG_SW_PARAM_INT;


   procedure CHG_SW_PARAM_FLOAT (
      PARAM_ID    : in CCT.T_UNSIGNED_16;
      PARAM_VALUE : in T_SW_PARAM_FLOAT
   ) is
      req : T_LSL_COMM_SW_PARAM_FLOAT;
      ans : T_LSL_ANS_RQACK;
   begin
      ENTER_LSL_CRITSEC;

      if (IASW_COMMON.CVT(PARAM_ID) not in T_SW_PARAM_FLOAT_ID) then
         raise EX_DFEE_CALLER_ERROR;
      end if;
      req.INDICATOR := COMM_CODE_PARAM_FLOAT_CHG;
      req.PARAM_ID  := PARAM_ID;
      req.PARAM_VAL := PARAM_VALUE;
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => ans'address,
         ANSWER_LEN  => 0,
         ANSWER_ID   => 0
      );
      -- Update context
      CONTEXT.SWPF(IASW_COMMON.CVT(PARAM_ID)) := PARAM_VALUE;

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end CHG_SW_PARAM_FLOAT;


   procedure CHG_CPU_SETTING (
      VALUE   : in CCT.T_UNSIGNED_16;
      OEM_RA6 : in BOOLEAN := TRUE
   ) is
      req : T_LSL_COMMAND1;
      ans : T_LSL_ANS_RQACK;
   begin
      ENTER_LSL_CRITSEC;

      req.INDICATOR := COMM_CODE_CPU_COMMAND;
      req.COMMAND   := VALUE;
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => ans'address,
         ANSWER_LEN  => 0,
         ANSWER_ID   => 0,
         OEM_RA6     => OEM_RA6
      );
      -- Update context
      CONTEXT.CPU := VALUE;

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end CHG_CPU_SETTING;


   procedure CHG_HW_SETTING (
      COMMAND : in CCT.T_UNSIGNED_16;
      PARAM   : in CCT.T_UNSIGNED_16;
      OEM_RA6 : in BOOLEAN := TRUE
   ) is
      req : T_LSL_COMMAND2;
      ans : T_LSL_ANS_RQACK;
   begin
      ENTER_LSL_CRITSEC;

      req.INDICATOR := COMM_CODE_HW_COMMAND;
      req.COMMAND   := COMMAND;
      req.PARAMETER := PARAM;
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => ans'address,
         ANSWER_LEN  => 0,
         ANSWER_ID   => 0,
         OEM_RA6     => OEM_RA6
      );
      -- Update context
      case COMMAND is
   when 0 =>
         case PARAM is
      when 5 =>  -- HV on
            CONTEXT.HV_IS_ON := TRUE;
            CONTEXT.HWST.HV_ACC_SET   := MIN_ACC_HV;
            CONTEXT.HWST.HV_DRIFT_SET := MIN_DRIFT_HV;
      when 6 =>  -- HV off
            CONTEXT.HV_IS_ON := FALSE;
      when others =>
            null;
         end case;
   when 1 =>   CONTEXT.HWST.ANODE_SETTINGS    := PARAM;
   when 2 =>   CONTEXT.HWST.DISCRIM_LOW_LEVEL := PARAM;
   when 3 =>   CONTEXT.HWST.HV_ACC_SET        := PARAM;
   when 4 =>   CONTEXT.HWST.HV_DRIFT_SET      := PARAM;
   when others =>   null;
      end case;

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end CHG_HW_SETTING;


   procedure HV_READY( OEM_RA6 : in BOOLEAN := TRUE ) is
   begin
      CHG_HW_SETTING(0, 4, OEM_RA6);
   end HV_READY;

   procedure HV_ON( OEM_RA6 : in BOOLEAN := TRUE ) is
   begin
      CHG_HW_SETTING(0, 5, OEM_RA6);
   end HV_ON;

   procedure HV_OFF( OEM_RA6 : in BOOLEAN := TRUE ) is
   begin
      CHG_HW_SETTING(0, 6, OEM_RA6);
   end HV_OFF;

   procedure SET_ANODE(
     PARAM   : in CCT.T_UNSIGNED_16;
     OEM_RA6 : in BOOLEAN := TRUE ) is
   begin
      CHG_HW_SETTING(1, PARAM, OEM_RA6);
   end SET_ANODE;

   procedure SET_DISCRIM_LOW_LEVEL(
      PARAM   : in CCT.T_UNSIGNED_16;
      OEM_RA6 : in BOOLEAN := TRUE ) is
   begin
      CHG_HW_SETTING(2, PARAM, OEM_RA6);
   end SET_DISCRIM_LOW_LEVEL;

   procedure SET_HV_ACC(
      PARAM   : in CCT.T_UNSIGNED_16;
      OEM_RA6 : in BOOLEAN := TRUE ) is
   begin
      CHG_HW_SETTING(3, PARAM, OEM_RA6);
   end SET_HV_ACC;

   procedure SET_HV_DRIFT(
      PARAM   : in CCT.T_UNSIGNED_16;
      OEM_RA6 : in BOOLEAN := TRUE ) is
   begin
      CHG_HW_SETTING(4, PARAM, OEM_RA6);
   end SET_HV_DRIFT;


   function REDUCED_HV(
   -- returns HV-DELTA, but not less than HVMIN.
      HV    : in CCT.T_UNSIGNED_16;
      HVMIN : in CCT.T_UNSIGNED_16;
      HVDELTA : in CCT.T_UNSIGNED_16
   ) return CCT.T_UNSIGNED_16 is
   begin
      if HV <= HVDELTA then
         return HVMIN;
      elsif (HV-HVDELTA) <= HVMIN then
         return HVMIN;
      end if;
      return HV-HVDELTA;
   end REDUCED_HV;


   procedure SEND_HV_OFF_RELAY_SIGNAL is
   begin
      for I in 1..2 loop
         DPE_CSSW.SWITCH_RELAY(K_HV_OFF_RELAY_ID);
         while not DPE_CSSW.RELAY_PULSE_IS_IDLE(K_HV_OFF_RELAY_ID) loop
            DPE_CSSW.DELAY_RELATIVE(1);
         end loop;
         DPE_CSSW.DELAY_RELATIVE(4);
      end loop;
      CONTEXT.HV_IS_ON := FALSE;
   exception
      when others =>
         IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_HV_OFF_RELAY);
   end SEND_HV_OFF_RELAY_SIGNAL;


   procedure ACCEPT_STATE( STATE : in T_DFEE_STATE ) is
   -- this is called when DFEE HK reports differente state
   -- then expected, eg. SAFE when HV is sw.off autonomously in DFEE
   -- because of high count rate.
   -- ---
   -- since ver 1.51 it is used only in IASW_HSL, when it receives
   -- dummy events "End of Data" on HSL.
   begin
      ENTER_LSL_CRITSEC;
      DFEE_STATE_IN_HK_VALID := FALSE;
      CONTEXT.STATE := STATE;
      CONTEXT.ST_PAR := 0;
      if STATE < K_STATE_SETUP then
         CONTEXT.HV_IS_ON := FALSE;
      end if;
      LEAVE_LSL_CRITSEC;
   end ACCEPT_STATE;


   procedure SET_STATE (
      STATE   : in T_DFEE_STATE;
      PARAM   : in CCT.T_UNSIGNED_16 := 0;
      OEM_RA6 : in BOOLEAN := TRUE
   ) is
      req : T_LSL_COMMAND2;
      ans : T_LSL_ANS_RQACK;
      do_it_twice : BOOLEAN :=
        (STATE = K_STATE_SAFE) and (CONTEXT.STATE = K_STATE_MEMORY);
   begin
      begin
         ENTER_LSL_CRITSEC;

         DFEE_STATE_IN_HK_VALID := FALSE;

         req.INDICATOR := COMM_CODE_SET_STATE;
         req.COMMAND   := STATE;
         req.PARAMETER := PARAM;
         DPE_DFEE_DIALOG(
            COMMAND     => req'address,
            COMMAND_LEN => req'size / 16,
            ANSWER      => ans'address,
            ANSWER_LEN  => 0,
            ANSWER_ID   => 0,
            OEM_RA6     => OEM_RA6
         );
         -- Update context
         CONTEXT.STATE := STATE;
         CONTEXT.ST_PAR := PARAM;
         if STATE < K_STATE_SETUP then
            CONTEXT.HV_IS_ON := FALSE;
         end if;
         -- repeat "goto Safe" to satisfy propper reset of FPGA.
         if do_it_twice then
            DPE_DFEE_DIALOG(
               COMMAND     => req'address,
               COMMAND_LEN => req'size / 16,
               ANSWER      => ans'address,
               ANSWER_LEN  => 0,
               ANSWER_ID   => 0,
               OEM_RA6     => OEM_RA6
            );
         end if;

         LEAVE_LSL_CRITSEC;
      exception
         when others =>
            LEAVE_LSL_CRITSEC;
            raise;
      end;
      -- load int. and float tables from DFEE if this is the first goto SETUP
      if (STATE = K_STATE_SETUP) and not CONTEXT.SWP_VALID then
         DPE_CSSW.DELAY_RELATIVE(1);
         GET_SW_INT_PARAMS( SWPI => CONTEXT.SWPI );
         DPE_CSSW.DELAY_RELATIVE(1);
         GET_SW_FLOAT_PARAMS( SWPF => CONTEXT.SWPF );
         DPE_CSSW.DELAY_RELATIVE(1);
         SET_ENERGY_TABLE( ENTAB => CONTEXT.ENTAB, UPD_CONTEXT => FALSE );
         DPE_CSSW.DELAY_RELATIVE(1);
         EXEC_ENERGY_TABLE;
         CONTEXT.SWP_VALID := TRUE;
      end if;

   end SET_STATE;


   procedure SET_GREY_FILTER (
      GF_ID         : in CCT.T_UNSIGNED_16
   ) is
      req : T_LSL_COMMAND1;
      ans : T_LSL_ANS_RQACK;
   begin
      ENTER_LSL_CRITSEC;

      req.INDICATOR := COMM_CODE_SET_GREY_FILTER;
      req.COMMAND   := GF_ID;
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => ans'address,
         ANSWER_LEN  => 0,
         ANSWER_ID   => 0
      );

      LEAVE_LSL_CRITSEC;
   exception
      when others =>
         LEAVE_LSL_CRITSEC;
         raise;
   end SET_GREY_FILTER;


   procedure SET_ENERGY_TABLE(
      ENTAB         : in IASW_COMMON.T_COMPRESSED_ENERGY_TABLE;
      UPD_CONTEXT   : in BOOLEAN := TRUE;
      USE_SEM       : in BOOLEAN := TRUE
   ) is
      ans : T_LSL_ANS_RQACK;
   begin
      if UPD_CONTEXT then
         SET_DEFAULT_ENERGY_TABLE(ENTAB);
      end if;

      if USE_SEM then
         ENTER_LSL_CRITSEC;
      end if;

      LSL_RQ_ENTAB.INDICATOR  := COMM_CODE_ENTAB_1;
      for i in 1..128 loop
         LSL_RQ_ENTAB.HALF_ENTAB(i) := ENTAB(i);
      end loop;
      DPE_DFEE_DIALOG(
         COMMAND     => LSL_RQ_ENTAB'address,
         COMMAND_LEN => LSL_RQ_ENTAB'size / 16,
         ANSWER      => ans'address,
         ANSWER_LEN  => 0,
         ANSWER_ID   => 0
      );

      LSL_RQ_ENTAB.INDICATOR  := COMM_CODE_ENTAB_2;
      for i in 1..128 loop
         LSL_RQ_ENTAB.HALF_ENTAB(i) := ENTAB(i+128);
      end loop;
      DPE_DFEE_DIALOG(
         COMMAND     => LSL_RQ_ENTAB'address,
         COMMAND_LEN => LSL_RQ_ENTAB'size / 16,
         ANSWER      => ans'address,
         ANSWER_LEN  => 0,
         ANSWER_ID   => 0
      );

      if USE_SEM then
         LEAVE_LSL_CRITSEC;
      end if;
   exception
      when others =>
         if USE_SEM then
            LEAVE_LSL_CRITSEC;
         end if;
         raise;
   end SET_ENERGY_TABLE;


   procedure EXEC_ENERGY_TABLE(
      USE_SEM       : in BOOLEAN := TRUE
   ) is
      req : T_LSL_COMMAND1;
      ans : T_LSL_ANS_RQACK;
   begin
      if USE_SEM then
         ENTER_LSL_CRITSEC;
      end if;

      req.INDICATOR := COMM_CODE_EXEC_ENTAB;
      req.COMMAND   := 0;
      DPE_DFEE_DIALOG(
         COMMAND     => req'address,
         COMMAND_LEN => req'size / 16,
         ANSWER      => ans'address,
         ANSWER_LEN  => 0,
         ANSWER_ID   => 0
      );

      if USE_SEM then
         LEAVE_LSL_CRITSEC;
      end if;
   exception
      when others =>
         if USE_SEM then
            LEAVE_LSL_CRITSEC;
         end if;
         raise;
   end EXEC_ENERGY_TABLE;


   procedure SET_DEFAULT_ENERGY_TABLE(
      ENTAB : in IASW_COMMON.T_COMPRESSED_ENERGY_TABLE
   ) is
   begin
      CONTEXT.ENTAB := ENTAB;
   end SET_DEFAULT_ENERGY_TABLE;


   procedure READ_CRC(COMPARE,SAVE : in BOOLEAN) is   -- local
      MEM_BLOCK_HDR : T_MEM_BLOCK_SPEC;
      C1,C2,C3,C4 : CCT.T_UNSIGNED_16;
      OEM_F2 : INTEGER := 0;
   begin
      MEM_BLOCK_HDR.ADDR   := SW_PARAMS_INT_ADDR;
      MEM_BLOCK_HDR.LENGTH := SW_PARAMS_INT_NO;
      GET_MEMORY_CRC(MEM_BLOCK_HDR, C1);
      --
      DPE_CSSW.DELAY_RELATIVE(1);
      MEM_BLOCK_HDR.ADDR   := SW_PARAMS_FLOAT_ADDR;
      MEM_BLOCK_HDR.LENGTH := SW_PARAMS_FLOAT_NO * 2;
      GET_MEMORY_CRC(MEM_BLOCK_HDR, C2);
      --
      DPE_CSSW.DELAY_RELATIVE(1);
      MEM_BLOCK_HDR.ADDR   := 16#9000#;
      MEM_BLOCK_HDR.LENGTH := 16#1000#;
      GET_MEMORY_CRC(MEM_BLOCK_HDR, C3);
      --
      DPE_CSSW.DELAY_RELATIVE(1);
      MEM_BLOCK_HDR.ADDR   := 16#B000#;
      MEM_BLOCK_HDR.LENGTH := 16#2000#;
      GET_MEMORY_CRC(MEM_BLOCK_HDR, C4);
      --
      if COMPARE then
         --it is new in ver.1.64 that OEM contains information about
         --which CRC does not agree.
         if C1 /= CONTEXT.SWPI_CRC then
            OEM_F2 := OEM_F2 + 1;
         end if;
         if C2 /= CONTEXT.SWPF_CRC then
            OEM_F2 := OEM_F2 + 2;
         end if;
         if C3 /= CONTEXT.COD1_CRC then
            OEM_F2 := OEM_F2 + 4;
         end if;
         if C4 /= CONTEXT.COD2_CRC then
            OEM_F2 := OEM_F2 + 8;
         end if;
         if OEM_F2 /= 0 then
            IASW_TM.PUT_OEM(IASW_TM.EV_DFEE_CONTEXT,OEM_F2);
            raise EX_DFEE_CRC_HAS_CHANGED;
         end if;
      end if;
      --
      if SAVE then
         CONTEXT.SWPI_CRC := C1;
         CONTEXT.SWPF_CRC := C2;
         CONTEXT.COD1_CRC := C3;
         CONTEXT.COD2_CRC := C4;
      end if;
   end READ_CRC;


   procedure GOTO_STATE(STATE : in T_DFEE_STATE) is
   -- changes to the state required regardless of the current state.
   begin
      loop
         exit when CONTEXT.STATE = STATE;  --the only correct exit
         case CONTEXT.STATE is
      when K_STATE_STANDBY =>
            SET_STATE(K_STATE_SAFE);
      when K_STATE_SAFE =>
            if STATE > K_STATE_SETUP then
               SET_STATE(K_STATE_SETUP);
            else
               SET_STATE(STATE);
            end if;
      when K_STATE_SETUP =>
            if STATE < K_STATE_SETUP then
               SET_STATE(K_STATE_SAFE);
            else
               SET_STATE(STATE);
            end if;
      when K_STATE_MEMORY =>
            SET_STATE(K_STATE_SAFE);
      when others =>
            SET_STATE(K_STATE_SETUP);
         end case;
      end loop;
   end GOTO_STATE;


   procedure SAVE_CONTEXT is
      ST0 : T_DFEE_STATE;
   begin
      ST0 := CONTEXT.STATE;   -- initial state
      GOTO_STATE(K_STATE_MEMORY);

      READ_CRC(COMPARE => FALSE, SAVE => TRUE);

      -- return to the initial state
      GOTO_STATE(ST0);
   end SAVE_CONTEXT;


   procedure LOAD_CONTEXT is
      subtype T_MEM_BLOCK_SWPI is T_MEM_BLOCK(1..SW_PARAMS_INT_NO);
      subtype T_MEM_BLOCK_SWPF is T_MEM_BLOCK(1..SW_PARAMS_FLOAT_NO*2);

      ii : INTEGER;
      NPAT : INTEGER;
      ST0 : T_DFEE_STATE;

      function CVT is new UNCHECKED_CONVERSION(
         SOURCE => T_SW_PARAMS_INT_FIELD, TARGET => T_MEM_BLOCK_SWPI);
      function CVT is new UNCHECKED_CONVERSION(
         SOURCE => T_SW_PARAMS_FLOAT_FIELD, TARGET => T_MEM_BLOCK_SWPF);
   begin
      ST0 := CONTEXT.STATE;   -- initial state

      -- load memory patches
      NPAT := IASW_PATCHES.GET_NUM_PATCHES;
      if NPAT > 0 then
         GOTO_STATE(K_STATE_MEMORY);
         for i in 1..NPAT loop
            ii := i;
            GET_PATCH(ii, GL_MEM_BLOCK_HDR, GL_MEM_BLOCK);
            DPE_CSSW.DELAY_RELATIVE(1);
            LOAD_MEMORY(
               MEM_BLOCK_HDR => GL_MEM_BLOCK_HDR,
               MEM_BLOCK     => GL_MEM_BLOCK,
               UPD_CONTEXT   => FALSE
            );
         end loop;
      end if;

      if CONTEXT.SWP_VALID then
         GOTO_STATE(K_STATE_MEMORY);
         -- load int s/w parameters
         GL_MEM_BLOCK_HDR.ADDR   := SW_PARAMS_INT_ADDR;
         GL_MEM_BLOCK_HDR.LENGTH := SW_PARAMS_INT_NO;
         DPE_CSSW.DELAY_RELATIVE(1);
         LOAD_MEMORY(
            MEM_BLOCK_HDR => GL_MEM_BLOCK_HDR,
            MEM_BLOCK     => CVT(CONTEXT.SWPI), --MEM_BLOCK1,
            UPD_CONTEXT   => FALSE
         );
         -- load float s/w parameters
         GL_MEM_BLOCK_HDR.ADDR   := SW_PARAMS_FLOAT_ADDR;
         GL_MEM_BLOCK_HDR.LENGTH := SW_PARAMS_FLOAT_NO * 2;
         DPE_CSSW.DELAY_RELATIVE(1);
         LOAD_MEMORY(
            MEM_BLOCK_HDR => GL_MEM_BLOCK_HDR,
            MEM_BLOCK     => CVT(CONTEXT.SWPF), --MEM_BLOCK2,
            UPD_CONTEXT   => FALSE
         );
      end if;

      -- load energy linearisation table
      GOTO_STATE(K_STATE_SETUP);
      DPE_CSSW.DELAY_RELATIVE(1);
      SET_ENERGY_TABLE( ENTAB => CONTEXT.ENTAB, UPD_CONTEXT => FALSE );
      DPE_CSSW.DELAY_RELATIVE(1);
      EXEC_ENERGY_TABLE;

      -- verify conformance of DFEE context with the state from before shutdown
      GOTO_STATE(K_STATE_MEMORY);
      READ_CRC( COMPARE => TRUE, SAVE => FALSE );

      -- HW settings
      GOTO_STATE(K_STATE_SETUP);
      DPE_CSSW.DELAY_RELATIVE(1);
      CHG_HW_SETTING( COMMAND => 1, PARAM => CONTEXT.HWST.ANODE_SETTINGS );
      CHG_HW_SETTING( COMMAND => 2, PARAM => CONTEXT.HWST.DISCRIM_LOW_LEVEL );

      -- return to the initial state
      GOTO_STATE(ST0);
   end LOAD_CONTEXT;


   procedure CHECK_STATE(
   -- called in TASK_HK just after reading of HK block.
   -- OEM is generated if the state of DFEE reported in DFEE-HK is not
   -- as expected.  If the state dropped down from Data Taking to Setup
   -- or to Safe, then the data taking is stopped.  If DFEE drops to
   -- Safe, rather than to Setup, it is the signal that HV had to be
   -- switched off due to high events rate - OEM will be issued.
   -- If DPE expects SAFE or SETUP, but DFEE reports one of the data
   -- taking modes, "goto Safe" will be issued.  This strange situation
   -- may be effect of some untipical changes of low voltage.
   -- ---
   -- 27.02.2001 it was decided that no automatic acceptance of SETUP
   -- after data taking mode.
      HK_STATE : in T_DFEE_STATE;
      DATA_TAKING_STOPPED : out BOOLEAN
   ) is
      EXP_STATE : T_DFEE_STATE;
      OK             : BOOLEAN := TRUE;
      ACCEPT_THIS    : BOOLEAN := FALSE;
      FORCE_SAFE     : BOOLEAN := FALSE;
      SETUP_IN_DT    : BOOLEAN := FALSE;
      TEST_POSSIBLE  : BOOLEAN;
   begin
      DATA_TAKING_STOPPED := FALSE;  -- preliminary

      ENTER_LSL_CRITSEC;
      -- crit.section needed to have the two data in sync.
      TEST_POSSIBLE := DFEE_STATE_IN_HK_VALID;
      EXP_STATE     := GET_STATE;
      LEAVE_LSL_CRITSEC;

      if not TEST_POSSIBLE then
         -- it is so when state change command was executed between
         -- GET_HK() and CHECK_STATE().
         -- this will be very rare.
         return;
      end if;

      if (EXP_STATE <= K_STATE_SETUP) and (HK_STATE >= K_STATE_DATA_TAKING) then
         FORCE_SAFE := TRUE;
         OK := FALSE;
      else
         case EXP_STATE is
      when K_STATE_STANDBY =>
            -- there is no STANDBY in DFEE.
            OK := (HK_STATE <= K_STATE_SAFE);
      when K_STATE_SAFE |
           K_STATE_MEMORY =>
            OK := (HK_STATE = EXP_STATE);
      when K_STATE_SETUP =>
            OK := (HK_STATE = K_STATE_SETUP) or
                  (HK_STATE = K_STATE_SETUP_HVR);
      when K_STATE_SETUP_HVR =>
            -- in the current implementation DPE never expects this state
            OK := (HK_STATE = K_STATE_SETUP) or
                  (HK_STATE = K_STATE_SETUP_HVR);
      when K_STATE_DATA_TAKING |
           K_STATE_CALIBRATION |
           K_STATE_DIAGNOSTIC =>
            OK := (HK_STATE = EXP_STATE) or
                  (HK_STATE = K_STATE_SETUP);
            SETUP_IN_DT := (HK_STATE = K_STATE_SETUP);  -- end of Cal/Diag
            ACCEPT_THIS := (HK_STATE = K_STATE_SAFE);   -- HV off by DFEE
            DATA_TAKING_STOPPED := (HK_STATE /= EXP_STATE);

      when others =>
            -- no other exist
            null;
         end case;
      end if;

      if SETUP_IN_DT then
         -- normally DPE sees dummy fillers "End Of Data" on HSL, before
         -- state SETUP is seen in housekeeping.  then task HSL issues
         -- ACCEPT_STATE(SETUP) and the program will not enter this path.
         -- this may however easily happen that SETUP in HK will be noted
         -- first.  it is ok, provided that this situation is not permanent.
         -- so, singular occurencies of SETUP are accepted.  afterthat
         -- OEM will be generated - "unexpected state of DFEE".
         --
         if SETUP_IN_DT_CNT < 1 then
            SETUP_IN_DT_CNT := SETUP_IN_DT_CNT + 1;
         else
            OK := FALSE;
         end if;
      else
         SETUP_IN_DT_CNT := 0;
      end if;

      if ACCEPT_THIS then
         ACCEPT_STATE(HK_STATE);
      end if;

      if FORCE_SAFE then
         begin
            SET_STATE(K_STATE_SAFE);
         exception
            when others =>
               -- the reason for failure is not interesting in this moment
               -- it is enough that SET_STATE internally generates propper OEM.
               null;
         end;
      end if;

      if OK then
         -- inform the OEMs filtering system about error condition off.
         IASW_TM.DFEE_STATE_ERROR_CONDITION_OFF;
      else
         IASW_TM.PUT_OEM(
            IASW_TM.EV_DFEE_STATE,
            IASW_COMMON.CVT(EXP_STATE),
            IASW_COMMON.CVT(HK_STATE)
         );
      end if;
   end CHECK_STATE;


   procedure MODIFY_EXPECTED_STATE_IF_NEEDED is   -- local
   -- modifies the DFEE state expected in function of accessibility
   -- of DFEE.  accessibility in turn is function of low voltage status
   -- and of possibility of communication via LSL.
   begin
      if not CONTEXT.DFEE_ACCESSIBLE then
         CONTEXT.STATE := K_STATE_STANDBY;
      elsif CONTEXT.STATE < K_STATE_SAFE then
         CONTEXT.STATE := K_STATE_SAFE;
      else
         -- no indications exist for changing the expected state
         null;
      end if;
   end MODIFY_EXPECTED_STATE_IF_NEEDED;


   procedure BCP4_TICK is
   --called regularly every 1 second
   begin
      if CONTEXT.LV_IS_ON then
         if CONTEXT.DFEE_HEETING_TIME < 15 then
            CONTEXT.DFEE_HEETING_TIME := CONTEXT.DFEE_HEETING_TIME + 1;
         else
            CONTEXT.DFEE_HEETED := TRUE;
         end if;
      end if;
   end BCP4_TICK;


   procedure DFEE_ANSWERED is   -- local
   -- called whwnever DFEE answers properly on LSL
   begin
      CONTEXT.DFEE_LIVED_AT_LEAST_ONCE := TRUE;
      --new since ver.1.63

      CONTEXT.LSL_POSSIBLY_DEAD := FALSE;
      CONTEXT.LSL_DEAD := FALSE;
      CONTEXT.DFEE_ACCESSIBLE := TRUE;
      MODIFY_EXPECTED_STATE_IF_NEEDED;
   end DFEE_ANSWERED;


   procedure DFEE_DOES_NOT_ANSWER is   -- local
   -- called in DPE_DFEE_DIALOG when there was no answer on LSL
   -- after severall repetitions.
   begin
      if CONTEXT.STATE >= K_STATE_SAFE then
         CONTEXT.LSL_POSSIBLY_DEAD := CONTEXT.LV_IS_ON;
         CONTEXT.DFEE_ACCESSIBLE := FALSE;
         MODIFY_EXPECTED_STATE_IF_NEEDED;
      else
         null;
      end if;
   end DFEE_DOES_NOT_ANSWER;


   procedure LOW_VOLTAGE_IS_OFF is  -- global
   -- called in IASW_STATE.ANALYSE when low voltage is observed too low
   begin
      CONTEXT.LSL_POSSIBLY_DEAD := FALSE;
      CONTEXT.DFEE_ACCESSIBLE := FALSE;
      CONTEXT.DFEE_HEETING_TIME := 0;
      CONTEXT.DFEE_HEETED := FALSE;
      CONTEXT.LV_IS_ON := FALSE;
      CONTEXT.HV_IS_ON := FALSE;
      MODIFY_EXPECTED_STATE_IF_NEEDED;
   end LOW_VOLTAGE_IS_OFF;


   procedure LOW_VOLTAGE_IS_ON is  -- global
   -- called in IASW_STATE.ANALYSE when low voltage is observed o.k.
   begin
      CONTEXT.LV_IS_ON := TRUE;
      CONTEXT.LSL_DEAD := CONTEXT.LSL_POSSIBLY_DEAD;
      if CONTEXT.LSL_DEAD and CONTEXT.HV_IS_ON then
         -- lost communication with DFEE, HV must be switched off with relay.
         SEND_HV_OFF_RELAY_SIGNAL;
      end if;
      MODIFY_EXPECTED_STATE_IF_NEEDED;
   end LOW_VOLTAGE_IS_ON;


   function GET_STATE return T_DFEE_STATE is
   begin
      return CONTEXT.STATE;
   end GET_STATE;


   function HV_IS_ON return BOOLEAN is
   begin
      return CONTEXT.HV_IS_ON;
   end HV_IS_ON;


   function LSL_DEAD return BOOLEAN is
   begin
      return CONTEXT.LSL_DEAD;
   end LSL_DEAD;


   function DFEE_HEETED return BOOLEAN is
   begin
      return CONTEXT.DFEE_HEETED;
   end DFEE_HEETED;


   function DFEE_LIVED_AT_LEAST_ONCE return BOOLEAN is
   --new since ver.1.63
   begin
      return CONTEXT.DFEE_LIVED_AT_LEAST_ONCE;
   end DFEE_LIVED_AT_LEAST_ONCE;


   procedure GET_HW_SETTINGS( HWS : out T_HW_SETTINGS ) is
   begin
      HWS := CONTEXT.HWST;
   end GET_HW_SETTINGS;


   procedure GET_LAST_COMMANDED_HW_SETTINGS( HWS : out T_HW_SETTINGS ) is
   begin
      HWS := CONTEXT.HWTC;
   end GET_LAST_COMMANDED_HW_SETTINGS;


   procedure SET_SW_PARAM(
      PARAM_ID:  in IASW_COMMON.T_FID;
      PARAM_VAL: in IASW_COMMON.T_TC_PARAM_VAL;
      PARAM_LEN: in IASW_COMMON.T_SW_PARAM_LEN
   ) is
      FP : T_SW_PARAM_FLOAT;
         for FP use at PARAM_VAL(2)'address;   --just after the param.#
      --Comment on 'use at': it is used here to locate an element of
      --type float inside the array of words.
   begin
      if PARAM_VAL'length < PARAM_LEN then
         raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
      end if;

      if PARAM_ID = 1 then         -- set sw int params

         if PARAM_LEN /= 2 then
            raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
         end if;
         CHG_SW_PARAM_INT(
            PARAM_ID    => PARAM_VAL(1),
            PARAM_VALUE => PARAM_VAL(2) );
         SW_PARAM_OBT := DPE_CSSW.GET_CURRENT_OBT;

      elsif PARAM_ID = 2 then      -- set sw float params

         if PARAM_LEN /= 3 then
            raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
         end if;
         CHG_SW_PARAM_FLOAT(
            PARAM_ID    => PARAM_VAL(1),
            PARAM_VALUE => FP );
         SW_PARAM_OBT := DPE_CSSW.GET_CURRENT_OBT;

      elsif PARAM_ID = 31 then    -- ANODE setting

         if PARAM_LEN /= 1 then
            raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
         end if;
         CHG_HW_SETTING(
            COMMAND => 1,
            PARAM   => PARAM_VAL(1) );
         CONTEXT.HWTC.ANODE_SETTINGS    := CONTEXT.HWST.ANODE_SETTINGS;

      elsif PARAM_ID = 32 then    -- Low Level of Discrim. setting

         if PARAM_LEN /= 1 then
            raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
         end if;
         CHG_HW_SETTING(
            COMMAND => 2,
            PARAM   => PARAM_VAL(1) );
         CONTEXT.HWTC.DISCRIM_LOW_LEVEL := CONTEXT.HWST.DISCRIM_LOW_LEVEL;

      elsif PARAM_ID = 33 then    -- Delta(=Acc) HV setting

         if PARAM_LEN /= 1 then
            raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
         end if;
         CHG_HW_SETTING(
            COMMAND => 3,
            PARAM   => PARAM_VAL(1) );
         CONTEXT.HWTC.HV_ACC_SET        := CONTEXT.HWST.HV_ACC_SET;

      elsif PARAM_ID = 34 then    -- Cathode(=Drift) HV setting

         if PARAM_LEN /= 1 then
            raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
         end if;
         CHG_HW_SETTING(
            COMMAND => 4,
            PARAM   => PARAM_VAL(1) );
         CONTEXT.HWTC.HV_DRIFT_SET      := CONTEXT.HWST.HV_DRIFT_SET;

      elsif PARAM_ID = 4 then     -- HV Ready Command

         CHG_HW_SETTING(
            COMMAND => 0,
            PARAM   => 4 );

      elsif PARAM_ID = 5 then     -- HV On Command

         CHG_HW_SETTING(
            COMMAND => 0,
            PARAM   => 5 );
         CONTEXT.HWTC.HV_ACC_SET        := CONTEXT.HWST.HV_ACC_SET;
         CONTEXT.HWTC.HV_DRIFT_SET      := CONTEXT.HWST.HV_DRIFT_SET;

      elsif PARAM_ID = 6 then     -- HV Off Command

         CHG_HW_SETTING(
            COMMAND => 0,
            PARAM   => 6 );

      elsif PARAM_ID = 7 then     -- LOAD_CONTEXT Command

         LOAD_CONTEXT;

      elsif PARAM_ID = 8 then     -- SAVE_CONTEXT Command

         SAVE_CONTEXT;

      elsif PARAM_ID = 9 then     -- Setting of 8/16MHz mode  (new since 1.47)

         if PARAM_LEN /= 1 then
            raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
         end if;
         CHG_CPU_SETTING(
            VALUE   => PARAM_VAL(1) );

      elsif PARAM_ID = 14 then     -- Setting of ACC,DRIFT delta HV  (new since 1.49)

         if PARAM_LEN /= 2 then
            raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
         end if;
         DELTA_ACC_HV   := PARAM_VAL(1);
         DELTA_DRIFT_HV := PARAM_VAL(2);

      elsif PARAM_ID = 255 then   -- Test of dialog DPE-DFEE

         TEST_LSL_DIALOG(PARAM_VAL, PARAM_LEN);

      else
         raise IASW_COMMON.EX_SW_PARAM_ERROR_ID;
      end if;
   end SET_SW_PARAM;


   procedure GET_SW_PARAM(
      PARAM_ID:  in  IASW_COMMON.T_FID;
      PARAM_VAL: out IASW_COMMON.T_TM_PARAM_VAL;
      PARAM_LEN: out IASW_COMMON.T_SW_PARAM_LEN
   ) is
      SWPI : T_SW_PARAMS_INT_FIELD;
         for SWPI use at PARAM_VAL'address;
pragma If(((SWPI'size+15)/16) > PARAM_VAL'length);
   error: Violated safety of 'use at'
pragma End if;

      SWPF : T_SW_PARAMS_FLOAT_FIELD;
         for SWPF use at PARAM_VAL'address;
pragma If(((SWPF'size+15)/16) > PARAM_VAL'length);
   error: Violated safety of 'use at'
pragma End if;

   begin
      if PARAM_ID = 11 then            -- get all sw int params

         PARAM_LEN := SW_PARAMS_INT_NO;
         if PARAM_VAL'length < SW_PARAMS_INT_NO then
            raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
         end if;
         GET_SW_INT_PARAMS(SWPI);

      elsif PARAM_ID = 12 then         -- get all sw float params

         PARAM_LEN := SW_PARAMS_FLOAT_NO * 2;
         if PARAM_VAL'length < (SW_PARAMS_FLOAT_NO*2) then
            raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
         end if;
         GET_SW_FLOAT_PARAMS(SWPF);

      elsif PARAM_ID = 14 then         -- get delta HV's

         PARAM_VAL(1) := DELTA_ACC_HV;
         PARAM_VAL(2) := DELTA_DRIFT_HV;
         PARAM_LEN := 2;

      else
         PARAM_LEN := 0;
         raise IASW_COMMON.EX_SW_PARAM_ERROR_ID;
      end if;
   end GET_SW_PARAM;


   procedure GET_DPE_STATE(DS : out IASW_COMMON.T_DPE_STATE) is
   begin
      DS.PARAMETERS_OBT := SW_PARAM_OBT;
   end GET_DPE_STATE;


   procedure GET_HK_DIAG(
      FLAGS : out IASW_COMMON.T_16_BITS;
      DIAG  : out T_DIAGNOSTIC_HK
   ) is
   begin
      DIAG.DFEE_STATE := CONTEXT.STATE;
   end GET_HK_DIAG;


   procedure CLEAR_DIAGNOSTIC is
   begin
      DFEE_DIAGNOSTIC.TX_TOTAL_NO   := 0;
      DFEE_DIAGNOSTIC.RX_ERROR_NO   := 0;
      DFEE_DIAGNOSTIC.RX_CRC_NO     := 0;
      DFEE_DIAGNOSTIC.RX_TIMEOUT_NO := 0;
      DFEE_DIAGNOSTIC.RX_FAILURE_NO := 0;
   end CLEAR_DIAGNOSTIC;


   procedure GET_AND_CLEAR_DIAGNOSTIC_DATA( DIAG : out T_DFEE_DIAGNOSTIC ) is
   begin
      DIAG := DFEE_DIAGNOSTIC;
      CLEAR_DIAGNOSTIC;
   end GET_AND_CLEAR_DIAGNOSTIC_DATA;


   procedure SET_DEFAULT_CONTEXT is
   begin
      CONTEXT.STATE := K_STATE_STANDBY;
      CONTEXT.SWP_VALID := FALSE;
      CONTEXT.HWST.ANODE_SETTINGS    := 2#0000#;   --was 1111 prior to ver.1.62
      CONTEXT.HWST.DISCRIM_LOW_LEVEL := 15;        --was 0 prior to ver.1.62
      CONTEXT.HWST.HV_ACC_SET        := MIN_ACC_HV;
      CONTEXT.HWST.HV_DRIFT_SET      := MIN_DRIFT_HV;
      CONTEXT.HWTC := CONTEXT.HWST;
      CONTEXT.CPU := 0;
      CONTEXT.HV_IS_ON := FALSE;
      CONTEXT.LV_IS_ON := FALSE;
      CONTEXT.DFEE_ACCESSIBLE := FALSE;
      CONTEXT.DFEE_HEETING_TIME := 0;
      CONTEXT.DFEE_HEETED := FALSE;
      CONTEXT.LSL_DEAD := FALSE;
      CONTEXT.LSL_POSSIBLY_DEAD := FALSE;

      CONTEXT.DFEE_LIVED_AT_LEAST_ONCE := FALSE;
      --in ver.1.63 moved here from IASW_STATE

   end SET_DEFAULT_CONTEXT;


   procedure CONSTRUCTOR is
   begin
      -- Configure Low Speed Line
      SET_LSL_PARAMS (
         CLK_FREQ_ID   => 2,   --4K
         BDR_FACTOR_ID => 2    --16x
      );

      -- Initialise variables
      SET_DEFAULT_CONTEXT;
      CLEAR_DIAGNOSTIC;
      SW_PARAM_OBT := (0,0,0);
      DFEE_STATE_IN_HK_VALID := FALSE;
      SETUP_IN_DT_CNT := 0;
   end CONSTRUCTOR;

end IASW_DFEE;
