
with CSSW_COMMON_TYPES;  use CSSW_COMMON_TYPES;
with CSSW_IF_TYPES;
with IASW_COMMON;
with IASW_MODES;
with IASW_GF;

package IASW_HSL is

   package CIT renames CSSW_IF_TYPES;
   package CCT renames CSSW_COMMON_TYPES;

   subtype T_EVENTS_RATE is CCT.T_UNSIGNED_16;

   type T_DIAGNOSTIC_DATA is record
      HSL_READOUTS,
      CRC_ERRORS,
      TIMEOUTS,
      OVERFLOWS,
      LOST_SYNCHRONIZATION
         : CCT.T_UNSIGNED_16;
   end record;


   --------------------------------------------------------------------
   -- Local types and data, made global because of possible patching --
   --------------------------------------------------------------------

   EX_HSL_ERROR              : exception;
   EX_EVENT_SIZE_NOT_DEFINED : exception;

   -- selector of conditional compilation
   USE_EVENT_MARKER : constant BOOLEAN := TRUE;

   -- Identifiers of SW events attached to HSL HW events.
   HSL_RECEP_SIGNAL         : constant CIT.T_EVENT_NUMBER := 1;
   --HSL_CRC_ERROR_EVENT      : constant CIT.T_EVENT_NUMBER := 2;
   HSL_OVERFLOW_ERROR_EVENT : constant CIT.T_EVENT_NUMBER := 3;

   -- The HSL channel used for transmission.
   HSL_CHANNEL       : constant CIT.T_HIGH_SPEED_LINE := CIT.A;
   HSL_WORD_INTERVAL : constant CIT.T_WORD_TRANSFER_INTERVAL := CIT.CLK_032;

   -- Analog channel used to signal ,,Half of DPEE's FIFO is full''
   HALF_FIFO_ADC_CHANNEL : constant CIT.T_ANALOG_CHANNEL :=
                           CIT.ANALOG_ACQUISITION_12;

   -- max duration of the hsl transfer
   HSL_TIMEOUT    : constant CIT.T_RTC_TICKS := 2;

   -- max time of silence on hsl after terminating data taking,
   -- afterthat End-of-Data is assumed
   EVENTS_TIMEOUT : constant CCT.T_UNSIGNED_16 := 2; -- measured in BCP1 periods

   -- number of events needed for a full check of synchronization
   EVENTS_TO_CHECK  : constant CCT.T_UNSIGNED_16 := 3;

   -- The size (in 16-bit words) of current event size.  In case of
   -- test data taking it includes also the envelope - one leading marker.
   EVENT_SIZE : IASW_COMMON.T_EVENT_SIZE;

   -- event size including envelope (actually one extra word - event-marker)
   HSL_EVENT_SIZE : IASW_COMMON.T_EVENT_SIZE;

   -- Data used in checking the validity of events.
   EVENT_MARKER : constant CCT.T_UNSIGNED_16 := 16#F000#;
   LAST_EVENT_CC, LAST_EVENT_TC : CCT.T_UNSIGNED_16;
   HK_CC : CCT.T_UNSIGNED_16; -- Current value of HKC.

   -- flag indicating activity of TASK_HSL
   HSL_ACTIVE : BOOLEAN;
   -- It is set when starting data mode.
   -- It is cleared:
   --   in CONSTRUCTOR,
   --   when End Of Data dummy-event,
   --   when HSL timeout.
   --   (please note that only TASK_HSL can clear this flag).
   -- It is queried in:
   --   TASK_HSL to determine its activity,
   --   TASK_MAIN to report in HK TM,
   --   TASK_SC_OUT to make a decision about:
   --     Entering "Gentle Stop Mode" in which buffer is flushed and the
   --     last, partially filled TM block is also sent,
   --     Terminating data mode if there is no more data on input,
   --     Calling procedure changing Grey Filter.

   SYNCHR_USED      : BOOLEAN;    -- set for modes: Data Tak., Cal., Diag,
   SYNCHR_REQUIRED  : BOOLEAN;    -- set when error detected
   GF_ANALYSIS_REQUIRED : BOOLEAN;
   BCP1_FLAG : BOOLEAN;


   -- Buffer for communication with HSL.

   subtype T_HSL_BUF_INDEX is IASW_COMMON.T_EVENT_SIZE;  --CCT.T_UNSIGNED_16;

   -- max. number of words read from HSL at a time
   HSL_ABS_MAX_READ_SIZE : constant T_HSL_BUF_INDEX := 4096;

   -- upper limit of HSL readout use by the program
   -- this can be set smaller by the testing TC(5,3)/TID=1/FID=253
   HSL_MAX_READ_SIZE : T_HSL_BUF_INDEX := HSL_ABS_MAX_READ_SIZE;

   -- allocation size of the buffer
   HSL_BUFFER_SIZE   : constant T_HSL_BUF_INDEX :=
                       HSL_ABS_MAX_READ_SIZE + EVENTS_TO_CHECK * 43 + 1;
   --changed 42 to 43 in ver.1.62 (HSL readouts will always be 4096 since now on)

   -- Note, that real size of the buffer is greater than HSL_ABS_MAX_READ_SIZE
   -- The reason is that we need an extra space for at least
   -- EVENTS_TO_CHECK events.
   -- the last +1 is for the marker ending the last event.

   HSL_BUFFER_INDEX        : T_HSL_BUF_INDEX range 0 .. HSL_BUFFER_SIZE;
   -- this is the index to the array HSL_BUFFER() pointing the word which
   -- will be next peeked.

   HSL_BUFFER_WORDS : T_HSL_BUF_INDEX range 0 .. HSL_BUFFER_SIZE;
   -- this is the number of valid words in the array HSL_BUFFER().

   -- Address State in which HSL_BUFFER is defined
   HSL_BUFFER_AS : constant CIT.T_ADDRESS_STATE := 1;
   -- WARNING! Logical and phisical addresses of HSL_BUFFER should agree in
   -- the low significant 16-bits.  It can be chcked in the listing
   -- produced by linker.
   --      (comment taken from the file 'jemx.lnk'):
   -- assign 1.5o,15,3 -- To place HSL_BUFFER correctly.
   -- Always check 'jemx.map' if the placement of HSL_BUFFER is correct !!
   -- 1) find definition of 'HSL_BUFFER',  it can be like this:
   --                OPND       PHYS
   --                ADDR       ADDR          SYMBOL
   --        .....
   --                5C43      15C43         +HSL_BUFFER_$AAOISA503R
   -- 2) both addresses should be equal on last 4 hex digits,
   -- 3) log.addr should start with 5xxx, phys.addr should start with 15xxx,
   --    (compare with assign statement).
   -- 4) there should be enough memory for the buffer in the continous area
   --    defined by the assign statement.  HSL_BUFFER is 4096+3*42 (dec) words
   --    long. if log.addr > 5F00 then it start to be dangerous.
   --    assign 1.5o,15,3  could be a solution in this case.

   -- and finaly the buffer itself.
   HSL_BUFFER : IASW_COMMON.T_EVENT_WRAPPER(0 .. HSL_BUFFER_SIZE - 1);

   -- Statistics.

   EVENTS_TIME     : CCT.T_UNSIGNED_16;  -- BCP1 ticks between two readouts
                                        -- from HSL
   EVENTS_RATE     : T_EVENTS_RATE;     -- number of events received in
                                        -- last BCP1 period
   EVENTS_RATE_CNT : T_EVENTS_RATE;     -- counter of events received

   TOTAL_EVENTS_COUNT : CIT.T_U32_BITS; -- Total number of events read,
                                        -- never cleared.

   -- Diagnostic data for the package.

   DIAGNOSTIC_DATA : T_DIAGNOSTIC_DATA;
   HSL_DIAG_CODE   : CCT.T_UNSIGNED_16;  -- id# of the event or CSSW exception

   -- moved from IASW_SC_OUT:
   DEFAULT_PRIMARY_MIN_GF   : constant IASW_GF.T_GREY_FILTER_INDICATOR := 10;
   DEFAULT_SECONDARY_MAX_GF : constant IASW_GF.T_GREY_FILTER_INDICATOR := 25;
   subtype T_SC_FORMAT_SELECTOR is INTEGER range 1..2;  --primary/secondary
   SCF_TAB : array (T_SC_FORMAT_SELECTOR) of IASW_MODES.T_SCIENCE_FORMAT;
   SCF_SELECTED : T_SC_FORMAT_SELECTOR;
   GREY_FILTER_REPORTED : IASW_GF.T_GREY_FILTER_INDICATOR;
   PRIMARY_MIN_GF, SECONDARY_MAX_GF : IASW_GF.T_GREY_FILTER_INDICATOR;

   -- Lost Synchro. Diagnostic Dump
   type T_LOST_SYNCHRO_DUMP is record
      EVENT_SIZE : IASW_COMMON.T_EVENT_SIZE;
      HSL_BUFFER_WORDS : T_HSL_BUF_INDEX;
      HSL_BUFFER_INDEX : T_HSL_BUF_INDEX;
      LSD_OFFSET : T_HSL_BUF_INDEX;
      HSL_BUFFER : IASW_COMMON.T_EVENT_WRAPPER(1 .. 200);
   end record;
   LSD : T_LOST_SYNCHRO_DUMP;
   AUTOMATIC_LSD : BOOLEAN;

   -- starting signal + data mode params.
   type T_START_SIGNAL is record
      ON : BOOLEAN;
      EV_SIZE : IASW_COMMON.T_EVENT_SIZE;
      SCF1,SCF2 : IASW_MODES.T_SCIENCE_FORMAT;
      GFL1,GFL2 : CCT.T_UNSIGNED_16;
      SYNC : BOOLEAN;
   end record;
   START_SIGNAL : T_START_SIGNAL;

   -- stoping signal
   type T_STOP_SIGNAL is record
      ON : BOOLEAN;
   end record;
   STOP_SIGNAL : T_STOP_SIGNAL;


   ----------------------------------------------------------------
   -- Local procedures, made global because of possible patching --
   ----------------------------------------------------------------

   procedure CLEAR_LSD;
   procedure SET_LSD;
   procedure SEND_LSD;
   procedure CLEAR_HSL_BUFFER;
   procedure SWITCH_SCF;
   procedure FORMAT_SWITCHING_ANALISIS;
   function HSL_EVENTS_LEFT return CCT.T_UNSIGNED_16;
   procedure CONFIGURE_HSL(WTI : in CIT.T_WORD_TRANSFER_INTERVAL);
   procedure CONNECT_TO_HSL;
   function READ_HSL return BOOLEAN;
   function NEXT_EVENT_SYNCHRONIZED(
      INDEX   : in T_HSL_BUF_INDEX
      --PREV_CC : in CCT.T_UNSIGNED_16;
      --PREV_TC : in CCT.T_UNSIGNED_16
   ) return BOOLEAN;
   function EVENTS_SYNCHRONIZED(
      HK_CC  : in CCT.T_UNSIGNED_16;
      EVENTS : in CCT.T_UNSIGNED_16;
      INDEX  : in T_HSL_BUF_INDEX
   ) return BOOLEAN;
   procedure GET_EVENT(
      EVENT  : in out IASW_COMMON.T_EVENT;
      RESULT : out BOOLEAN );
   procedure LOCAL_START_DATA_TAKING;
   procedure LOCAL_STOP_DATA_TAKING;
   procedure CLEAR_DIAGNOSTIC_DATA;

   -----------------------
   -- Global procedures --
   -----------------------

   procedure SET_HSL_PARAMS (
      WT_INTERVAL  : in CCT.T_UNSIGNED_16;
      MAX_BLK_SIZE : in CCT.T_UNSIGNED_16 );

   procedure TASK_HSL;

   procedure START_DATA_TAKING(
      EV_SIZE   : in IASW_COMMON.T_EVENT_SIZE;
      SCF1,SCF2 : in IASW_MODES.T_SCIENCE_FORMAT;
      GFL1,GFL2 : in CCT.T_UNSIGNED_16;
      SYNC      : in BOOLEAN );

   procedure HARD_STOP;

   function STOPPED return BOOLEAN;

   procedure BCP1_TICK;

   procedure NEW_HK_CC(CC : in CCT.T_UNSIGNED_16);

   function GET_EVENTS_RATE return T_EVENTS_RATE;

   procedure SET_SW_PARAM(
      PARAM_ID:  in IASW_COMMON.T_FID;
      PARAM_VAL: in IASW_COMMON.T_TC_PARAM_VAL;
      PARAM_LEN: in IASW_COMMON.T_SW_PARAM_LEN);

   procedure GET_SW_PARAM(
      PARAM_ID:  in  IASW_COMMON.T_FID;
      PARAM_VAL: out IASW_COMMON.T_TM_PARAM_VAL;
      PARAM_LEN: out IASW_COMMON.T_SW_PARAM_LEN);

   procedure GET_AND_CLEAR_DIAGNOSTIC_DATA (DATA : out T_DIAGNOSTIC_DATA);

   procedure GET_DPE_STATE(DS : out IASW_COMMON.T_DPE_STATE);

   procedure GET_HK_DIAG( FLAGS : out IASW_COMMON.T_16_BITS );

   procedure CONSTRUCTOR;

end IASW_HSL;
