----------------------------------------------------------------------------
-- Module Name:        IASW_MAIN - package body
-- Project:            INTEGRAL, JEM-X, DPE
-- Author:             Dariusz Jaszkowski
-- Creation Date:      16.III.1998
-- Change History:
--    (Version/Date/Author/Description)
-- Function:
--    (Say what the component does)
--    Task Main - activates tasks HK and STATE when necessary,
--    Task TC - receives and executes telecommands,
-- Interfaces:
--    (Define the inputs and outputs)
-- Dependencies:
--    (Describe the preconditions for using this component)
-- Processing:
--    (Summarise the processing using pseudo-code or PDL)
----------------------------------------------------------------------------

with SYSTEM;
with UNCHECKED_CONVERSION;
with CSSW_COMMON_TYPES; use CSSW_COMMON_TYPES;
with CSSW_IF_TYPES;     use CSSW_IF_TYPES;
with DPE_CSSW;
with IASW_COMMON;
with IASW_VER;
with IASW_TIME;         use IASW_TIME;
with IASW_MODES;
with IASW_SC_OUT;
with IASW_DFEE;
with IASW_HK;
with IASW_HSL;
with IASW_MRTU;
with IASW_REQ_OUT;
with IASW_TM;
with IASW_STATE;


package body IASW_MAIN is

   -------------------------------------------------------------------------------
   -- procedure TASK_MAIN;
   --
   procedure TASK_MAIN is
      -- event ids
      BCP1_EVENT_ID : constant CIT.T_EVENT_NUMBER := CIT.T_EVENT_NUMBER'first;
      BCP4_EVENT_ID : constant CIT.T_EVENT_NUMBER := CIT.T_EVENT_NUMBER'first+1;

      -- task event list:
      LISTENED_EVENTS : CIT.T_EVENT_LIST := (others => false);
      ACQUIRED_EVENTS : CIT.T_EVENT_LIST;

      SW_DIAG_REPORT_ID : constant IASW_COMMON.T_PARAM_ID :=
         (TID => 1, FID => 16);

      BCP4_CNT : CIT.T_RTC_TICKS;
      DIAGCOUNTER : CCT.T_UNSIGNED_16;
      REQUEST: IASW_REQ_OUT.T_ON_RQ_TM_REQUEST;
   begin
      LISTENED_EVENTS(BCP1_EVENT_ID) := true;
      DPE_CSSW.CONNECT_TO_HW_EVENT(CIT.BCP1, BCP1_EVENT_ID);

      -- wait for first BCP1

      loop
         DPE_CSSW.WAIT_AND_CLEAR_EVENTS(LISTENED_EVENTS, BCP1_TIMEOUT, ACQUIRED_EVENTS);
         if ACQUIRED_EVENTS(BCP1_EVENT_ID) then
            exit;   -- O.K. stop waiting
         else
            IASW_COMMON.INC(MAIN_DIAGNOSTIC.BCP1_TIMEOUT_COUNT);
            IASW_TM.PUT_OEM(IASW_TM.EV_CLOCK_BCP1_LONG);
         end if;
      end loop;

      -- O.K. the first BCP1 received

      LISTENED_EVENTS(BCP4_EVENT_ID) := true;   -- entry for BCP1 is already true
      DPE_CSSW.CONNECT_TO_HW_EVENT(CIT.BCP4, BCP4_EVENT_ID);

      BCP4_CNT := 0;
      DIAGCOUNTER := 0;

      loop
         begin
            IASW_COMMON.I_AM_ALIVE(IASW_COMMON.TASK_MAIN_NO);

            -- wait for BCP4 or BCP1, whichever comes first
            DPE_CSSW.WAIT_AND_CLEAR_EVENTS(LISTENED_EVENTS, BCP4_TIMEOUT, ACQUIRED_EVENTS);

            if ACQUIRED_EVENTS(BCP4_EVENT_ID) then
               -- actions after BCP4
               IASW_DFEE.BCP4_TICK;
               BCP4_CNT := BCP4_CNT + 1;
            end if;

            if ACQUIRED_EVENTS(BCP1_EVENT_ID) then
               -- actions after BCP1
               IASW_TIME.   BCP1_TICK;
               IASW_HSL.    BCP1_TICK;
               IASW_SC_OUT. BCP1_TICK;
               IASW_STATE.  BCP1_TICK;
               IASW_TM.     BCP1_TICK;

               if IASW_DIAGNOSTIC_PERIOD > 0 then
                  -- diagnostic TM is to be generated every period.
                  DIAGCOUNTER := DIAGCOUNTER + 1;
                  if DIAGCOUNTER >= IASW_DIAGNOSTIC_PERIOD then
                     DPE_CSSW.ENTER_CRITICAL_REGION;
                     if not IASW_REQ_OUT.REQUESTS_QUEUE_FULL then
                        REQUEST.CODE := IASW_REQ_OUT.TASK_PARAMETER_REPORT;
                        REQUEST.PARAM_ID := SW_DIAG_REPORT_ID;
                        REQUEST.OEMF2 := -1;
                        REQUEST.OEMF3 := -1;
                        IASW_REQ_OUT.ADD_REQUEST(REQUEST);
                     end if;
                     DPE_CSSW.LEAVE_CRITICAL_REGION;
                     DIAGCOUNTER := 0;
                  end if;
               end if;

               if BCP4_CNT < 7 then
                  -- last BCP1 period was too short
                  IASW_COMMON.INC(MAIN_DIAGNOSTIC.BCP1_ERROR_COUNT);
                  IASW_TM.PUT_OEM(IASW_TM.EV_CLOCK_BCP1_SHORT);
               else
                  IASW_TM.BCP1_ERROR_CONDITION_OFF;
               end if;
               BCP4_CNT := 0;
            end if;

            if BCP4_CNT > 9 then
               -- no BCP1 after expected time
               IASW_COMMON.INC(MAIN_DIAGNOSTIC.BCP1_TIMEOUT_COUNT);
               IASW_TM.PUT_OEM(IASW_TM.EV_CLOCK_BCP1_LONG);
               BCP4_CNT := 0;
            end if;

            if ACQUIRED_EVENTS(BCP1_EVENT_ID) or
               ACQUIRED_EVENTS(BCP4_EVENT_ID)
            then
               -- action after BCP4_CNT got a new value:
               case BCP4_CNT is
                  when 0 | 1 | 2 | 3 | 4 =>
                     if IASW_STATE.HAVE_PENDING_ANALYSE then
                        IASW_STATE.SET_WANTED_SHUTDOWN_LEVEL;
                     end if;

                  when 6 =>
                     IASW_COMMON.START_TASK(IASW_COMMON.TASK_HK_ID);

                  when others =>
                     null;
               end case;
            else
               -- no BCP4 after timeout (2 sec)
               IASW_COMMON.INC(MAIN_DIAGNOSTIC.RTC_TIMEOUT_COUNT);
               IASW_TM.PUT_OEM(IASW_TM.EV_CSSW_NO_RTC, DPE_CSSW.CURRENT_TASK);
            end if;

         exception
            when others =>
               IASW_TM.PUT_OEM(IASW_TM.EV_IASW_EXCEPTION, DPE_CSSW.CURRENT_TASK);
         end;
      end loop;
   end TASK_MAIN;


   procedure TC_MODE_TRANSITION(TC_DATA: in CIT.T_TC_APPL_DATA_FIELD) is
   begin
      IASW_STATE.TC55(TC_DATA);
   end TC_MODE_TRANSITION;


   procedure TC_LOAD_TASK_PARAMS(TC_DATA: in CIT.T_TC_APPL_DATA_FIELD) is
      type T_TC_BLOCK is record
         PARAM_ID  : IASW_COMMON.T_PARAM_ID;
         PARAM_VAL : IASW_COMMON.T_TC_PARAM_VAL;
      end record;

      TC : T_TC_BLOCK;
         for TC use at TC_DATA.WORDS'address;
pragma If(((TC'size+15)/16) > TC_DATA.WORDS'length);
   error: Violated safety of 'use at'
pragma End if;

      PARAM_LEN : IASW_COMMON.T_SW_PARAM_LEN;
   begin
      if TC_DATA.LEN < 1 then  -- to be sure that TID/FID is given
         raise EX_LENGTH;
      end if;

      PARAM_LEN := TC_DATA.LEN - 1;

      case TC.PARAM_ID.TID is
         when TID_DFEE =>
            IASW_DFEE.SET_SW_PARAM(TC.PARAM_ID.FID, TC.PARAM_VAL, PARAM_LEN);
         when TID_MAIN =>
            SET_SW_PARAM(TC.PARAM_ID.FID, TC.PARAM_VAL, PARAM_LEN);
         when TID_HSL =>
            IASW_HSL.SET_SW_PARAM(TC.PARAM_ID.FID, TC.PARAM_VAL, PARAM_LEN);
         when TID_SC_OUT =>
            IASW_SC_OUT.SET_SW_PARAM(TC.PARAM_ID.FID, TC.PARAM_VAL, PARAM_LEN);
         when others =>  -- invalid TID value
            raise EX_TID_FID;
      end case;
   end TC_LOAD_TASK_PARAMS;


   procedure TC_REPORT_TASK_PARAMS(TC_DATA: in CIT.T_TC_APPL_DATA_FIELD) is
      PARAM_ID : IASW_COMMON.T_PARAM_ID;
         for PARAM_ID use at TC_DATA.WORDS'address;
pragma If(((PARAM_ID'size+15)/16) > TC_DATA.WORDS'length);
   error: Violated safety of 'use at'
pragma End if;

      REQUEST : IASW_REQ_OUT.T_ON_RQ_TM_REQUEST;
   begin
      if TC_DATA.LEN /= 1 then  -- to be sure that TID/FID is given
         raise EX_LENGTH;
      end if;

      REQUEST.CODE := IASW_REQ_OUT.TASK_PARAMETER_REPORT;
      REQUEST.PARAM_ID := PARAM_ID;
      REQUEST.OEMF2 := IASW_COMMON.CVT(TC_OEM_DATA_FIELD2);
      REQUEST.OEMF3 := IASW_COMMON.CVT(TC_OEM_DATA_FIELD3);
      IASW_REQ_OUT.ADD_REQUEST(REQUEST);
   end TC_REPORT_TASK_PARAMS;


   procedure TC_LOAD_MEMORY(TC_DATA: in CIT.T_TC_APPL_DATA_FIELD) is
      HDRLEN : constant INTEGER := IASW_COMMON.T_TC_LOAD_MEM_HDR'size / 16;  -- =3
      MAXDATALEN : constant INTEGER := TC_DATA.WORDS'length - HDRLEN;
      type T_TC_BLOCK is record
         MEM_HDR : IASW_COMMON.T_TC_LOAD_MEM_HDR;
         MEM_CONTENT : IASW_DFEE.T_MEM_BLOCK(1..MAXDATALEN); --was ..IASW_DFEE.K_MAX_LOAD_MEM_BLK_LEN);
      end record;

      TC : T_TC_BLOCK;
         for TC use at TC_DATA.WORDS'address;
pragma If(((TC'size+15)/16) > TC_DATA.WORDS'length);
   error: Violated safety of 'use at'
pragma End if;

      MEM_DESCR : IASW_DFEE.T_MEM_BLOCK_SPEC;
      MEM_BLK_LEN : INTEGER;
   begin
      if TC_DATA.LEN < 4 then
         raise EX_LENGTH;
      elsif TC.MEM_HDR.MID /= MID_DFEE then
         raise EX_MID;
      end if;

      MEM_BLK_LEN := TC_DATA.LEN - HDRLEN;
      if (TC.MEM_HDR.ADDR not in IASW_DFEE.T_DFEE_MEM_ADDRESS) or
         (MEM_BLK_LEN not in 1..MAXDATALEN) or  -- was ..IASW_DFEE.K_MAX_LOAD_MEM_BLK_LEN) or
         ((TC.MEM_HDR.ADDR + IASW_COMMON.CVT(MEM_BLK_LEN)-1)
         not in IASW_DFEE.T_DFEE_MEM_ADDRESS)
      then
          raise EX_ADDRESS;
      end if;

      MEM_DESCR.ADDR := TC.MEM_HDR.ADDR;
      MEM_DESCR.LENGTH := MEM_BLK_LEN;
      IASW_DFEE.LOAD_MEMORY(MEM_DESCR, TC.MEM_CONTENT);
   end TC_LOAD_MEMORY;


   procedure TC_DUMP_MEMORY(TC_DATA: in CIT.T_TC_APPL_DATA_FIELD) is
      MEM_HDR : IASW_COMMON.T_TC_MEM_HDR;
         for MEM_HDR use at TC_DATA.WORDS'address;
pragma If(((MEM_HDR'size+15)/16) > TC_DATA.WORDS'length);
   error: Violated safety of 'use at'
pragma End if;

      REQUEST : IASW_REQ_OUT.T_ON_RQ_TM_REQUEST;
   begin
      if TC_DATA.LEN /= 4 then
         raise EX_LENGTH;
      elsif MEM_HDR.MID /= MID_DFEE then
         raise EX_MID;
      end if;

      REQUEST.CODE := IASW_REQ_OUT.MEMORY_DUMP;
      REQUEST.MEM_HDR := MEM_HDR;
      REQUEST.OEMF2 := IASW_COMMON.CVT(TC_OEM_DATA_FIELD2);
      REQUEST.OEMF3 := IASW_COMMON.CVT(TC_OEM_DATA_FIELD3);
      IASW_REQ_OUT.ADD_REQUEST(REQUEST);
   end TC_DUMP_MEMORY;


   procedure TC_GET_MEMORY_CRC(TC_DATA: in CIT.T_TC_APPL_DATA_FIELD) is
      MEM_HDR : IASW_COMMON.T_TC_MEM_HDR;
         for MEM_HDR use at TC_DATA.WORDS'address;
pragma If(((MEM_HDR'size+15)/16) > TC_DATA.WORDS'length);
   error: Violated safety of 'use at'
pragma End if;

      REQUEST : IASW_REQ_OUT.T_ON_RQ_TM_REQUEST;
   begin
      if TC_DATA.LEN /= 4 then
         raise EX_LENGTH;
      elsif MEM_HDR.MID /= MID_DFEE then
         raise EX_MID;
      end if;

      REQUEST.CODE := IASW_REQ_OUT.MEMORY_CRC;
      REQUEST.MEM_HDR := MEM_HDR;
      REQUEST.OEMF2 := IASW_COMMON.CVT(TC_OEM_DATA_FIELD2);
      REQUEST.OEMF3 := IASW_COMMON.CVT(TC_OEM_DATA_FIELD3);
      IASW_REQ_OUT.ADD_REQUEST(REQUEST);
   end TC_GET_MEMORY_CRC;


   procedure TC_BROADCAST_PACKET(TC_DATA: in CIT.T_TC_APPL_DATA_FIELD) is
      BCPKT : IASW_COMMON.T_BROADCAST_PACKET;
         for BCPKT use at TC_DATA.WORDS'address;
pragma If(((BCPKT'size+15)/16) > TC_DATA.WORDS'length);
   error: Violated safety of 'use at'
pragma End if;

   begin
      if TC_DATA.LEN < (BCPKT'size / 16) then
         raise EX_LENGTH;
      end if;

      IASW_STATE.UPDATE_BCPKT(BCPKT);
   end;


   procedure PREPARE_TC_OEM_INFO(TC : in CIT.T_TCSP) is   -- local
      type T_F2 is record
         PARAM    : CCT.T_UNSIGNED_8;
         LSB_APID : CCT.T_UNSIGNED_8;
      end record;
      for T_F2 use record
         PARAM    at 0 range 0..7;
         LSB_APID at 0 range 8..15;
      end record;

      function CVT is new UNCHECKED_CONVERSION(
         SOURCE => T_F2, TARGET => CCT.T_UNSIGNED_16);

      function CVT is new UNCHECKED_CONVERSION(
         SOURCE => CIT.T_TC_PACKET_HEAD_DATA, TARGET => CCT.T_UNSIGNED_16);

      F2 : T_F2;
      F3 : CIT.T_TC_PACKET_HEAD_DATA;
   begin
      F2.PARAM := 0;
      F2.LSB_APID := CCT.T_UNSIGNED_8(IASW_COMMON.APID_TC mod 256);
      F3 := TC.PCK_HEAD;
      F3.SPARE := 0;
      TC_OEM_DATA_FIELD2 := CVT(F2);
      TC_OEM_DATA_FIELD3 := CVT(F3);
   end PREPARE_TC_OEM_INFO;


   procedure EXECUTE_TC(TC : in CIT.T_TCSP) is   -- local
   begin
      IASW_STATE.CHECK_TC_CORRECTNESS_VERSUS_STATE(TC);

      if TC.DAT_HEAD = (TYP => 5, SUB => 3) then
         TC_LOAD_TASK_PARAMS(TC.DATA);

      elsif TC.DAT_HEAD = (TYP => 5, SUB => 4) then
         TC_REPORT_TASK_PARAMS(TC.DATA);

      elsif TC.DAT_HEAD = (TYP => 5, SUB => 5) then
         TC_MODE_TRANSITION(TC.DATA);

      elsif TC.DAT_HEAD = (TYP => 6, SUB => 1) then
         TC_LOAD_MEMORY(TC.DATA);

      elsif TC.DAT_HEAD = (TYP => 6, SUB => 2) then
         TC_DUMP_MEMORY(TC.DATA);

      elsif TC.DAT_HEAD = (TYP => 6, SUB => 3) then
         TC_GET_MEMORY_CRC(TC.DATA);

      elsif TC.DAT_HEAD = (TYP => 13, SUB => 1) then
         null;

      elsif TC.DAT_HEAD = (TYP => 15, SUB => 1) then
         TC_BROADCAST_PACKET(TC.DATA);

      else
         raise EX_TYPE_SUBTYPE;
      end if;

   exception

      when EX_TYPE_SUBTYPE =>
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_REJECTED_COUNT);
         IASW_TM.PUT_OEM(IASW_TM.EV_TC_REJ_TYPE_SUBTYPE,
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD2),
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD3));

      when EX_TID_FID |
           IASW_COMMON.EX_SW_PARAM_ERROR_ID =>
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_REJECTED_COUNT);
         IASW_TM.PUT_OEM(IASW_TM.EV_TC_REJ_TID_FID,
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD2),
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD3));

      when EX_MID =>
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_REJECTED_COUNT);
         IASW_TM.PUT_OEM(IASW_TM.EV_TC_REJ_MID,
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD2),
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD3));

      when EX_ADDRESS =>
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_REJECTED_COUNT);
         IASW_TM.PUT_OEM(IASW_TM.EV_TC_REJ_ADDRESS,
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD2),
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD3));

      when EX_LENGTH |
           IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH =>
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_REJECTED_COUNT);
         IASW_TM.PUT_OEM(IASW_TM.EV_TC_REJ_LENGTH,
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD2),
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD3));

      when EX_DATA |
           IASW_COMMON.EX_SW_PARAM_ERROR_VALUE |
           IASW_DFEE.EX_DFEE_CALLER_ERROR |
           IASW_DFEE.EX_PATCHES_LENGTH_ERROR =>
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_REJECTED_COUNT);
         IASW_TM.PUT_OEM(IASW_TM.EV_TC_REJ_DATA,
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD2),
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD3));

      when EX_BUSY |
           IASW_REQ_OUT.EX_REQUEST_QUEUE_FULL =>
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_REJECTED_COUNT);
         IASW_TM.PUT_OEM(IASW_TM.EV_TC_REJ_BUSY,
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD2),
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD3));

      when EX_SAFETY =>
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_REJECTED_COUNT);
         IASW_TM.PUT_OEM(IASW_TM.EV_TC_REJ_SAFETY,
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD2),
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD3));

      when EX_STATE =>
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_REJECTED_COUNT);
         IASW_TM.PUT_OEM(IASW_TM.EV_TC_REJ_STATE,
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD2),
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD3));

      when EX_EXEC_FAILED |
           IASW_DFEE.EX_LSL_DEAD |
           IASW_DFEE.EX_DFEE_BUSY |
           IASW_DFEE.EX_DFEE_DETECTS_ERROR |
           IASW_DFEE.EX_DFEE_UNEXP_ANSWER =>
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_REJECTED_COUNT);
         IASW_TM.PUT_OEM(IASW_TM.EV_TC_REJ_EXEC_FAILED,
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD2),
            IASW_COMMON.CVT(TC_OEM_DATA_FIELD3));

      when IASW_DFEE.EX_PATCHES_TABLE_FULL =>
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_REJECTED_COUNT);
         IASW_TM.PUT_OEM(IASW_TM.EV_IASW_PATCHES,
            DPE_CSSW.CURRENT_TASK);

      when others =>
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_REJECTED_COUNT);
         IASW_TM.PUT_OEM(IASW_TM.EV_IASW_EXCEPTION,
            DPE_CSSW.CURRENT_TASK);
   end EXECUTE_TC;


   procedure WAIT_FOR_VALID_TC(TCSP : in out CIT.T_OPT_TCSP) is   -- local
   begin
      loop
         DPE_CSSW.GET_TC(IASW_COMMON.APID_TC, ICB_READ_TIMEOUT, TCSP);
         exit when TCSP.VALID;
         if DPE_CSSW.GET_ICB_STATE(CIT.TC_IASW).USED > 0 then
            IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_TIMEOUT_COUNT);
         end if;
      end loop;
   end WAIT_FOR_VALID_TC;


   procedure TASK_TC is
      TCSP : CIT.T_OPT_TCSP;   -- this must be on stack
   begin
      IASW_DFEE.CONNECT_TO_LSL;
      loop
         IASW_COMMON.I_AM_ALIVE(IASW_COMMON.TASK_TC_NO);
         WAIT_FOR_VALID_TC(TCSP);
         IASW_COMMON.INC(MAIN_DIAGNOSTIC.TC_COUNT);
         PREPARE_TC_OEM_INFO(TCSP.TC);
         IASW_STATE.ENTER_STATE_CRITSEC;
         EXECUTE_TC(TCSP.TC);
         IASW_STATE.LEAVE_STATE_CRITSEC;
      end loop;
   end TASK_TC;


   procedure SET_SW_PARAM(
      PARAM_ID :  in IASW_COMMON.T_FID;
      PARAM_VAL : in IASW_COMMON.T_TC_PARAM_VAL;
      PARAM_LEN : in IASW_COMMON.T_SW_PARAM_LEN
   ) is
   begin
      if PARAM_VAL'length < PARAM_LEN then
         raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
      end if;

      case PARAM_ID is

         when 15 =>  -- period of automatic generation of IASW Diagnostic report
            if PARAM_LEN /= 1 then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            else
               IASW_DIAGNOSTIC_PERIOD := PARAM_VAL(1);
            end if;

         when others =>
            IASW_STATE.SET_SW_PARAM(PARAM_ID, PARAM_VAL, PARAM_LEN);

      end case;
   end SET_SW_PARAM;


   procedure GET_SW_PARAM(
      PARAM_ID :  in  IASW_COMMON.T_FID;
      PARAM_VAL : out IASW_COMMON.T_TM_PARAM_VAL;
      PARAM_LEN : out IASW_COMMON.T_SW_PARAM_LEN
   ) is
      -- record of the IASW diagnostic data:
      type T_IASW_DIAG is record
          MARKER1 : CCT.T_UNSIGNED_16;
          TC_DIAG : IASW_MAIN.T_TC_DIAGNOSTIC;
          MARKER2 : CCT.T_UNSIGNED_16;
          MRTU_DIAG : IASW_MRTU.T_MRTU_DIAGNOSTIC;
          MARKER3 : CCT.T_UNSIGNED_16;
          DFEE_DIAG : IASW_DFEE.T_DFEE_DIAGNOSTIC;
          MARKER4 : CCT.T_UNSIGNED_16;
          HSL_DIAG : IASW_HSL.T_DIAGNOSTIC_DATA;
          MARKER5 : CCT.T_UNSIGNED_16;
      end record;
      IASW_DIAG_DATA : T_IASW_DIAG;
         for IASW_DIAG_DATA use at PARAM_VAL'address;
pragma If(((IASW_DIAG_DATA'size+15)/16) > PARAM_VAL'length);
   error: Violated safety of 'use at'
pragma End if;

      IASW_DIAG_LEN : constant INTEGER := IASW_DIAG_DATA'size/16;

      procedure GET_IASW_DIAG(IASW_DIAG_DATA: out T_IASW_DIAG) is
      begin
         IASW_DIAG_DATA.MARKER1 := 16#1111#;
         IASW_DIAG_DATA.MARKER2 := 16#2222#;
         IASW_DIAG_DATA.MARKER3 := 16#3333#;
         IASW_DIAG_DATA.MARKER4 := 16#4444#;
         IASW_DIAG_DATA.MARKER5 := 16#5555#;
         IASW_MAIN.GET_AND_CLEAR_DIAGNOSTIC_DATA(IASW_DIAG_DATA.TC_DIAG);
         IASW_MRTU.GET_AND_CLEAR_DIAGNOSTIC_DATA(IASW_DIAG_DATA.MRTU_DIAG);
         IASW_DFEE.GET_AND_CLEAR_DIAGNOSTIC_DATA(IASW_DIAG_DATA.DFEE_DIAG);
         IASW_HSL. GET_AND_CLEAR_DIAGNOSTIC_DATA(IASW_DIAG_DATA.HSL_DIAG);
      end GET_IASW_DIAG;

   begin
      PARAM_LEN := 0;   --preliminary

      case PARAM_ID is
         when 11 =>  -- IASW version
            PARAM_LEN := 2;
            if PARAM_VAL'length < 2 then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            end if;
            PARAM_VAL(1) := IASW_VER.VER_MAJOR;
            PARAM_VAL(2) := IASW_VER.VER_MINOR;

         when 15 =>  -- period of automatic generation of IASW Diagnostic report
            PARAM_LEN := 1;
            if PARAM_VAL'length < 1 then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            end if;
            PARAM_VAL(1) := IASW_DIAGNOSTIC_PERIOD;

         when 16 =>  -- IASW Diagnostic report
            PARAM_LEN := IASW_DIAG_LEN;
            if PARAM_VAL'length < IASW_DIAG_LEN then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            end if;
            GET_IASW_DIAG(IASW_DIAG_DATA);

         when others =>
            IASW_STATE.GET_SW_PARAM(PARAM_ID, PARAM_VAL, PARAM_LEN);
      end case;
   end GET_SW_PARAM;


   procedure CLEAR_DIAGNOSTIC_DATA is
   begin
      MAIN_DIAGNOSTIC.TC_COUNT           := 0;
      MAIN_DIAGNOSTIC.TC_REJECTED_COUNT  := 0;
      MAIN_DIAGNOSTIC.TC_TIMEOUT_COUNT   := 0;
      MAIN_DIAGNOSTIC.BCP1_TIMEOUT_COUNT := 0;
      MAIN_DIAGNOSTIC.BCP1_ERROR_COUNT   := 0;
      MAIN_DIAGNOSTIC.RTC_TIMEOUT_COUNT  := 0;
   end;


   procedure GET_AND_CLEAR_DIAGNOSTIC_DATA(DATA: out T_TC_DIAGNOSTIC) is
   begin
      DATA := MAIN_DIAGNOSTIC;
      CLEAR_DIAGNOSTIC_DATA;
   end GET_AND_CLEAR_DIAGNOSTIC_DATA;


   procedure CONSTRUCTOR is
   begin
      CLEAR_DIAGNOSTIC_DATA;
      IASW_DIAGNOSTIC_PERIOD := 0;
   end CONSTRUCTOR;

end IASW_MAIN;
