----------------------------------------------------------------------------
-- Module Name:        IASW_STATE - package body
-- Project:            INTEGRAL, JEM-X, DPE
-- Author:             Dariusz Jaszkowski
-- Creation Date:      16.III.1998
-- Change History:
--    (Version/Date/Author/Description)
--    see Release Notes - the separate document
-- Function:
--    (Say what the component does)
--    Task STATE performs shutdown and recovery of data taking.
-- Interfaces:
--    (Define the inputs and outputs)
-- Dependencies:
--    (Describe the preconditions for using this component)
-- Processing:
--    (Summarise the processing using pseudo-code or PDL)
--    This package contains many data accessed by several tasks.
--    STATE_SEMAPHORE is used to synchronize access to those data.  The
--    following actions are performed inside the critical region:
--    1) Whole activity of TASK_STATE,
--    2) Execution of any telecommand not generating on-req. telemetry,
--    3) Execution of proc. ANALYSE_STATE (called in TASK_MAIN,TASK_TC,
--       TASK_HK).
----------------------------------------------------------------------------

with system;
with UNCHECKED_CONVERSION;
with CSSW_COMMON_TYPES; use CSSW_COMMON_TYPES;
with CSSW_IF_TYPES;     use CSSW_IF_TYPES;
with DPE_CSSW;
with IASW_COMMON;
with IASW_DFEE;
with IASW_HSL;
with IASW_MODES;
with IASW_TIME;         use IASW_TIME;
with IASW_TM;
with IASW_MAIN;
with IASW_GF;


package body IASW_STATE is

   function CVT is new UNCHECKED_CONVERSION(
      SOURCE => T_BCPKT_FLAGS, TARGET => CCT.T_UNSIGNED_16);

   function CVT is new UNCHECKED_CONVERSION(
      SOURCE => CCT.T_UNSIGNED_16, TARGET => T_BCPKT_FLAGS);

   function CVT is new UNCHECKED_CONVERSION(
      SOURCE => T_HK_FLAGS, TARGET => CCT.T_UNSIGNED_16);

   function CVT is new UNCHECKED_CONVERSION(
      SOURCE => CCT.T_UNSIGNED_16, TARGET => T_HK_FLAGS);

   function CVT is new UNCHECKED_CONVERSION(
      SOURCE => T_AUTO_RECOVERY_CONTROL_INFO, TARGET => CCT.T_UNSIGNED_16);

   function CVT is new UNCHECKED_CONVERSION(
      SOURCE => CCT.T_UNSIGNED_16, TARGET => T_AUTO_RECOVERY_CONTROL_INFO);


   procedure ENTER_STATE_CRITSEC is   -- public
   begin
      loop
         exit when IASW_COMMON.CAN_ENTER_CRITSEC(STATE_SEMAPHORE'address);
         DPE_CSSW.DELAY_RELATIVE(1);
      end loop;
   end ENTER_STATE_CRITSEC;

   function CAN_ENTER_STATE_CRITSEC return BOOLEAN is   -- local
   -- if this function returns TRUE, you are in the critical section
   -- and it is necessary to use LEAVE_STATE_CRITSEC after the action
   begin
      return IASW_COMMON.CAN_ENTER_CRITSEC(STATE_SEMAPHORE'address);
   end;

   procedure LEAVE_STATE_CRITSEC is   -- public
   begin
      STATE_SEMAPHORE := 0;
   end LEAVE_STATE_CRITSEC;


   function DEEPER_SHUTDOWN_LEVEL return T_SHUTDOWN_LEVEL is   -- local
   begin
      if ACTUAL_SHUTDOWN_LEVEL > WANTED_SHUTDOWN_LEVEL then
         return ACTUAL_SHUTDOWN_LEVEL;
      else
         return WANTED_SHUTDOWN_LEVEL;
      end if;
   end DEEPER_SHUTDOWN_LEVEL;


   -- methodes of the object T_RAD_MONITOR:

   procedure RADMON_INIT(   -- local
      RM : out T_RAD_MONITOR;
      HL,LL : in T_RAD_MONITOR_COUNTER
   ) is
   begin
      RM.VALUE := 0;
      RM.HIGH_LIMIT := HL;
      RM.LOW_LIMIT := LL;
      RM.IS_HIGH := FALSE;
      RM.CNT_OPPOSIT := 0;
   end RADMON_INIT;

   procedure RADMON_NEW_VALUE(   -- local
      RM : in out T_RAD_MONITOR;
      V : in T_RAD_MONITOR_COUNTER
   ) is
   begin
      RM.VALUE := V;
   end RADMON_NEW_VALUE;

   procedure RADMON_BCP1_TICK( RM : in out T_RAD_MONITOR ) is   -- local
      CNT_LIMIT : CCT.T_UNSIGNED_16;
   begin
      if ((RM.VALUE < RM.LOW_LIMIT) and RM.IS_HIGH) or
         ((RM.VALUE > RM.HIGH_LIMIT) and not RM.IS_HIGH)
      then
         RM.CNT_OPPOSIT := RM.CNT_OPPOSIT + 1;
         if RM.IS_HIGH then
            CNT_LIMIT := RADMON_EXIT_CNT;
         else
            CNT_LIMIT := RADMON_ENTRY_CNT;
         end if;
         if RM.CNT_OPPOSIT >= CNT_LIMIT then
            RM.IS_HIGH := not RM.IS_HIGH;
            RM.CNT_OPPOSIT := 0;
         end if;
      else
         RM.CNT_OPPOSIT := 0;
      end if;
   end RADMON_BCP1_TICK;

   function RADMON_HIGH(   -- local
      RM : in T_RAD_MONITOR;
      FILTER, CONDITION : in BOOLEAN
   ) return BOOLEAN is
   begin
      if FILTER then
         return CONDITION;
      else
         return RM.IS_HIGH;
      end if;
   end RADMON_HIGH;


   --methodes of the object T_TIME_INTERVAL:

   function BCPKT_TO_RBI_OBT(   -- local
      OBT1 : in IASW_COMMON.T_BCPKT_OBT
   ) return CIT.T_OBT is
      OBT2 : CIT.T_OBT;
   begin
      OBT2.HIGH := OBT1.HIGH;
      OBT2.MID  := OBT1.LOW;
      OBT2.LOW  := 0;
      return OBT2;
   end BCPKT_TO_RBI_OBT;

   procedure INTERVAL_INIT(   -- local
      INTERVAL : out T_TIME_INTERVAL
   ) is
   begin
      INTERVAL.ENTRY_TIME := K_OBT_UNDEFINED;
      INTERVAL.ENTRY_OK   := FALSE;
      INTERVAL.EXIT_TIME  := K_OBT_UNDEFINED;
      INTERVAL.EXIT_OK    := FALSE;
   end INTERVAL_INIT;

   procedure INTERVAL_SET(   -- local
      INTERVAL : in out T_TIME_INTERVAL;
      T1,T2 : in IASW_COMMON.T_BCPKT_OBT
   ) is
      use IASW_COMMON;  --to be able to compare IASW_COMMON.T_BCPKT_OBT-s
   begin
      -- valid since ver 1.33.  Introduced, because people from Alenia
      -- claim that times in BCPKT should be used as they are,
      -- without storing.  Times are still stored, just not to change
      -- the program too much, but they are stored unconditionally, which
      -- results as if the program used times stright from BCPKT.
      INTERVAL.ENTRY_TIME := BCPKT_TO_RBI_OBT(T1);
      INTERVAL.ENTRY_OK := (T1 /= K_BCPKT_OBT_UNDEFINED);
      INTERVAL.EXIT_TIME := BCPKT_TO_RBI_OBT(T2);
      INTERVAL.EXIT_OK := (T2 /= K_BCPKT_OBT_UNDEFINED);
   end INTERVAL_SET;

   function TIME_INSIDE_INTERVAL(   -- local
      INTERVAL : in T_TIME_INTERVAL;
      TIME : in CIT.T_OBT
   ) return BOOLEAN is
      GT_ENTRY, LT_EXIT : BOOLEAN;
   begin
      if INTERVAL.ENTRY_OK then
         GT_ENTRY := (INTERVAL.ENTRY_TIME < TIME);
      else
         GT_ENTRY := TRUE;
      end if;

      if INTERVAL.EXIT_OK then
         LT_EXIT := (TIME < INTERVAL.EXIT_TIME);
      else
         LT_EXIT := TRUE;
      end if;

      if INTERVAL.ENTRY_OK and INTERVAL.EXIT_OK then
         if INTERVAL.EXIT_TIME < INTERVAL.ENTRY_TIME then
            return GT_ENTRY or LT_EXIT;
         end if;
      end if;

      return GT_ENTRY and LT_EXIT and (INTERVAL.ENTRY_OK or INTERVAL.EXIT_OK);
   end TIME_INSIDE_INTERVAL;


   procedure CHANGE_STATE(   -- local
   --called by: TASK_TC, TASK_STATE
   --parameter TC_LEN counts existing param. words
   --starting from word (MODE/PARAM_1)
      TC_LEN:    in INTEGER;
      MODE:      in CCT.T_UNSIGNED_8;
      PARAM_1:   in CCT.T_UNSIGNED_8  := 0;
      PARAM_2:   in CCT.T_UNSIGNED_16 := 0;
      PARAM_3:   in CCT.T_UNSIGNED_16 := 0;
      PARAM_4:   in CCT.T_UNSIGNED_16 := 0;
      PARAM_5:   in CCT.T_UNSIGNED_16 := 0;
      OEM_RA6:   in BOOLEAN := TRUE
   ) is
      K_MARKER_FOR_DEFAULT  : constant CCT.T_UNSIGNED_16 := 16#FFFF#;
      K_MARKER_FOR_PREVIOUS : constant CCT.T_UNSIGNED_16 := 16#FF00#;
      DFEE_STATE : IASW_DFEE.T_DFEE_STATE;
      PREV_STATE : IASW_DFEE.T_DFEE_STATE;
      DFEE_PARAM : CCT.T_UNSIGNED_16;
      SYNC : BOOLEAN;
      PRIMARY_SCF, SECONDARY_SCF : IASW_MODES.T_SCIENCE_FORMAT;
      PRIMARY_GF_LIMIT, SECONDARY_GF_LIMIT : CCT.T_UNSIGNED_16;
      TASK_ID : CIT.T_IASW_TASK_ID;
   begin
      TASK_ID := DPE_CSSW.CURRENT_TASK;

      if TC_LEN < 1 then  -- to be sure that MODE/PARAM#1 is given
         raise IASW_MAIN.EX_LENGTH;
      end if;

      --preliminary assumptions
      DFEE_PARAM := 0;
      SYNC := FALSE;
      PRIMARY_SCF        := K_MARKER_FOR_DEFAULT;
      SECONDARY_SCF      := K_MARKER_FOR_DEFAULT;
      PRIMARY_GF_LIMIT   := K_MARKER_FOR_DEFAULT;
      SECONDARY_GF_LIMIT := K_MARKER_FOR_DEFAULT;

      case MODE is
         when K_MODE_SAFE =>
            DFEE_STATE := IASW_DFEE.K_STATE_SAFE;

         when K_MODE_MEMORY =>
            DFEE_STATE := IASW_DFEE.K_STATE_MEMORY;

         when K_MODE_SETUP =>
            DFEE_STATE := IASW_DFEE.K_STATE_SETUP;

         when K_MODE_DATA_TAKING =>
            DFEE_STATE := IASW_DFEE.K_STATE_DATA_TAKING;
            SYNC := TRUE;

            if TC_LEN >= 2 then
               PRIMARY_SCF := PARAM_2;
            end if;
            if TC_LEN >= 3 then
               SECONDARY_SCF := PARAM_3;
            end if;
            if TC_LEN >= 4 then
               PRIMARY_GF_LIMIT := PARAM_4;
            end if;
            if TC_LEN >= 5 then
               SECONDARY_GF_LIMIT := PARAM_5;
            end if;

         when K_MODE_CALIBRATION =>
            if TC_LEN < 2 then
               raise IASW_MAIN.EX_LENGTH;
            end if;

            DFEE_STATE := IASW_DFEE.K_STATE_CALIBRATION;
            DFEE_PARAM := PARAM_2;
            SYNC := TRUE;

         when K_MODE_DIAGNOSTIC =>
            if TC_LEN < 2 then
               raise IASW_MAIN.EX_LENGTH;
            end if;

            DFEE_STATE := IASW_DFEE.K_STATE_DIAGNOSTIC;
            DFEE_PARAM := PARAM_2;
            SYNC := TRUE;

         when K_MODE_DATA_TAKING_TEST =>
            DFEE_STATE := IASW_DFEE.K_STATE_DATA_TAKING;

         when K_MODE_CALIBRATION_TEST =>
            if TC_LEN < 2 then
               raise IASW_MAIN.EX_LENGTH;
            end if;

            DFEE_STATE := IASW_DFEE.K_STATE_CALIBRATION;
            DFEE_PARAM := PARAM_2;

         when K_MODE_DIAGNOSTIC_TEST =>
            if TC_LEN < 2 then
               raise IASW_MAIN.EX_LENGTH;
            end if;

            DFEE_STATE := IASW_DFEE.K_STATE_DIAGNOSTIC;
            DFEE_PARAM := PARAM_2;

         when others =>  -- Invalid state
            raise IASW_MAIN.EX_DATA;
      end case;

      -- since ver.1.50 this procedure takes full control
      -- over interpretation of values:
      --   FFFF - return to default (the same for lack of param)
      --   FF00 - return to the last used value

      -- change values FFFF to the defaults
      if PRIMARY_SCF = K_MARKER_FOR_DEFAULT then
         PRIMARY_SCF := IASW_MODES.GET_DEFAULT_SCF(DFEE_STATE, SYNC);
      end if;
      if SECONDARY_SCF = K_MARKER_FOR_DEFAULT then
         SECONDARY_SCF := IASW_MODES.GET_DEFAULT_SCF(DFEE_STATE, SYNC);
      end if;
      if PRIMARY_GF_LIMIT = K_MARKER_FOR_DEFAULT then
         PRIMARY_GF_LIMIT := IASW_HSL.DEFAULT_PRIMARY_MIN_GF;
      end if;
      if SECONDARY_GF_LIMIT = K_MARKER_FOR_DEFAULT then
         SECONDARY_GF_LIMIT := IASW_HSL.DEFAULT_SECONDARY_MAX_GF;
      end if;

      -- change values FF00 to the previously set values
      if PRIMARY_SCF = K_MARKER_FOR_PREVIOUS then
         PRIMARY_SCF := SCF1;
      end if;
      if SECONDARY_SCF = K_MARKER_FOR_PREVIOUS then
         SECONDARY_SCF := SCF2;
      end if;
      if PRIMARY_GF_LIMIT = K_MARKER_FOR_PREVIOUS then
         PRIMARY_GF_LIMIT := GFL1;
      end if;
      if SECONDARY_GF_LIMIT = K_MARKER_FOR_PREVIOUS then
         SECONDARY_GF_LIMIT := GFL2;
      end if;

      -- check validity of formats
      if not IASW_MODES.IS_SCF_VALID(PRIMARY_SCF, DFEE_STATE, SYNC) then
         raise IASW_MAIN.EX_DATA;
      end if;
      if not IASW_MODES.IS_SCF_VALID(SECONDARY_SCF, DFEE_STATE, SYNC) then
         raise IASW_MAIN.EX_DATA;
      end if;

      -- check validity of grey filter limits
      if PRIMARY_GF_LIMIT not in IASW_GF.T_GREY_FILTER_INDICATOR then
         raise IASW_MAIN.EX_DATA;
      end if;
      if SECONDARY_GF_LIMIT not in IASW_GF.T_GREY_FILTER_INDICATOR then
         raise IASW_MAIN.EX_DATA;
      end if;

      if (MODE = K_MODE_DATA_TAKING) and
         (TASK_ID /= IASW_COMMON.TASK_STATE_ID)
      then
         -- this is TC from operator, save parameters:
         SCF1 := PRIMARY_SCF;
         SCF2 := SECONDARY_SCF;
         GFL1 := PRIMARY_GF_LIMIT;
         GFL2 := SECONDARY_GF_LIMIT;
      end if;

      PREV_STATE := IASW_DFEE.GET_STATE;

      if DFEE_STATE > IASW_DFEE.K_STATE_SETUP then
         IASW_MODES.START_DATA_TAKING(
            STATE => DFEE_STATE,
            SYNC  => SYNC,
            SF1   => PRIMARY_SCF,
            SF2   => SECONDARY_SCF,
            GFL1  => PRIMARY_GF_LIMIT,
            GFL2  => SECONDARY_GF_LIMIT
         );
      elsif PREV_STATE > IASW_DFEE.K_STATE_SETUP then
         IASW_MODES.STOP_DATA_TAKING;
      end if;

      IASW_DFEE.SET_STATE(DFEE_STATE, DFEE_PARAM, OEM_RA6);

      if DFEE_STATE = IASW_DFEE.K_STATE_DATA_TAKING then
         IASW_GF.SET_GF;  -- sets the previously used or 100% depending on the flag
      end if;

      if TASK_ID /= IASW_COMMON.TASK_STATE_ID then
         -- this is command from ground - clear the state as if there were
         -- no shutdowns before.  operator takes responsibility for control.
         ACTUAL_SHUTDOWN_LEVEL := 0;
         MINIMAL_SHUTDOWN_LEVEL := 0;

         --added in ver.1.62
         INTERRUPTED_DFEE_STATE := IASW_DFEE.K_STATE_STANDBY;

      end if;
   end CHANGE_STATE;


   procedure TC55(TC_DATA: in CIT.T_TC_APPL_DATA_FIELD) is
   --called by: TASK_TC

      TC55_DATA : T_TC55_DATA;
         for TC55_DATA use at TC_DATA.WORDS'address;
pragma If(((TC55_DATA'size+15)/16) > TC_DATA.WORDS'length);
   error: Violated safety of 'use at'
pragma End if;

   begin
      CHANGE_STATE(
         TC_DATA.LEN,
         TC55_DATA.MODE,
         TC55_DATA.PARAM_1,
         TC55_DATA.PARAM_2,
         TC55_DATA.PARAM_3,
         TC55_DATA.PARAM_4,
         TC55_DATA.PARAM_5 );
   end TC55;


   procedure SHUTDOWN(   -- local
   --called by: TASK_STATE
      ACTUAL_LEVEL: in out T_SHUTDOWN_LEVEL;
      WANTED_LEVEL: in T_SHUTDOWN_LEVEL
   ) is
   begin
      -- any level --> level #70 or #80
      if WANTED_LEVEL >= 16#70# then

         IASW_MODES.STOP_DATA_TAKING;
         ACTUAL_LEVEL := WANTED_LEVEL;
         -- level#70 means DFEE off, level#80 means LSL dead
         -- there is no sense to communicate with DFEE.
         return;
      end if;

      -- level #0 --> level #10
      if (WANTED_LEVEL >= 16#10#) and (ACTUAL_LEVEL < 16#10#) then

         -- store DFEE State interrupted by a shutdown
         -- this will indicate the target state when recovering
         INTERRUPTED_DFEE_STATE := IASW_DFEE.GET_STATE;
         if INTERRUPTED_DFEE_STATE > IASW_DFEE.K_STATE_DATA_TAKING then
            -- Calibration nor Diag.Dump will not be resumed automatically
            INTERRUPTED_DFEE_STATE := IASW_DFEE.K_STATE_SAFE;
         end if;

         -- store settings of HV
         IASW_DFEE.GET_LAST_COMMANDED_HW_SETTINGS(DFEE_HV.SETTINGS);

         -- switch off data taking
         if IASW_DFEE.GET_STATE > IASW_DFEE.K_STATE_SETUP then
            CHANGE_STATE(1, K_MODE_SETUP);
         end if;

         ACTUAL_LEVEL := 16#10#;
      end if;

      -- level #10 --> level #20
      if (WANTED_LEVEL >= 16#20#) and (ACTUAL_LEVEL < 16#20#) then
         ACTUAL_LEVEL := 16#20#;
      end if;

      -- level #20 --> level #30
      if (WANTED_LEVEL >= 16#30#) and (ACTUAL_LEVEL < 16#30#) then

         if IASW_DFEE.GET_STATE >= IASW_DFEE.K_STATE_SETUP then
            NEED_HV_ON := IASW_DFEE.HV_IS_ON and
               (INTERRUPTED_DFEE_STATE = IASW_DFEE.K_STATE_DATA_TAKING);
            --goto SAFE
            CHANGE_STATE(1, K_MODE_SAFE);
         else
            NEED_HV_ON := FALSE;
         end if;

         ACTUAL_LEVEL := 16#30#;
      end if;

      -- level #30 --> level #40
      if (WANTED_LEVEL >= 16#40#) and (ACTUAL_LEVEL < 16#40#) then

         --save DFEE context + goto STANDBY
         if IASW_DFEE.GET_STATE > IASW_DFEE.K_STATE_STANDBY then
            if WANTED_LEVEL = 16#40# then
               -- there is a chance that DFEE lives.
               if AUTO_RECOVERY_CONTROL_INFO.RECOVERY_LEVEL >= K_RECOVER_CONTEXT then
                  -- the RECOVERY_LEVEL is tested here
                  -- to avoid saving context if no recovery is provided,
                  -- When the flag is off, it is likely not
                  -- desired to change context stored manually by operator.
                  IASW_DFEE.SAVE_CONTEXT;
               end if;
               -- make sure to leave DFEE in SAFE
               if IASW_DFEE.GET_STATE > IASW_DFEE.K_STATE_SAFE then
                  CHANGE_STATE(1, K_MODE_SAFE);
               end if;
            end if;
         end if;

         ACTUAL_LEVEL := 16#40#;
      end if;

      -- level #40 --> level #50
      if (WANTED_LEVEL >= 16#50#) and (ACTUAL_LEVEL < 16#50#) then
         ACTUAL_LEVEL := 16#50#;
      end if;

      -- level #50 --> level #60
      if WANTED_LEVEL >= 16#60# then
         ACTUAL_LEVEL := 16#60#;
      end if;
   end SHUTDOWN;


   procedure RESTORE(   --local
   --called by: TASK_STATE
      ACTUAL_LEVEL : in out T_SHUTDOWN_LEVEL;
      WANTED_LEVEL : in T_SHUTDOWN_LEVEL
   ) is
   begin

      -- level #50 --> level #40
      if 16#40# < MINIMAL_SHUTDOWN_LEVEL then
         return;
      end if;
      if (WANTED_LEVEL <= 16#40#) and (ACTUAL_LEVEL > 16#40#) then
         ACTUAL_LEVEL := 16#40#;
      end if;

      -- level #40 --> level #30
      if 16#30# < MINIMAL_SHUTDOWN_LEVEL then
         return;
      end if;
      if (WANTED_LEVEL <= 16#30#) and (ACTUAL_LEVEL > 16#30#) then

         if INTERRUPTED_DFEE_STATE >= IASW_DFEE.K_STATE_SAFE then
            if IASW_DFEE.GET_STATE < IASW_DFEE.K_STATE_SAFE then
               -- wait for DFEE heeting and at least one response with HK
               return;
            else
               -- context
               if AUTO_RECOVERY_CONTROL_INFO.RECOVERY_LEVEL >= K_RECOVER_CONTEXT then
                  IASW_DFEE.LOAD_CONTEXT;
               end if;
               -- CPU mode
               if AUTO_RECOVERY_CONTROL_INFO.CPU_MODE /= 0 then
                  CHANGE_STATE(1, K_MODE_SETUP);
                  IASW_DFEE.CHG_CPU_SETTING(CCT.T_UNSIGNED_16(
                     AUTO_RECOVERY_CONTROL_INFO.CPU_MODE));
                  if INTERRUPTED_DFEE_STATE < IASW_DFEE.K_STATE_SETUP then
                     CHANGE_STATE(1, K_MODE_SAFE);
                  end if;
               end if;
               -- make sure to leave DFEE in SAFE
               if IASW_DFEE.GET_STATE > IASW_DFEE.K_STATE_SAFE then
                  CHANGE_STATE(1, K_MODE_SAFE);
               end if;
            end if;
         end if;

         ACTUAL_LEVEL := 16#30#;
      end if;

      -- level #30 --> level #20
      if 16#20# < MINIMAL_SHUTDOWN_LEVEL then
         return;
      end if;
      if (WANTED_LEVEL <= 16#20#) and (ACTUAL_LEVEL > 16#20#) then

         if INTERRUPTED_DFEE_STATE >= IASW_DFEE.K_STATE_SETUP then
            begin
               if ACTUAL_LEVEL > 16#2E# then
                  if IASW_DFEE.GET_STATE < IASW_DFEE.K_STATE_SETUP then
                     CHANGE_STATE(1, K_MODE_SETUP);
                  end if;
                  ACTUAL_LEVEL := 16#2E#;
               end if;

               if NEED_HV_ON and
                  (AUTO_RECOVERY_CONTROL_INFO.RECOVERY_LEVEL >= K_RECOVER_HV)
               then
                  if ACTUAL_LEVEL > 16#2C# then
                     IASW_DFEE.HV_READY;  --this goes smoothly
                     ACTUAL_LEVEL := 16#2C#;
                  end if;
                  if ACTUAL_LEVEL > 16#2A# then
                     IASW_DFEE.HV_ON;     --this goes smoothly
                     ACTUAL_LEVEL := 16#2A#;
                  end if;
                  if ACTUAL_LEVEL > 16#28# then
                     IASW_DFEE.SET_HV_DRIFT(  --this may answer "BUSY"
                        PARAM   => IASW_DFEE.REDUCED_HV(
                           HV    => DFEE_HV.SETTINGS.HV_DRIFT_SET,
                           HVMIN => IASW_DFEE.MIN_DRIFT_HV,
                           HVDELTA => IASW_DFEE.DELTA_DRIFT_HV),
                        OEM_RA6 => FALSE );
                     ACTUAL_LEVEL := 16#28#;
                  end if;
                  if ACTUAL_LEVEL > 16#26# then
                     IASW_DFEE.SET_HV_ACC(  --this may answer "BUSY"
                        PARAM   => IASW_DFEE.REDUCED_HV(
                           HV    => DFEE_HV.SETTINGS.HV_ACC_SET,
                           HVMIN => IASW_DFEE.MIN_ACC_HV,
                           HVDELTA => IASW_DFEE.DELTA_ACC_HV),
                        OEM_RA6 => FALSE );
                     ACTUAL_LEVEL := 16#26#;
                  end if;
               end if;
            exception
         when IASW_DFEE.EX_DFEE_BUSY =>
               return;   --action will be repeated soon
            end;
         end if;

         ACTUAL_LEVEL := 16#20#;
      end if;

      -- level #20 --> level #10
      if 16#10# < MINIMAL_SHUTDOWN_LEVEL then
         return;
      end if;
      if (WANTED_LEVEL <= 16#10#) and (ACTUAL_LEVEL > 16#10#) then
         ACTUAL_LEVEL := 16#10#;
      end if;

      -- level #10 --> level #0
      if 0 < MINIMAL_SHUTDOWN_LEVEL then
         return;
      end if;
      if (WANTED_LEVEL = 0) and (ACTUAL_LEVEL > 0) then

         if (INTERRUPTED_DFEE_STATE = IASW_DFEE.K_STATE_DATA_TAKING) and
            (AUTO_RECOVERY_CONTROL_INFO.RECOVERY_LEVEL >= K_RECOVER_DATA_TAKING)
         then
            begin
               CHANGE_STATE( 5, K_MODE_DATA_TAKING, 0, SCF1, SCF2, GFL1, GFL2, FALSE);
               -- this may answer "BUSY"

               NEED_HV_ON := FALSE;
            exception
         when IASW_DFEE.EX_DFEE_BUSY =>
               return;   --action will be repeated soon
            end;
         end if;

         ACTUAL_LEVEL := 0;
      end if;
   end RESTORE;


   function START_OF_A_NEW_POINTING return BOOLEAN is   --local
   use IASW_COMMON;  --to be able to compare IASW_COMMON.T_PID-s
      OTF_ACTIVE_EDGE, OTF_ACTIVE_STATE : BOOLEAN;
      PID_ACTIVE_EDGE, PID_ACTIVE_STATE : BOOLEAN;
      PID1V, PID_NOW : IASW_COMMON.T_PID;
   begin
      if not BCPKT_VALID then
      -- no valid BCPKT to check
         return FALSE;
      end if;

      OTF_ACTIVE_STATE := (BCPKT.OTF /= 0) and (BCPKT.AOCS_MODE = 5);
      OTF_ACTIVE_EDGE := OTF_ACTIVE_STATE and not PREV_OTF;
      PREV_OTF := OTF_ACTIVE_STATE;
      PID_NOW := BCPKT.PID;
      PID_ACTIVE_STATE := (PID_NOW /= (0,0));
      PID_ACTIVE_EDGE := PID_ACTIVE_STATE and (PID_NOW /= PREV_PID);
      PREV_PID := PID_NOW;

      if (OTF_ACTIVE_EDGE and PID_ACTIVE_STATE) or  -- standard condition
         (PID_ACTIVE_EDGE and OTF_ACTIVE_STATE)     -- spare condition
      then
         PID1V := PREV_VALID_PID;
         PREV_VALID_PID := PID_NOW;
         return (PID_NOW /= PID1V) and not BCPKT_FILTER.OTF;
      end if;

      return FALSE;
   end START_OF_A_NEW_POINTING;


   procedure RESTART_DATA_TAKING is   --local
   --this is to clear big buffer before new observation.
   begin
      IASW_TM.PUT_OEM(
         IASW_TM.EV_AUTO_RESTART,
         IASW_COMMON.CVT(BCPKT.PID.REVOLUTION),
         IASW_COMMON.CVT(BCPKT.PID.POINTING));
      CHANGE_STATE(1, K_MODE_SETUP);
      DPE_CSSW.DELAY_RELATIVE(16);
      -- the delay (since 1.60) is introduced to give task HSL chance to
      -- read dummy events.
      CHANGE_STATE(5, K_MODE_DATA_TAKING, 0, SCF1, SCF2, GFL1, GFL2);
   exception
      when others =>
         IASW_TM.PUT_OEM(IASW_TM.EV_IASW_EXCEPTION, DPE_CSSW.CURRENT_TASK);
   end RESTART_DATA_TAKING;


   procedure TASK_STATE is
      SUSP : BOOLEAN;
   begin
      IASW_DFEE.CONNECT_TO_LSL;
      loop
         SUSP := FALSE;   -- preliminary

         ENTER_STATE_CRITSEC;

         begin
            IASW_COMMON.I_AM_ALIVE(IASW_COMMON.TASK_STATE_NO);

            if START_OF_A_NEW_POINTING then
               if (DEEPER_SHUTDOWN_LEVEL = 0) and
                  (IASW_DFEE.GET_STATE = IASW_DFEE.K_STATE_DATA_TAKING)
               then
                  RESTART_DATA_TAKING;
               end if;
            end if;

            if (WANTED_SHUTDOWN_LEVEL < ACTUAL_SHUTDOWN_LEVEL) and
               (MINIMAL_SHUTDOWN_LEVEL < ACTUAL_SHUTDOWN_LEVEL) and
               (AUTO_RECOVERY_CONTROL_INFO.RECOVERY_LEVEL > K_RECOVER_NONE) and
               (DEEPER_SHUTDOWN_LEVEL < 16#60#)
            then

               if (not REC_OEM_CONTROL.OEM_SENT) or
                  (REC_OEM_CONTROL.ASL /= ACTUAL_SHUTDOWN_LEVEL) or
                  (REC_OEM_CONTROL.WSL /= WANTED_SHUTDOWN_LEVEL)
               then
                  IASW_TM.PUT_OEM(
                     IASW_TM.EV_AUTO_RECOVERY,
                     IASW_COMMON.CVT(ACTUAL_SHUTDOWN_LEVEL),
                     IASW_COMMON.CVT(WANTED_SHUTDOWN_LEVEL));
                  REC_OEM_CONTROL.OEM_SENT := TRUE;
                  REC_OEM_CONTROL.ASL := ACTUAL_SHUTDOWN_LEVEL;
                  REC_OEM_CONTROL.WSL := WANTED_SHUTDOWN_LEVEL;
               end if;

               begin
                  RESTORE(ACTUAL_SHUTDOWN_LEVEL, WANTED_SHUTDOWN_LEVEL);
               exception
                  when others =>
                     IASW_TM.PUT_OEM(
                        IASW_TM.EV_AUTO_RCV_FAILED,
                        IASW_COMMON.CVT(ACTUAL_SHUTDOWN_LEVEL),
                        IASW_COMMON.CVT(WANTED_SHUTDOWN_LEVEL));
                     MINIMAL_SHUTDOWN_LEVEL := ACTUAL_SHUTDOWN_LEVEL;
                     -- there will be no more tries to recovere more than this,
                     -- until sh. levels will be cleared by operator state command.
               end;

            elsif WANTED_SHUTDOWN_LEVEL > ACTUAL_SHUTDOWN_LEVEL then

               --allow for uncoditional OEM(recovery) in the future
               REC_OEM_CONTROL.OEM_SENT := FALSE;

               IASW_TM.PUT_OEM(
                  IASW_TM.EV_AUTO_SHUTDOWN,
                  IASW_COMMON.CVT(ACTUAL_SHUTDOWN_LEVEL),
                  IASW_COMMON.CVT(WANTED_SHUTDOWN_LEVEL));
               begin
                  SHUTDOWN(ACTUAL_SHUTDOWN_LEVEL, WANTED_SHUTDOWN_LEVEL);
               exception
                  when others =>
                     IASW_TM.PUT_OEM(
                        IASW_TM.EV_AUTO_SHD_FAILED,
                        IASW_COMMON.CVT(ACTUAL_SHUTDOWN_LEVEL),
                        IASW_COMMON.CVT(WANTED_SHUTDOWN_LEVEL));
               end;

            else
               SUSP := TRUE;
            end if;

         exception
            when others =>
               IASW_TM.PUT_OEM(IASW_TM.EV_IASW_EXCEPTION, DPE_CSSW.CURRENT_TASK);
         end;

         LEAVE_STATE_CRITSEC;

         if SUSP then
            DPE_CSSW.TERMINATE_TASK;
         else
            DPE_CSSW.DELAY_RELATIVE(8);
         end if;

      end loop;
   end TASK_STATE;


   procedure ANALYSE_LOW_VOLTAGE( MRTU_HK : in IASW_COMMON.T_MRTU_HK ) is
      FAILURE : BOOLEAN := FALSE;
      V : CCT.T_UNSIGNED_16;
   begin
      V := MRTU_HK(IASW_COMMON.VOLTAGE_P5V_DIG);
      if HK_FILTER.VOLTAGE_P5V_DIG then
         FAILURE := FAILURE or HK_CONDITION.VOLTAGE_P5V_DIG;
      elsif V <= 16#FF# then
         FAILURE := FAILURE or V not in K_DIG_P5_V_MIN..K_DIG_P5_V_MAX;
      end if;

      V := MRTU_HK(IASW_COMMON.CURRENT_P5V_DIG);
      if HK_FILTER.CURRENT_P5V_DIG then
         FAILURE := FAILURE or HK_CONDITION.CURRENT_P5V_DIG;
      elsif V <= 16#FF# then
         FAILURE := FAILURE or V not in K_DIG_P5_A_MIN..K_DIG_P5_A_MAX;
      end if;

      V := MRTU_HK(IASW_COMMON.VOLTAGE_P5V_ANA);
      if HK_FILTER.VOLTAGE_P5V_ANA then
         FAILURE := FAILURE or HK_CONDITION.VOLTAGE_P5V_ANA;
      elsif V <= 16#FF# then
         FAILURE := FAILURE or V not in K_ANA_P5_V_MIN..K_ANA_P5_V_MAX;
      end if;

      V := MRTU_HK(IASW_COMMON.VOLTAGE_M5V_ANA);
      if HK_FILTER.VOLTAGE_M5V_ANA then
         FAILURE := FAILURE or HK_CONDITION.VOLTAGE_M5V_ANA;
      elsif V <= 16#FF# then
         FAILURE := FAILURE or V not in K_ANA_M5_V_MIN..K_ANA_M5_V_MAX;
      end if;

      V := MRTU_HK(IASW_COMMON.VOLTAGE_P12V);
      if HK_FILTER.VOLTAGE_P12V then
         FAILURE := FAILURE or HK_CONDITION.VOLTAGE_P12V;
      elsif V <= 16#FF# then
         FAILURE := FAILURE or V not in K_P12V_V_MIN..K_P12V_V_MAX;
      end if;

      V := MRTU_HK(IASW_COMMON.CURRENT_P12V);
      if HK_FILTER.CURRENT_P12V then
         FAILURE := FAILURE or HK_CONDITION.CURRENT_P12V;
      elsif V <= 16#FF# then
         FAILURE := FAILURE or V not in K_P12V_A_MIN..K_P12V_A_MAX;
      end if;

      V := MRTU_HK(IASW_COMMON.VOLTAGE_M12V);
      if HK_FILTER.VOLTAGE_M12V then
         FAILURE := FAILURE or HK_CONDITION.VOLTAGE_M12V;
      elsif V <= 16#FF# then
         FAILURE := FAILURE or V not in K_M12V_V_MIN..K_M12V_V_MAX;
      end if;

      V := MRTU_HK(IASW_COMMON.CURRENT_M12V);
      if HK_FILTER.CURRENT_M12V then
         FAILURE := FAILURE or HK_CONDITION.CURRENT_M12V;
      elsif V <= 16#FF# then
         FAILURE := FAILURE or V not in K_M12V_A_MIN..K_M12V_A_MAX;
      end if;

      if FAILURE then
         IASW_DFEE.LOW_VOLTAGE_IS_OFF;
      else
         IASW_DFEE.LOW_VOLTAGE_IS_ON;
      end if;

      LOW_VOLTAGE_FAILURE := FAILURE;
   end ANALYSE_LOW_VOLTAGE;


   procedure ANALYSE_PRESSURE( MRTU_HK : in IASW_COMMON.T_MRTU_HK ) is
      FAILURE : BOOLEAN := FALSE;
      V : CCT.T_UNSIGNED_16;
   begin
      V := MRTU_HK(IASW_COMMON.DET_PRESSURE_1);
      if HK_FILTER.DET_PRESSURE_1 then
         FAILURE := FAILURE or HK_CONDITION.DET_PRESSURE_1;
      elsif V <= 16#FF# then
         FAILURE := FAILURE or (V < DET_PRESS_1_MIN);
      end if;

      V := MRTU_HK(IASW_COMMON.DET_PRESSURE_2);
      if HK_FILTER.DET_PRESSURE_2 then
          FAILURE := FAILURE or HK_CONDITION.DET_PRESSURE_2;
      elsif V <= 16#FF# then
         FAILURE := FAILURE or (V < DET_PRESS_2_MIN);
      end if;

      PRESSURE_FAILURE := FAILURE;
   end ANALYSE_PRESSURE;


   procedure ANALYSE_TEMPERATURE( MRTU_HK : in IASW_COMMON.T_MRTU_HK ) is
      FAILURE : BOOLEAN := FALSE;
      V : CCT.T_UNSIGNED_16;
   begin
      V := MRTU_HK(IASW_COMMON.DET_TEMP_1);
      if HK_FILTER.DET_TEMP_1 then
         FAILURE := FAILURE or HK_CONDITION.DET_TEMP_1;
      elsif V <= 16#FF# then
         FAILURE := FAILURE or (V > DET_TEMP_1_MAX);
      end if;

      V := MRTU_HK(IASW_COMMON.DET_TEMP_2);
      if HK_FILTER.DET_TEMP_2 then
          FAILURE := FAILURE or HK_CONDITION.DET_TEMP_2;
      elsif V <= 16#FF# then
         FAILURE := FAILURE or (V > DET_TEMP_2_MAX);
      end if;

      TEMP_FAILURE := FAILURE;
   end ANALYSE_TEMPERATURE;


   procedure ANALYSE_MRTU_DATA( MRTU_HK : in IASW_COMMON.T_MRTU_HK ) is
   begin
      --DFEE_LIVED_AT_LEAST_ONCE := TRUE;
      --in ver.1.63 moved to IASW_DFEE

      ANALYSE_PRESSURE   (MRTU_HK);
      ANALYSE_TEMPERATURE(MRTU_HK);
      ANALYSE_LOW_VOLTAGE(MRTU_HK);
   end ANALYSE_MRTU_DATA;


   function SWITCH_OFF_IMM return BOOLEAN is   --local
   begin
      if BCPKT_FILTER.SWITCH_OFF_IMM then
         return BCPKT_CONDITION.SWITCH_OFF_IMM;
      end if;

      return BCPKT_VALID and (BCPKT.SWITCH_OFF_IMM /= 0);
   end SWITCH_OFF_IMM;


   function ESAM return BOOLEAN is   --local
   begin
      if BCPKT_FILTER.ESAM then
         return BCPKT_CONDITION.ESAM;
      end if;

      return BCPKT_VALID and (BCPKT.ESAM /= 0);
   end ESAM;


   function ECLIPSE return BOOLEAN is   --local
   begin
      if BCPKT_FILTER.ECLIPSE then
         return BCPKT_CONDITION.ECLIPSE;
      end if;

      return TIME_INSIDE_INTERVAL(
         INTERVAL => ECLIPSE_INTERVAL,
         TIME     => DPE_CSSW.GET_CURRENT_OBT);
   end ECLIPSE;


   function RADIATION_BELTS return BOOLEAN is   --local
   begin
      if not (BCPKT_VALID and (BCPKT.DRMC /= 0)) then
         -- the main sens of the above condition is
         -- "if radiation counters are valid then"
         -- it may be confusing that it is allowed to enter this fragment
         -- with BCPKT_VALID=FALSE.  everything is ok.  in such case
         -- the RADMON(x).IS_HIGH are all FALSE.
         if RADMON_HIGH( RADMON(1),
               BCPKT_FILTER.RAD_MONITOR_1,
               BCPKT_CONDITION.RAD_MONITOR_1) or
            RADMON_HIGH( RADMON(2),
               BCPKT_FILTER.RAD_MONITOR_2,
               BCPKT_CONDITION.RAD_MONITOR_2) or
            RADMON_HIGH( RADMON(3),
               BCPKT_FILTER.RAD_MONITOR_3,
               BCPKT_CONDITION.RAD_MONITOR_3)
         then
            return TRUE;
         end if;
      end if;

      if BCPKT_FILTER.RAD_BELTS then
         return BCPKT_CONDITION.RAD_BELTS;
      end if;

      return TIME_INSIDE_INTERVAL(
         INTERVAL => RADIATION_BELTS_INTERVAL,
         TIME     => DPE_CSSW.GET_CURRENT_OBT);
   end RADIATION_BELTS;


   function TM_SHARE return CCT.T_UNSIGNED_16 is   --local
      TMS : CCT.T_UNSIGNED_8;
   begin
      if BCPKT_FILTER.TM_SHARE then
         if BCPKT_CONDITION.TM_SHARE then
            TMS := 16;
         else
            TMS := 8;
         end if;

      elsif BCPKT_VALID then
         case CIT.K_JEMX is
            when CIT.K_JEMX_1 =>
               TMS := BCPKT.TM_SHARE_JEMX1;
            when CIT.K_JEMX_2 =>
               TMS := BCPKT.TM_SHARE_JEMX2;
            when others =>
               TMS := 0;
         end case;

         if TMS >= 1 then
            TMS := TMS - 1;   -- -1 for housekeeping TM package
         end if;
         if TMS > 16 then         -- this must be an error
            TMS := 16;
         end if;

      else  --no BCPKT available, set default
         TMS := 8;
      end if;

      return CCT.T_UNSIGNED_16(TMS);
   end TM_SHARE;


   procedure ANALYSE_STATE is   --local
   --it is to determine the WANTED_SHUTDOWN_LEVEL in function of
   --external conditions (analog readouts, broadcasted info).
   --called by: TASK_MAIN, TASK_HK, TASK_TC
   begin

      if LOW_VOLTAGE_FAILURE then
         if DEEPER_SHUTDOWN_LEVEL in 16#40# .. 16#50# then
            --normal switch off after eclipse warning
            WANTED_SHUTDOWN_LEVEL := 16#50#;
         elsif IASW_DFEE.DFEE_LIVED_AT_LEAST_ONCE then
            --unexpected failure of low voltage
            WANTED_SHUTDOWN_LEVEL := 16#70#;
         else
            --DFEE was never On
            WANTED_SHUTDOWN_LEVEL := 16#50#;
         end if;

      elsif IASW_DFEE.LSL_DEAD then
         WANTED_SHUTDOWN_LEVEL := 16#80#;

      elsif ESAM or SWITCH_OFF_IMM or PRESSURE_FAILURE or TEMP_FAILURE then
         WANTED_SHUTDOWN_LEVEL := 16#60#;

      elsif ECLIPSE then
         WANTED_SHUTDOWN_LEVEL := 16#40#;

      elsif RADIATION_BELTS then
         WANTED_SHUTDOWN_LEVEL := 16#30#;

      else
         WANTED_SHUTDOWN_LEVEL := 0;
      end if;

      IASW_COMMON.START_TASK(IASW_COMMON.TASK_STATE_ID);
   end ANALYSE_STATE;


   procedure SET_WANTED_SHUTDOWN_LEVEL is
   --this is only an envelope for ANALYSE_STATE, making that the least
   --is executed inside the critical section.
   --called by: TASK_MAIN, TASK_HK
   begin
      if not CAN_ENTER_STATE_CRITSEC then
         PENDING_ANALYSE := TRUE;
         return;
      end if;

      begin
         PENDING_ANALYSE := FALSE;
         ANALYSE_STATE;
      exception
         when others =>
            null;
      end;

      LEAVE_STATE_CRITSEC;
   end SET_WANTED_SHUTDOWN_LEVEL;


   function HAVE_PENDING_ANALYSE return BOOLEAN is
   begin
      return PENDING_ANALYSE;
   end HAVE_PENDING_ANALYSE;


   procedure UPDATE_BCPKT(NEW_BCPKT : in IASW_COMMON.T_BROADCAST_PACKET) is
   --called by: TASK_TC
   --this is called from within the critical region (STATE_SEMAPHORE)
   begin
      BCPKT := NEW_BCPKT;
      BCPKT_VALID := TRUE;

      -- radiation belts entry and exit times:
      INTERVAL_SET( RADIATION_BELTS_INTERVAL,
         BCPKT.NEXT_BELT_ENTRY,
         BCPKT.NEXT_BELT_EXIT);

      -- eclipse entry  and exit times:
      INTERVAL_SET( ECLIPSE_INTERVAL,
         BCPKT.NEXT_ECLIPSE_ENTRY,
         BCPKT.NEXT_ECLIPSE_EXIT);

      if BCPKT.DRMC = 0 then   --don't Disregard Rad.Mon.Counters
         RADMON_NEW_VALUE(RADMON(1), BCPKT.RAD_MONITOR_1);
         RADMON_NEW_VALUE(RADMON(2), BCPKT.RAD_MONITOR_2);
         RADMON_NEW_VALUE(RADMON(3), BCPKT.RAD_MONITOR_3);
      end if;

      IASW_GF.SET_TM_BLOCKS_PER_BCP1(TM_SHARE);

      ANALYSE_STATE;
   end UPDATE_BCPKT;


   procedure GET_LAST_BCPKT(
   --called by: TASK_MAIN
      LAST_BCPKT : out IASW_COMMON.T_BROADCAST_PACKET;
      VALID : out BOOLEAN
   ) is
   begin
      if BCPKT_VALID then
         LAST_BCPKT := BCPKT;
      end if;
      VALID := BCPKT_VALID;
   end GET_LAST_BCPKT;


   procedure BCP1_TICK is
   begin
      RADMON_BCP1_TICK(RADMON(1));
      RADMON_BCP1_TICK(RADMON(2));
      RADMON_BCP1_TICK(RADMON(3));
   end BCP1_TICK;


   procedure SET_SW_PARAM(
   --called by: TASK_TC
   --this is called from within the critical region (STATE_SEMAPHORE)
      PARAM_ID : in IASW_COMMON.T_FID;
      PARAM_VAL: in IASW_COMMON.T_TC_PARAM_VAL;
      PARAM_LEN: in IASW_COMMON.T_SW_PARAM_LEN
   ) is
   begin
      if PARAM_VAL'length < PARAM_LEN then
         raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
      end if;

      case PARAM_ID is
         when 12 =>  -- set flag "Automatic Recovery"
            if PARAM_LEN /= 1 then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            else
               AUTO_RECOVERY_CONTROL_INFO := CVT(PARAM_VAL(1));
            end if;

         when 14 =>  -- set thresholds for preassure checking
            if PARAM_LEN /= 2 then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            else
               DET_PRESS_1_MIN := PARAM_VAL(1);
               DET_PRESS_2_MIN := PARAM_VAL(2);
            end if;

         when 24 =>  -- set thresholds for temperature checking
            if PARAM_LEN /= 2 then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            else
               DET_TEMP_1_MAX := PARAM_VAL(1);
               DET_TEMP_2_MAX := PARAM_VAL(2);
            end if;

         when 13 =>  -- counters of occurencies high/low radiation monitors
            if PARAM_LEN /= 2 then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            else
               RADMON_ENTRY_CNT := PARAM_VAL(1);
               RADMON_EXIT_CNT  := PARAM_VAL(2);
            end if;

         when 17 =>  -- set tresholds for Radiation Monitor Count Rates
            if PARAM_LEN/=6 then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            else
               RADMON_INIT(RADMON(1), PARAM_VAL(1), PARAM_VAL(2));
               RADMON_INIT(RADMON(2), PARAM_VAL(3), PARAM_VAL(4));
               RADMON_INIT(RADMON(3), PARAM_VAL(5), PARAM_VAL(6));
            end if;

         when 18 =>  -- set filter for broadcast packet data
            if PARAM_LEN /= 2 then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            else
               BCPKT_FILTER := CVT(PARAM_VAL(1));
               BCPKT_CONDITION := CVT(PARAM_VAL(2));
            end if;

         when 19 =>  -- set filter for hk data
            if PARAM_LEN /= 2 then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            else
               HK_FILTER := CVT(PARAM_VAL(1));
               HK_CONDITION := CVT(PARAM_VAL(2));
            end if;

         -- deal with some extra testing commands

         when 251 =>     -- HV Off with use of relay

            IASW_DFEE.SEND_HV_OFF_RELAY_SIGNAL;
            IASW_DFEE.SET_STATE(IASW_DFEE.K_STATE_SAFE);  -- since ver. 1.49

         when 252 =>     -- Reset LSL module

            DPE_CSSW.RESET_SERIAL_LINE_MODULE;

         when 253 =>     -- Set HSL params

            if PARAM_LEN < 2 then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            end if;
            IASW_HSL.SET_HSL_PARAMS(
               WT_INTERVAL  => PARAM_VAL(1),
               MAX_BLK_SIZE => PARAM_VAL(2));

         when 254 =>     -- Set LSL params

            if PARAM_LEN < 2 then
               raise IASW_COMMON.EX_SW_PARAM_ERROR_LENGTH;
            end if;
            IASW_DFEE.SET_LSL_PARAMS(
               CLK_FREQ_ID   => PARAM_VAL(1),
               BDR_FACTOR_ID => PARAM_VAL(2));

         when others =>
            raise IASW_COMMON.EX_SW_PARAM_ERROR_ID;
      end case;
   end SET_SW_PARAM;


   procedure GET_SW_PARAM(
   --called by: TASK_REQ_OUT
      PARAM_ID : in  IASW_COMMON.T_FID;
      PARAM_VAL: out IASW_COMMON.T_TM_PARAM_VAL;
      PARAM_LEN: out IASW_COMMON.T_SW_PARAM_LEN
   ) is
   begin
      case PARAM_ID is
         when 12 =>  -- get flag "Automatic Recovery"
            PARAM_VAL(1) := CVT(AUTO_RECOVERY_CONTROL_INFO);
            PARAM_LEN := 1;

         when 14 =>  -- get thresholds for preassure checking
            PARAM_VAL(1) := DET_PRESS_1_MIN;
            PARAM_VAL(2) := DET_PRESS_2_MIN;
            PARAM_LEN := 2;

         when 24 =>  -- get thresholds for temperature checking
            PARAM_VAL(1) := DET_TEMP_1_MAX;
            PARAM_VAL(2) := DET_TEMP_2_MAX;
            PARAM_LEN := 2;

         when 13 =>  -- get counters of occurencies high/low radiation monitors
            PARAM_VAL(1) := RADMON_ENTRY_CNT;
            PARAM_VAL(2) := RADMON_EXIT_CNT;
            PARAM_LEN := 2;

         when 17 =>  -- get tresholds for Radiation Monitor Count Rates
            PARAM_VAL(1) := RADMON(1).HIGH_LIMIT;
            PARAM_VAL(2) := RADMON(1).LOW_LIMIT;
            PARAM_VAL(3) := RADMON(2).HIGH_LIMIT;
            PARAM_VAL(4) := RADMON(2).LOW_LIMIT;
            PARAM_VAL(5) := RADMON(3).HIGH_LIMIT;
            PARAM_VAL(6) := RADMON(3).LOW_LIMIT;
            PARAM_LEN := 6;

         when others =>
            raise IASW_COMMON.EX_SW_PARAM_ERROR_ID;
      end case;
   end GET_SW_PARAM;


   procedure GET_DPE_STATE(DS: out IASW_COMMON.T_DPE_STATE) is
   --called by: TASK_MAIN
   begin
      DS.SHUTDOWN_LEVEL(1) := CCT.T_UNSIGNED_8(ACTUAL_SHUTDOWN_LEVEL);
      DS.SHUTDOWN_LEVEL(2) := CCT.T_UNSIGNED_8(WANTED_SHUTDOWN_LEVEL);
      DS.AUT_RECV_INFO := CVT(AUTO_RECOVERY_CONTROL_INFO);
   end GET_DPE_STATE;


   procedure GET_HK_DIAG(
      FLAGS : out IASW_COMMON.T_16_BITS;
      DIAG  : out T_DIAGNOSTIC_HK
   ) is
   begin
      DIAG.BCPKT_FILTER    := CVT(BCPKT_FILTER);
      DIAG.BCPKT_CONDITION := CVT(BCPKT_CONDITION);
      DIAG.HK_FILTER       := CVT(HK_FILTER);
      DIAG.HK_CONDITION    := CVT(HK_CONDITION);
   end GET_HK_DIAG;


   function MAX_ALLOWED_STATE( LEVEL : in T_SHUTDOWN_LEVEL ) return CCT.T_UNSIGNED_8 is
   begin
      case LEVEL is
   when 0 =>
        return CCT.T_UNSIGNED_8'last;
   when 16#10#..16#2E# =>
        return K_MODE_SETUP;
   when others =>
        return K_MODE_SAFE;
      end case;
   end MAX_ALLOWED_STATE;


   procedure CHECK_TC_CORRECTNESS_VERSUS_STATE(TC : in CIT.T_TCSP) is
   --called by: TASK_TC
   --raises exception if command incorrect.

      PARAM_ID : IASW_COMMON.T_PARAM_ID;
         for PARAM_ID use at TC.DATA.WORDS'address;
pragma If(((PARAM_ID'size+15)/16) > TC.DATA.WORDS'length);
   error: Violated safety of 'use at'
pragma End if;

      TC55_DATA : T_TC55_DATA;
         for TC55_DATA use at TC.DATA.WORDS'address;
pragma If(((TC55_DATA'size+15)/16) > TC.DATA.WORDS'length);
   error: Violated safety of 'use at'
pragma End if;

   begin
      if TC.DAT_HEAD = (TYP => 5, SUB => 3) then
         if TC.DATA.LEN < 1 then  -- to be sure that TID/FID is given
            raise IASW_MAIN.EX_LENGTH;
         end if;

         case PARAM_ID.TID is
      when 0 =>    -- directed to DFEE
            case PARAM_ID.FID is
         when 31..34 | 4 | 5 =>   -- HV settings, Ready, On
               if DEEPER_SHUTDOWN_LEVEL > 0 then
                  raise IASW_MAIN.EX_SAFETY;
               end if;

         when 7 =>     -- Load Context command
               if IASW_DFEE.GET_STATE > IASW_DFEE.K_STATE_SETUP then
                  raise IASW_MAIN.EX_SAFETY;
               end if;

         when 8 =>     -- Save Context command
               if IASW_DFEE.GET_STATE > IASW_DFEE.K_STATE_SETUP then
                  raise IASW_MAIN.EX_SAFETY;
               end if;

         when others =>
               null;
            end case;

      when others =>
            null;
         end case;

      elsif TC.DAT_HEAD = (TYP => 5, SUB => 5) then
         if TC.DATA.LEN < 1 then  -- to be sure that MODE/- is given
            raise IASW_MAIN.EX_LENGTH;
         elsif (TC55_DATA.MODE > MAX_ALLOWED_STATE(DEEPER_SHUTDOWN_LEVEL)) then
            raise IASW_MAIN.EX_STATE;
         end if;

      else
         null;
      end if;
   end CHECK_TC_CORRECTNESS_VERSUS_STATE;


   procedure CONSTRUCTOR is
   begin
      --DFEE_LIVED_AT_LEAST_ONCE := FALSE;
      --in ver.1.63 moved to IASW_DFEE

      AUTO_RECOVERY_CONTROL_INFO := (
         RECOVERY_LEVEL => K_RECOVER_CONTEXT,
         CPU_MODE       => 3
      );
      NEED_HV_ON := FALSE;
      LOW_VOLTAGE_FAILURE := FALSE;
      PRESSURE_FAILURE := FALSE;
      TEMP_FAILURE := FALSE;

      --LV_FAILURE_CNT := 0;
      --deleted in ver.1.62 (this variable no more used)

      BCPKT_VALID := FALSE;
      INTERVAL_INIT(ECLIPSE_INTERVAL);
      INTERVAL_INIT(RADIATION_BELTS_INTERVAL);
      RADMON_INIT(RADMON(1), 5000, 3000);
      RADMON_INIT(RADMON(2), 5000, 3000);
      RADMON_INIT(RADMON(3), 5000, 3000);

      BCPKT_FILTER    := (others => FALSE);
      BCPKT_CONDITION := (others => FALSE);
      HK_FILTER       := (others => FALSE);
      HK_CONDITION    := (others => FALSE);

      WANTED_SHUTDOWN_LEVEL := 0;
      ACTUAL_SHUTDOWN_LEVEL := 0;
      MINIMAL_SHUTDOWN_LEVEL := 0;
      INTERRUPTED_DFEE_STATE := IASW_DFEE.K_STATE_STANDBY;
      PENDING_ANALYSE := FALSE;
      PREV_VALID_PID := (0,0);
      PREV_PID := (0,0);
      PREV_OTF := FALSE;

      -- this needed for the case of use of FF00 in
      -- the very first command "goto data taking".
      SCF1 := IASW_MODES.K_SCF_FULL;
      SCF2 := IASW_MODES.K_SCF_FULL;
      GFL1 := IASW_HSL.DEFAULT_PRIMARY_MIN_GF;
      GFL2 := IASW_HSL.DEFAULT_SECONDARY_MAX_GF;

      -- reset control info for generation of OEM(recovery)
      REC_OEM_CONTROL.OEM_SENT := FALSE;
      REC_OEM_CONTROL.ASL := 0;
      REC_OEM_CONTROL.WSL := 0;
   end CONSTRUCTOR;


end IASW_STATE;
