----------------------------------------------------------------------------
-- Module Name:        IASW_TIME - package body
-- Project:            INTEGRAL, JEM-X, DPE
-- Author:             Grzegorz Juchnikowski
-- Creation Date:      3.III.1998
-- Change History:
--    (Version/Date/Author/Description)
-- Function:
--    (Say what the component does)
--    The purpose of procedures in this package is to ease the
--    synchronization of the data events obtained from DFEE with
--    On-Board Time (OBT).
-- Interfaces:
--    (Define the inputs and outputs)
-- Dependencies:
--    (Describe the preconditions for using this component)
-- Processing:
--    (Summarise the processing using pseudo-code or PDL)
----------------------------------------------------------------------------

with SYSTEM;
with UNCHECKED_CONVERSION;
with CSSW_COMMON_TYPES;   use CSSW_COMMON_TYPES;
with CSSW_IF_TYPES;
with IASW_COMMON;
with DPE_CSSW;

package body IASW_TIME is

   procedure BCP1_TICK is
   begin
      -- we are reading OBT related to the last BCP4(=BCP2) from before BCP1,
      -- but the interesting one is the BCP4 after BCP1, which is 1 sec after
      -- the measured one.
      BCP1_OBT := DPE_CSSW.GET_BCP4_TIME + SEC_TO_OBT(SEC => 1, FSEC => 0);
   end BCP1_TICK;


   procedure SYNCHRONIZE_DFEE_WITH_OBT( NEW_DFEE_HK_CC : in CCT.T_UNSIGNED_16 ) is
   begin
      loop
         exit when IASW_COMMON.CAN_ENTER_CRITSEC(TIME_SEMAPHORE'address);
         DPE_CSSW.DELAY_RELATIVE(1);
      end loop;

      DFEE_HK_CC_STORED := NEW_DFEE_HK_CC;
      DFEE_OBT_STORED := BCP1_OBT;

      TIME_SEMAPHORE := 0;
   end SYNCHRONIZE_DFEE_WITH_OBT;


   function GET_OBT_OF_EVENT(
      DFEE_HK_CC        : in CCT.T_UNSIGNED_16;
      DFEE_TIME_COUNTER : in CCT.T_UNSIGNED_16
   ) return CIT.T_OBT is
      EX_TOO_BIG_DIFFERENCE : exception;
      BCP1no : INTEGER;
      CC0  : CCT.T_UNSIGNED_16;
      OBT0 : CIT.T_OBT;
   begin
      loop
         exit when IASW_COMMON.CAN_ENTER_CRITSEC(TIME_SEMAPHORE'address);
         DPE_CSSW.DELAY_RELATIVE(1);
      end loop;

      CC0 := DFEE_HK_CC_STORED;
      OBT0 := DFEE_OBT_STORED;

      TIME_SEMAPHORE := 0;

      -- calculate number of BCP1 periods between event and last synchro
      -- BCP1no := INTEGER( DFEE_HK_CC - CC0 );
      begin
         if DFEE_HK_CC >= CC0 then
            if (DFEE_HK_CC-CC0) > 4000 then
               raise EX_TOO_BIG_DIFFERENCE;
            end if;
            BCP1no := INTEGER(DFEE_HK_CC-CC0);
         else
            if (CC0-DFEE_HK_CC) > 4000 then
               raise EX_TOO_BIG_DIFFERENCE;
            end if;
            BCP1no := - INTEGER(CC0-DFEE_HK_CC);
         end if;
      exception
         when others =>    -- only EX_TOO_BIG_DIFFERENCE is possible
            return (0,0,0);
      end;

      return OBT0 + SEC_TO_OBT(
         SEC  => IASW_COMMON.INT2LONG(
                 BCP1no * 8 + INTEGER(DFEE_TIME_COUNTER / 16#2000#)),
         FSEC => (DFEE_TIME_COUNTER mod 16#2000#) * 8 );
   end GET_OBT_OF_EVENT;


   function SEC_TO_OBT(
      SEC  : in LONG_INTEGER;
      FSEC : in CCT.T_UNSIGNED_16 := 0
   ) return CIT.T_OBT is
      subtype T_B6 is IASW_COMMON.T_PACKED_ARRAY_OF_BYTES(0..5);
      subtype T_B4 is IASW_COMMON.T_PACKED_ARRAY_OF_BYTES(0..3);
      subtype T_B2 is IASW_COMMON.T_PACKED_ARRAY_OF_BYTES(0..1);
      T : T_B6;
      --Z : CIT.T_OBT;
      --   for Z use at T'address;
      BSEC : T_B4;
      --   for BSEC use at SEC'address;
      BFSEC : T_B2;
      --   for BFSEC use at FSEC'address;

      function CVT is new UNCHECKED_CONVERSION(
         SOURCE => LONG_INTEGER, TARGET => T_B4);
      function CVT is new UNCHECKED_CONVERSION(
         SOURCE => CCT.T_UNSIGNED_16, TARGET => T_B2);
      function CVT is new UNCHECKED_CONVERSION(
         SOURCE => T_B6, TARGET => CIT.T_OBT);
   begin
      BSEC  := CVT(SEC);
      BFSEC := CVT(FSEC);
      T(0) := BSEC(1);
      T(1) := BSEC(2);
      T(2) := BSEC(3);
      T(3) := BFSEC(0);
      T(4) := BFSEC(1);
      T(5) := 0;
      return CVT(T);  --Z;
   end SEC_TO_OBT;


   function "+"( A,B : in CIT.T_OBT ) return CIT.T_OBT is
      Z : CIT.T_OBT;
      C : BOOLEAN;
   begin
      C := FALSE;
      IASW_COMMON.ADDC(A.LOW,  B.LOW,  Z.LOW,  C);
      IASW_COMMON.ADDC(A.MID,  B.MID,  Z.MID,  C);
      IASW_COMMON.ADDC(A.HIGH, B.HIGH, Z.HIGH, C);
      return Z;
   end "+";


   function "<"( A,B : in CIT.T_OBT ) return BOOLEAN is
   -- returns TRUE if A<B.  if A and B differ more than half of full
   -- scale, then the result is reversed - it is assumed that the smaller
   -- is one full cycle younger.
   begin
      if A.HIGH < B.HIGH then
         if (B.HIGH - A.HIGH) < 16#8000# then
            return TRUE;
         end if;
      elsif A.HIGH > B.HIGH then
         if (A.HIGH - B.HIGH) >= 16#8000# then
            return TRUE;
         end if;
      else  -- A.HIGH = B.HIGH
         if A.MID < B.MID then
            return TRUE;
         elsif A.MID = B.MID then
            if A.LOW < B.LOW then
               return TRUE;
            end if;
         end if;
      end if;
      return FALSE;
   end "<";


   procedure CONSTRUCTOR is
   begin
      null;
   end CONSTRUCTOR;

end IASW_TIME;
